"""
Is the base set of sentries
Includes alignment and aligned distances
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de




import os
import os.path
import cPickle
import math
import copy

import Numeric
import MLab

import set_alig
import log_util
import scop_sets
import sus_dpi
import rpy


def compile_basesets(config_dic):
    """
    Read SCOP sets defintion, generates basesets and saves them
    """
    # get baseset
    read_scop = scop_sets.ReadScopBaseset(config_dic['min_set_len'])
    read_scop.load_file(config_dic['sets_file'])
    while(read_scop.line_counter < read_scop.nr_lines):
        last_lnr = read_scop.line_counter
        try:
            use_pdb_subdir = True
            baseset = read_scop.get_baseset(config_dic['pdb_dir'], use_pdb_subdir)    
            # align
            baseset.get_alignment(config_dic['stmap_dir'],config_dic['uniprot_dir'])
            # get aligned distance matrices, 
            baseset.get_alig_dist_array()
            # save
            save_baseset(baseset, config_dic['baseset_dir'])
        except:
            log_util.log_error('Could not compile baseset, input line nr:%d' % read_scop.line_counter)
            if read_scop.line_counter == last_lnr:
                # if same line on error, then increase line count
                read_scop.line_counter += 1
            continue



def save_baseset(baseset, baseset_dir):
    """
    Save basest object in baseset_dir
    """    
    pickle_protocol = 2 # use binary
    data_file = baseset.set_id + '.bset'
    path_data_file = os.path.join(baseset_dir, data_file)
    outf_obj = file(path_data_file, 'w')
    cPickle.dump(baseset, outf_obj, pickle_protocol)
    outf_obj.close()
    log_util.log_info('Saved baseset: %s' % path_data_file)


def load_baseset(set_id, baseset_dir):
    """
    Load baseset object file, given set_id
    """
    data_file = set_id + '.bset'
    path_data_file = os.path.join(baseset_dir, data_file)
    f_obj = file(path_data_file, 'r')
    baseset = cPickle.load(f_obj)
    f_obj.close()
    log_util.log_info('Loaded baseset: %s' % path_data_file)
    return baseset



def get_setsid(setsid_file):
    """
    Get list of sets id from setsid_file
    """
    f_obj = file(setsid_file, 'r')    
    line_lst = f_obj.readlines()
    f_obj.close()
    
    setsid_lst = []
    for line_st in line_lst:
        setsid_lst.append(line_st.strip())
    return setsid_lst



class BaseSet:
    """
    This is a Sentry set, with SetAlig, dist_mat_dic and alig_dist_array
    This is a data unit for clustering and set comparison
    Can be saved into a file
    """

    def __init__(self, sentryset):
        self.set_id = sentryset.set_id
        self.set_len = sentryset.set_len
        self.working_dir = None
        self.sentryset = sentryset
        # derived data
        self.alig = None
        self.dpi_dic = {}    # DPI dictionary for sentries with DPI, self.dpi_dic[sentry_id] = dpi_obj
        self.sentry_sus_lst = []
        self.sentry_nosus_lst = []
        # self.dist_mat_dic = {}  # dictionary of distance matrices dist_mat_dic[sentry_id] = dist_mat
        self.alig_dist_array = None
        self.alig_gc_dist_array = None   # side chain geometric center
        self.alig_sus_array = None
        self.alig_gc_sus_array = None
        self.pickle_protocol = 2 # use binary

        
        
    def reset_data(self):
        """
        Reset baseset, delete derived data
        """
        # derived data
        self.alig = None
        self.dpi_dic = {}    # DPI dictionary for sentries with DPI, self.dpi_dic[sentry_id] = dpi_obj
        self.sentry_sus_lst = []
        self.sentry_nosus_lst = []
        # self.dist_mat_dic = {}  # dictionary of distance matrices dist_mat_dic[sentry_id] = dist_mat
        self.alig_dist_array = None
        self.alig_gc_dist_array = None
        self.alig_sus_array = None
        self.alig_gc_sus_array = None        
        
        
    def delete_sentries(self, sentry_id_lst):
        """
        Delete the sentry from baseset
        reset the baseset
        """
        for sentry_id in sentry_id_lst:
            self.sentryset.delete_sentry(sentry_id)
            self.set_len = self.sentryset.set_len
            self.reset_data()
        
            
            
    def set_working_dir(self, base_dir):
        """
        Set the current working directory
        and make the directory        
        """
        set_dir = self.set_id + '_sunid'
        self.working_dir = os.path.join(base_dir,set_dir)
        os.mkdir(self.working_dir)
        
        
        
    def set_working_to_current_dir(self):
        self.working_dir = "."
        
        
        
    def get_alignment(self, stmap_dir, uniprot_dir):
        """
        Get alignment
        """
        # start set alignment object
        self.alig = set_alig.SetAlig()
        self.alig.get_alignment(self.sentryset, stmap_dir, uniprot_dir)
        # write alignment as fasta file
        # self.alig.write_fasta_alig()

       
         
    def get_aligment_from_msa(self, msa_file):
        """
        Read alignment from MSA file
        """
        self.alig = set_alig.SetAlig()
        self.alig.get_aligment_from_msa(self.sentryset, msa_file)

        
        
    def get_alignment_from_file(self, alig_filename):
        """
        Read alignment from file generated with write_alignment
        """
        self.alig = set_alig.SetAlig()
        self.alig.get_alignment_from_file(self.sentryset, alig_filename)
        
        

    def get_alig_dist_array(self, get_dmat = True, mat_dir = None):
        """
        Calculate the distance matrices and align them
        Computes alig_dist_array
        """
        if get_dmat:
            # Compute
            dist_mat_dic = self.get_ca_dist_mat()
        else:
            # Load precomputed distance matrices
            dist_mat_dic = self.load_ca_dist_mat(mat_dir)
        # align distance matrices
        self.alig_dist_array = self.alig.alig_mat(dist_mat_dic)
        log_util.log_info("Calculated aligned distance array set:%s" % self.set_id)

        
        
    def get_alig_gc_dist_array(self):
        gc_dist_mat_dic = self.get_gc_dist_mat()
        self.alig_gc_dist_array = self.alig.alig_mat(gc_dist_mat_dic)
        log_util.log_info("Calculated aligned GC distance array set:%s" % self.set_id)
        
        

    def get_ca_dist_mat(self):   
        """
        Compute distance matrices
        """
        dist_mat_dic = {}
        for sentry in self.sentryset.sentry_lst:
            dist_mat = sentry.ca_dist_mat()
            dist_mat_dic[sentry.sentry_id] = dist_mat
        log_util.log_info("Calculated distance matrices set:%s" % self.set_id)
        return dist_mat_dic       
        
        
        
    def get_gc_dist_mat(self):   
        """
        Compute distance matrices for side chain geometric centers
        """
        dist_mat_dic = {}
        for sentry in self.sentryset.sentry_lst:
            dist_mat = sentry.gc_dist_mat()
            dist_mat_dic[sentry.sentry_id] = dist_mat
        log_util.log_info("Calculated distance GC matrices set:%s" % self.set_id)
        return dist_mat_dic       
        
        
        
    def load_ca_dist_mat(self, mat_dir):
        """
        Load distance matrices from pickled objects
        If failed compute
        """
        dist_mat_dic = {}
        for sentry in self.sentryset.sentry_lst:
            sentry_id = sentry.sentry_id
            dist_mat_file = self._get_dmat_file(sentry_id, mat_dir)
            if os.path.exists(dist_mat_file):
                f_obj = file(dist_mat_file, 'r')
                dist_mat = cPickle.load(f_obj)
                f_obj.close()
                dist_mat_dic[sentry.sentry_id] = dist_mat
            else:
                # compute it
                dist_mat = sentry.ca_dist_mat()
                dist_mat_dic[sentry.sentry_id] = dist_mat
                log_util.log_warning("Could not find %s. Calculated distance matrix:%s" % (dist_mat_file,sentry_id))
        log_util.log_info("Loaded distance matrices set:%s" % self.set_id)
        return dist_mat_dic
        
    
    
    def save_ca_dist_mat(self, mat_dir):
        """
        Save Ca distance matrix, pickle the object
        """
        dist_mat_dic = self.get_ca_dist_mat()
        for sentry in self.sentryset.sentry_lst:
            sentry_id = sentry.sentry_id
            dist_mat = dist_mat_dic[sentry_id]
            dist_mat_file = self._get_dmat_file(sentry_id, mat_dir)
            f_obj = file(dist_mat_file, 'w')
            cPickle.dump(dist_mat, f_obj, self.pickle_protocol) 
            f_obj.close()
            log_util.log_info('Saved %s' % dist_mat_file)
            
            
            
    def _get_dmat_file(self, sentry_id, mat_dir):
        file_name = sentry_id + '.dmat'
        dist_mat_file =  os.path.join(mat_dir,file_name)
        return dist_mat_file
        
                
        
    def write_dist_mat(self, dist_mat_dic):
        """
        Debug, write dist_mat.
        """
        for sentry in self.sentryset.sentry_lst:
            dist_mat = dist_mat_dic[sentry.sentry_id]
            # column label
            out_str = "%7s   " % sentry.sentry_id
            for residue in sentry.res_lst:
                residue_str = " %5s:%s  " % (residue.label, residue.chain_id)
                out_str =  out_str + residue_str
            print out_str
            for i in range(sentry.length):
                # row label
                out_str = ""
                residue_i = sentry.res_lst[i]
                out_str = " %5s:%s  " % (residue_i.label, residue_i.chain_id)
                for j in range(sentry.length):                    
                    # values
                    coord_dist = dist_mat[i,j]
                    dist_str = "%8.3f  " % coord_dist
                    out_str = out_str + dist_str
                print out_str

    
    
    def get_alig_sus(self,dpi_file = None, get_dsusmat = True, mat_dir = None):
        """
        Get coordinate sus
        get matrix of distances sus
        align
        """
        if get_dsusmat:
            # Compute
            sus_mat_dic = self.get_dist_sus_mat(dpi_file, 'CA')
        else:
            # Load precomputed distance sus matrices
            sus_mat_dic = self.load_dist_sus_mat(mat_dir)
        self.alig_sus_array = self.alig.alig_mat(sus_mat_dic)
        log_util.log_info("Calculated aligned distance sus array set:%s" % self.set_id)


        
    def get_alig_gc_sus(self,dpi_file = None):
        """
        Get coordinate sus for side chain geometric center
        get matrix of distances sus
        align
        """
        sus_mat_dic = self.get_dist_sus_mat(dpi_file, 'GC')
        self.alig_gc_sus_array = self.alig.alig_mat(sus_mat_dic)
        log_util.log_info("Calculated aligned GC distance sus array set:%s" % self.set_id)
        
        
            
    def get_dist_sus_mat(self, dpi_file, atom_sel):
        """
        compute the distance sus matrix
        """
        sus_mat_dic = {}
        self._get_dpi_dic(dpi_file)
        for sentry_id in self.sentry_sus_lst:
            sentry = self.sentryset.get_sentry(sentry_id)
            sus_ok = sentry.set_sus(self.dpi_dic[sentry_id])
            if sus_ok == 0:
                del self.dpi_dic[sentry_id]
                self.sentry_sus_lst.remove(sentry_id)
                self.sentry_nosus_lst.append(sentry_id)
                log_util.log_error('Could not set sus for %s' % sentry_id)
                continue
            # distance sus
            if atom_sel == 'GC':
                dist_sus_mat = sentry.dist_gc_sus_mat()
            else:
                dist_sus_mat = sentry.dist_sus_mat()
            sus_mat_dic[sentry_id] = dist_sus_mat
        if atom_sel == 'GC':
            log_util.log_info("Calculated GC distance sus matrices set:%s" % self.set_id)        
        else:
            log_util.log_info("Calculated distance sus matrices set:%s" % self.set_id)
        return sus_mat_dic
    
        
        
    def load_dist_sus_mat(self, mat_dir):
        """
        Load distance sus matrices from pickled objects
        If failed append to self.sentry_nosus_lst
        """
        sus_mat_dic = {}
        for sentry in self.sentryset.sentry_lst:
            sentry_id = sentry.sentry_id
            dist_sus_mat_file = self._get_dsmat_file(sentry_id, mat_dir)
            if os.path.exists(dist_sus_mat_file):
                f_obj = file(dist_sus_mat_file, 'r')
                dist_sus_mat = cPickle.load(f_obj)
                f_obj.close()
                sus_mat_dic[sentry_id] = dist_sus_mat
                self.sentry_sus_lst.append(sentry_id)
            else:
                self.sentry_nosus_lst.append(sentry_id)
        log_util.log_info("Sentries with SUS: %s" % str(self.sentry_sus_lst))
        log_util.log_info("Sentries without SUS: %s" % str(self.sentry_nosus_lst))
        log_util.log_info("Loaded distance sus matrices set:%s" % self.set_id)
        return sus_mat_dic
    
        
    def save_dist_sus_mat(self, dpi_file, mat_dir):
        """
        Save distance sus matrix as pickled object
        """
        sus_mat_dic = self.get_dist_sus_mat(dpi_file)
        for sentry_id in self.sentry_sus_lst:
            dist_sus_mat = sus_mat_dic[sentry_id]
            dist_sus_mat_file = self._get_dsmat_file(sentry_id, mat_dir)
            f_obj = file(dist_sus_mat_file, 'w')
            cPickle.dump(dist_sus_mat, f_obj, self.pickle_protocol) 
            f_obj.close()
            log_util.log_info('Saved %s' % dist_sus_mat_file)
        

    def write_sequences_fasta(self, uniprot_dir, uniprot_id_lst , fasta_file):
        """
        Write all the uniprot and all sentry sequences into fasta file
        """
        self.alig = set_alig.SetAlig()
        self.alig.write_sequences_fasta(self.sentryset, uniprot_dir, uniprot_id_lst, fasta_file)
            
    
            
    ##########################################################################        
    # collect and plot different types of data
    
    
    def source_r(self, source_dir):
        r_file = os.path.join(source_dir, 'make_plots.R')
        rpy.r.source(file = r_file)

    
        
    # zdist
    def zdist_stats(self, subset_lst, subset_name, source_dir, atom_sel):
        """
        Compare zscores of distances to t distribution in a q-q plot
        and plot histogram of zscores        
        """
        if atom_sel == 'CA' and self.alig_dist_array == None:
            log_util.log_error('Cannot compare distance distribution, no distances')
        elif atom_sel == 'GC' and self.alig_gc_dist_array == None:
            log_util.log_error('Cannot compare distance distribution, no distances')
        zscore_subset = self._get_zdist(subset_lst, atom_sel)
        self.qqplot(zscore_subset, subset_name, source_dir)
        self.zdist_hist_plot(zscore_subset, subset_name, source_dir)
        log_util.log_info('Finished distance stats on %s' % subset_name)
        return zscore_subset
        
        
        
    def qqplot(self, zscore_subset, subset_name, source_dir):
        qqplot_file = subset_name + '_zdist_qqplot_.png'
        self.source_r(source_dir)
        rpy.r.png(filename = qqplot_file)
        tmp = rpy.r.qqnorm(y=zscore_subset, main = qqplot_file)
        rpy.r.qqline(y=zscore_subset)
        rpy.r.dev_off()
        
        
        
    def zdist_hist_plot(self, zscore_subset, subset_name, source_dir):
        density_file = subset_name + '_zdist_hist.png'
        self.source_r(source_dir)
        rpy.r.png(filename=density_file)
        hist_title = "zscore %s" % subset_name
        tmp = rpy.r.hist(x=zscore_subset, xlab = '', ylab='', main = density_file, breaks = 100)
        # density_data = rpy.r.density(zscore_subset)
        # rpy.r.lines(density_data)
        rpy.r.dev_off()
        
        
    # sus
    def sus_boxplots(self, boxplot_name):
        """
        Make boxplots of sus for all baseset
        """
        sentry_id_lst = self.sentryset.sentryid_dic.keys()
        self.sus_boxplots_subset(sentry_id_lst, boxplot_name)
        
        
        
    def sus_boxplots_subset(self, subset_lst, boxplot_name, source_dir):
        """
        Make boxplots of carbon alpha sus for entries in subset_lst
        """
        rev_subset_lst = copy.deepcopy(subset_lst)
        rev_subset_lst.reverse()
        
        if self.alig_sus_array == None:
            log_util.log_error('Cannot get sus boxplots, no aligned sus')
        # get dictionary of entries without sus
        sentry_nosus_dic = {}
        for sentry_id in self.sentry_nosus_lst:
            sentry_nosus_dic[sentry_id] = sentry_id
        sentry_sus_lst = []
        sentry_id_lst = []
        for sentry_id in rev_subset_lst:
            if sentry_nosus_dic.has_key(sentry_id):
                # skip if no sus
                continue
            sentry = self.sentryset.get_sentry(sentry_id)
            sus_v = sentry.get_CA_sus_v()
            sentry_sus_lst.append(sus_v)
            sentry_id_lst.append(sentry_id)

        self.source_r(source_dir)
        boxplot_file = boxplot_name + '.png'
        summary_str = rpy.r.susboxplot(suslst=sentry_sus_lst, idlst=sentry_id_lst, filename=boxplot_file)       
        log_util.log_info("Generated sus boxplot %s" % boxplot_file)
        boxplot_file = boxplot_name + '.pdf'
        summary_str = rpy.r.susboxplotpdf(suslst=sentry_sus_lst, idlst=sentry_id_lst, filename=boxplot_file)
        log_util.log_info("Generated sus boxplot %s" % boxplot_file)
        log_util.log_info("Sus: %s" % summary_str)
          
        
          
    # max relative distance difference X
    def get_x_mat(self, subset_lst, atom_sel = 'CA'):
        # if not defined then -1
        nr_sentries = len(subset_lst)
        x_mat = Numeric.zeros((self.alig.length, self.alig.length), Numeric.Float)
        if atom_sel == 'GC':
            dist_array = self.alig_gc_dist_array
            sus_array = self.alig_gc_sus_array
        else:
            dist_array = self.alig_dist_array
            sus_array = self.alig_sus_array
        for i in range(self.alig.length):
            for j in range(i):
                rd_lst = []
                for a_counter in range(nr_sentries):
                    a_id = subset_lst[a_counter]
                    a_dist = self._get_dist(a_id, i, j, dist_array)
                    a_distsusv = self._get_distsusv(a_id, i, j, sus_array)
                    for b_counter in range(a_counter):
                        b_id = subset_lst[b_counter]
                        b_dist = self._get_dist(b_id, i, j, dist_array)
                        b_distsusv = self._get_distsusv(b_id, i, j, sus_array)
                        if a_dist >= 0.0 and b_dist >= 0.0 and a_distsusv >= 0.0 and b_distsusv>= 0.0:
                            distdiff_sus = math.sqrt(a_distsusv + b_distsusv)
                            if distdiff_sus > 0.0:
                                rd_val = abs(a_dist-b_dist)/distdiff_sus
                            else:                            
                                rd_val = -1.0
                        else:
                            rd_val = -1.0
                        rd_lst.append(rd_val)
                x_val = max(rd_lst)
                x_mat[i,j] = x_val
                x_mat[j,i] = x_val
        return x_mat
                            
                        
        
    # dist sdev
    def get_dist_sdev_mat(self, subset_lst, atom_sel = 'CA'):
        # if not defined then -1
        nr_sentries = len(subset_lst)
        dist_sdev_mat = Numeric.zeros((self.alig.length,self.alig.length), Numeric.Float)
        if atom_sel == 'GC':
            alig_array = self.alig_gc_dist_array
        else:
            alig_array = self.alig_dist_array
        for i in range(self.alig.length):
            for j in range(i):
                dist_lst = []
                for sentry_id in subset_lst:
                    dist = self._get_dist(sentry_id, i, j, alig_array)
                    if dist >= 0.0:
                        dist_lst.append(dist)
                if len(dist_lst) > 0:
                    if len(dist_lst) == 1:
                        dist_sdev = 0
                    else:
                        dist_v = Numeric.array(dist_lst, Numeric.Float)
                        dist_sdev = MLab.std(dist_v)
                else:
                    dist_sdev = -1.0
                dist_sdev_mat[i,j] = dist_sdev
                dist_sdev_mat[j,i] = dist_sdev
        return dist_sdev_mat
                
                
        
    # square root of mean distance sus variance                
    def get_sqrtmean_dsusv_mat(self, subset_lst, atom_sel = 'CA'):
        # if not defined then -1
        sqrtmean_dsusv_mat = Numeric.zeros((self.alig.length,self.alig.length), Numeric.Float)
        if atom_sel == 'GC':
            alig_sus_array = self.alig_gc_sus_array
        else:
            alig_sus_array = self.alig_sus_array
        for i in range(self.alig.length):
            for j in range(i):
                dsusv_lst = []
                for sentry_id in subset_lst:
                    dsusv = self._get_distsusv(sentry_id, i, j, alig_sus_array)
                    if dsusv >=0.0:
                        dsusv_lst.append(dsusv)
                if len(dsusv_lst) > 0.0:
                    dsusv_v = Numeric.array(dsusv_lst, Numeric.Float)
                    sqrtmean_dsusv = math.sqrt(MLab.mean(dsusv_v))
                else:
                    sqrtmean_dsusv = -1.0
                sqrtmean_dsusv_mat[i,j] = sqrtmean_dsusv
                sqrtmean_dsusv_mat[j,i] = sqrtmean_dsusv
        return sqrtmean_dsusv_mat
        

    # max distance sus            
    def get_max_dsus_mat(self, subset_lst, atom_sel = 'CA'):
        # if not defined then -1
        max_dsus_mat = Numeric.zeros((self.alig.length,self.alig.length), Numeric.Float)
        if atom_sel == 'GC':
            alig_sus_array = self.alig_gc_sus_array
        else:
            alig_sus_array = self.alig_sus_array
        for i in range(self.alig.length):
            for j in range(i):
                dsus_lst = []
                for sentry_id in subset_lst:
                    dsus = self._get_distsus(sentry_id, i, j, alig_sus_array)
                    if dsus >=0.0:
                        dsus_lst.append(dsus)
                if len(dsus_lst) > 0:
                    dsus_v = Numeric.array(dsus_lst, Numeric.Float)
                    dsus_max = MLab.max(dsus_v)
                else:
                    dsus_max = -1.0
                max_dsus_mat[i,j] = dsus_max
                max_dsus_mat[j,i] = dsus_max
        return max_dsus_mat
                
                
    # relative distance sdev            
    def get_rel_dist_sdev_mat(self, dist_sdev_mat, n_mat):
        # get relative dist sdev mat
        # where values are deviedd by n_mat
        
        # 0 in n_mat convert to -1
        mask_array = Numeric.less_equal(n_mat, 0.0)
        Numeric.putmask(n_mat, mask_array, -1.0)
        
        # relative dist sdev
        rel_dist_sdev_mat = dist_sdev_mat/n_mat
        
        # set not defined to -1
        mask_array = Numeric.less(n_mat, 0.0)
        Numeric.putmask(rel_dist_sdev_mat, mask_array, -1.0)
        
        mask_array = Numeric.less(dist_sdev_mat, 0.0)
        Numeric.putmask(rel_dist_sdev_mat, mask_array, -1.0)
        
        # diagonal to 0
        dm = Numeric.identity(rel_dist_sdev_mat.shape[0])
        Numeric.putmask(rel_dist_sdev_mat, dm, 0.0)
        
        return rel_dist_sdev_mat
        
        
    
    # dist sdev and sus   
    def get_dist_sus_sdev_mat(self, subset_lst, dist_sdev_mat, atom_sel='CA'):
        # if not defiedn then -1
        nr_sentries = len(subset_lst)
        dist_sus_sdev_mat = Numeric.zeros((self.alig.length,self.alig.length), Numeric.Float)
        if atom_sel == 'GC':
            alig_sus_array = self.alig_gc_sus_array
        else:
            alig_sus_array = self.alig_sus_array
        for i in range(self.alig.length):
            for j in range(i):
                susv_lst = []
                for sentry_id in subset_lst:
                    susv = self._get_distsusv(sentry_id, i, j, alig_sus_array)
                    if susv >=0.0:
                        susv_lst.append(susv)
                dist_sdev = dist_sdev_mat[i,j]
                if len(susv_lst) > 0 and dist_sdev >= 0.0:
                    susv_v = Numeric.array(susv_lst, Numeric.Float)
                    susv_mean = MLab.mean(susv_v)
                    dist_sus_sdev= math.sqrt(math.pow(dist_sdev ,2) + susv_mean)
                else:
                    dist_sus_sdev = -1.0
                dist_sus_sdev_mat[i,j] = dist_sus_sdev
                dist_sus_sdev_mat[j,i] = dist_sus_sdev
        return dist_sus_sdev_mat
                

    # sdev ratio            
    def get_sdev_ratio_v(self, subset_lst, dist_sdev_mat, dist_sus_sdev_mat):
        """
        Ratio between distance sdev and distance and mean sus sdev
        """
        nr_sentries = len(subset_lst)
        r_lst = []
        for i in range(self.alig.length):
            for j in range(i):
                dist_sdev = dist_sdev_mat[i,j]
                dist_sus_sdev = dist_sus_sdev_mat[i,j]
                if dist_sdev >= 0.0 and dist_sus_sdev > 0.0:
                    r =  dist_sdev / dist_sus_sdev
                    r_lst.append(r)
                if dist_sus_sdev == 0.0:
                    # debug
                    log_util.log_warning("dist_sus_sdev is 0")
        dist_sdev_to_distsus_sdev_ratio_v = Numeric.array(r_lst, Numeric.Float)
        return dist_sdev_to_distsus_sdev_ratio_v                    

        
        
    def get_undef_mat(self, mat):
        """
        Get undefined values in matrix (-1)
        undefined values are set to 1, otherwise they are 0
        """
        undef_mat = Numeric.less(mat, 0)
        return undef_mat
    
        
            
    def get_submat(self, sentry_id, reskey_lst, mat):
        """
        return a submatrix of mat corresponding to the aligned positions of reskey_lst from sentry_id 
        """
        n = Numeric.shape(mat)[0]
        refpos_lst = self.reskey_to_refpos(sentry_id, reskey_lst)
        # refpos_lst.sort()
        submat_len = len(refpos_lst)
        # print "submat_len %d" % submat_len   # debug
        submat = Numeric.zeros((submat_len,submat_len), Numeric.Float)

        i = 0
        for i_refpos in refpos_lst:
            j = 0
            for j_refpos in refpos_lst:
                # print "i_refpos %d j_refpos %d, i %d, j %d" % (i_refpos,j_refpos, i, j) # debug
                submat[i,j] = mat[i_refpos,j_refpos] 
                j += 1            
            i += 1
        return submat
        
            
            
    def reskey_to_refpos(self, sentry_id, reskey_lst):
        refpos_lst = []
        sentry = self.sentryset.get_sentry(sentry_id)
        for reskey in reskey_lst:
            if sentry.res_dic.has_key(reskey):
                sentrypos = sentry.res_dic[reskey]
                if self.alig.sentry_refpos.has_key(sentry_id):
                    if self.alig.sentry_refpos[sentry_id].has_key(sentrypos):                        
                        refpos = self.alig.sentry_refpos[sentry_id][sentrypos]
                        refpos_lst.append(refpos)
                    else:
                        log_util.log_warning('No refpos %s, sentrypos %s in sentry_refpos for %s, skipping' % (str(refpos), str(sentrypos) ,sentry_id))
                        continue
                else:
                    log_util.log_warning('No sentry_id %s in sentry_refpos, skipping' % sentry_id)
                    continue                
            else:
                log_util.log_warning('No reskey %s in %s, skipping' % (reskey, sentry_id))
                continue
        return refpos_lst
            
    
    
    def get_aligned_reskeys(self, sentry_id, reskey_lst, sentry_id_lst):
        """
        given a list of residue keys from a given sentry_id, return a dictionary
        where keys is sentry_id and value is list of aligned reskeys
        """        
        sentrypos_dic = {}
        refpos_lst = self.reskey_to_refpos(sentry_id, reskey_lst)
        for sentry_id in sentry_id_lst:
            sentry = self.sentryset.get_sentry(sentry_id)
            sentrypos_dic[sentry_id] = []
            for refpos in refpos_lst:        
                sentrypos = self.alig.get_sentrypos_from_refpos(refpos, sentry_id)
                reskey = sentry._get_res_key(sentry.res_lst[sentrypos])
                sentrypos_dic[sentry_id].append(reskey)
        return sentrypos_dic

        
        
    def get_selection_for_aligned_reskeys(self, sentry_id, reskey_lst, sentry_id_lst):
        """
        given a list of residue keys from a given sentry_id, return a dictionary
        where keys is sentry_id and value is list of aligned reskeys
        """        
        selection_dic = {}
        refpos_lst = self.reskey_to_refpos(sentry_id, reskey_lst)
        for sentry_id in sentry_id_lst:
            sentry = self.sentryset.get_sentry(sentry_id)
            sel_str = ""
            last_chain = None
            for refpos in refpos_lst:        
                sentrypos = self.alig.get_sentrypos_from_refpos(refpos, sentry_id)
                res_obj = sentry.res_lst[sentrypos]
                chain_id = res_obj.chain_id
                res_label = res_obj.label
                if chain_id != last_chain and chain_id != " ":
                    item_str = "chain %s and resi %s or " % ( chain_id, res_label)
                    last_chain = chain_id
                else:
                    item_str = "resi %s or " % res_label
                sel_str = sel_str + item_str
            selection_dic[sentry_id]= sel_str
        return selection_dic
            
        
            
    def plot_relsdev(self, mat, plot_name):              
        # write mat as R matrix
        mat_file = os.path.join(self.working_dir, plot_name + '.mat')
        self.write_mat_R(mat, mat_file)
        # make plot
        plot_file = os.path.join(self.working_dir, plot_name + '.png')
        rpy.r.plotrelsdev(matfile=mat_file, plotfile=plot_file)    
        log_util.log_info("Saved relative distance sdev matrix and plot: %s %s" % (mat_file,plot_file))



    def plot_variation_X(self, mat, plot_name):              
        # write mat as R matrix
        mat_file = os.path.join(self.working_dir, plot_name + '.mat')
        self.write_mat_R(mat, mat_file)
        # make plot
        plot_file = os.path.join(self.working_dir, plot_name + '.png')
        rpy.r.plotvarx(matfile=mat_file, plotfile=plot_file)    
        log_util.log_info("Saved X matrix and plot: %s %s" % (mat_file,plot_file))
                
                         
        
    def plot_sdev(self, mat, plot_name):              
        # write mat as R matrix
        mat_file = os.path.join(self.working_dir, plot_name + '.mat')
        self.write_mat_R(mat, mat_file)
        # make plot
        plot_file = os.path.join(self.working_dir, plot_name + '.png')
        rpy.r.plotsdev(matfile = mat_file, plotfile = plot_file)    
        log_util.log_info("Saved distance sdev matrix and plot: %s %s" % (mat_file,plot_file))
        
    
        
    def plot_histsdev(self, mat_file, plot_name, source_dir):
        # make histogram plot
        plot_file = plot_name + '_hist.png'
        self.source_r(source_dir)
        rpy.r.histsdev(matfile= mat_file, plotfile = plot_file)
        log_util.log_info("Saved histogram: %s" % (plot_file))
    
        
        
    ##########################################################################        
    # usefull methods                
                                
    def write_mat_R(self, mat, mat_file):
        """
        Write values of float array mat to a text file
        To be read by R as matrix
        undefined values (-1) are set to NA
        """
        f_obj = file(mat_file, 'w')
        dim = Numeric.shape(mat)
        for i in range(dim[0]):
            for j in range(dim[1]):
                if mat[i,j]<0.0:
                    f_obj.write("NA ")
                else:
                    f_obj.write("%f " % mat[i,j])
            f_obj.write('\n')     
        f_obj.close()
        # log_util.log_info('Wrote %s' % mat_file)
        
        
        
    def write_mat_txt(self, mat, mat_file):
        """
        Write values of float array mat to a text file
        """
        f_obj = file(mat_file, 'w')
        dim = Numeric.shape(mat)
        for i in range(dim[0]):
            for j in range(dim[1]):
                f_obj.write("%s " % str(mat[i,j]))
            f_obj.write('\n')     
        f_obj.close()
    
        
        
    def save_obj_r(self, obj, obj_file):
        """
        Write object in R
        """
        rpy.r.saveobj(robj = obj, robjfile = obj_file)
        log_util.log_info('Saved %s' % obj_file)
    

    def get_sentry_id_lst(self):
        sentry_id_lst = []
        for sentry in self.sentryset.sentry_lst:
            sentry_id_lst.append(sentry.sentry_id)
        return sentry_id_lst
            
        
            
    ##########################################################################        
    # private methods
    
    def _get_dist(self, sentry_id, i, j, alig_array):
        sentry_index = self.sentryset.sentryid_dic[sentry_id]
        # coord_dist = self.alig_dist_array[i, j, sentry_index]
        coord_dist = alig_array[i, j, sentry_index]
        return coord_dist
                
    
    
    def _get_distsus(self, sentry_id, i, j, alig_sus_array):
        # return distsus
        sentry_index = self.sentryset.sentryid_dic[sentry_id]
        # distsus = self.alig_sus_array[i, j, sentry_index]
        distsus = alig_sus_array[i, j, sentry_index]
        return distsus

    
                
    def _get_distsusv(self, sentry_id, i, j, alig_sus_array):
        # return variance of distsus
        sentry_index = self.sentryset.sentryid_dic[sentry_id]
        # distsus = self.alig_sus_array[i, j, sentry_index]
        distsus = alig_sus_array[i, j, sentry_index]
        if distsus < 0:
            return distsus
        else:
            return distsus * distsus
    
            
                
    def _get_zdist(self, subset_lst, atom_sel):
        """
        Compute zscores for the distances in sentries on subset_lst
        """
        zscore_lst = []
        if atom_sel == 'GC':
            alig_array = self.alig_gc_dist_array
        else:
            alig_array = self.alig_dist_array
        for i in range(self.alig.length):
            for j in range(i):
                dist_lst = []
                for sentry_id in subset_lst:
                    coord_dist =  self._get_dist(sentry_id, i, j, alig_array)
                    if coord_dist >=0:
                        dist_lst.append(coord_dist)
                nr_dist = len(dist_lst)
                if nr_dist > 1:
                    dist_v = Numeric.array(dist_lst, Numeric.Float)
                    dist_mean = MLab.mean(dist_v)
                    dist_sdev = MLab.std(dist_v)
                    if dist_sdev > 0.0:
                        zscore_v = (dist_v - dist_mean) /dist_sdev
                        for c in range(len(zscore_v)):
                            zscore_lst.append(zscore_v[c])                        
        zscore_subset =  Numeric.array(zscore_lst, Numeric.Float)         
        return zscore_subset
                    
   
                                         
    def _get_dsmat_file(self, sentry_id, mat_dir):
        file_name = sentry_id + '.dsmat'
        dist_sus_mat_file =  os.path.join(mat_dir,file_name)
        return dist_sus_mat_file
        
    
    
    def _get_dpi_dic(self, dpi_file):
        suspdb_obj = sus_dpi.SUSpdb()
        suspdb_obj.read_sus_file(dpi_file)
        sentryid_lst = self.sentryset.sentryid_dic.keys()
        for sentry_id in sentryid_lst:
            pdb_id = sentry_id[1:5]
            if suspdb_obj.sus_dic.has_key(pdb_id):
                self.dpi_dic[sentry_id] = suspdb_obj.sus_dic[pdb_id]
                self.sentry_sus_lst.append(sentry_id)
            else:
                self.sentry_nosus_lst.append(sentry_id)
        log_util.log_info("Sentries with SUS: %s" % str(self.sentry_sus_lst))
        if len(self.sentry_nosus_lst) > 0:
            log_util.log_info("Sentries without SUS: %s" % str(self.sentry_nosus_lst))

        
    ##########################################################################        
    # old plots            
    def plot_sdev_subset(self, subset_lst, dist_variance_plot_name, source_dir):
        """
        Plot the matrix of distance standard deviation and distance standard deviation with mean sus
        """
        self.source_r(source_dir)
        
        dist_sdev_mat = self.get_dist_sdev_mat(subset_lst)
        dist_sus_sdev_mat = self.get_dist_sus_sdev_mat(subset_lst, dist_sdev_mat)
        ratio_v = self.get_sdev_ratio_v(subset_lst, dist_sdev_mat, dist_sus_sdev_mat)
        
        # get breaks, colors and legend
        (breaks_lst, col_lst, legend_lst, hbreaks_lst) = self._get_plot_sdev_data(dist_sdev_mat, dist_sus_sdev_mat)
        
        
        plot_title = dist_variance_plot_name + ' dist_sdev'
        plot_file = dist_variance_plot_name + '_dist_sdev.png'
        rpy.r.plotsdev(dist_sdev_mat, plot_file, plot_title, breaks_lst, col_lst, legend_lst)
        plot_file = dist_variance_plot_name + '_hist_dist_sdev.png'
        rpy.r.histsdev(dist_sdev_mat, plot_file, plot_title, hbreaks_lst)
        
        plot_title = dist_variance_plot_name + ' dist_sus_sdev'
        plot_file = dist_variance_plot_name + '_dist_sus_sdev.png'
        rpy.r.plotsdev(dist_sus_sdev_mat, plot_file, plot_title, breaks_lst, col_lst, legend_lst)
        plot_file = dist_variance_plot_name + '_hist_dist_sus_sdev.png'
        rpy.r.histsdev(dist_sus_sdev_mat, plot_file, plot_title, hbreaks_lst)
        
        log_util.log_info('Finished analysis of sdev')
        return (dist_sdev_mat, dist_sus_sdev_mat, ratio_v)
        
    
    
    def plot_surface_combined(self, mat1, mat2, comp_name, source_dir):
        n = len(mat1)
        comb_mat = Numeric.zeros((n,n), Numeric.Float)
        for i in range(n):
            for j in range(i):
                comb_mat[i,j] = mat1[i,j]
                comb_mat[j,i] = mat2[i,j]
        mat_file_name = comp_name + '.mat'
        self._write_mat_R(comb_mat, mat_file_name)
        self.source_r(source_dir)
        rpy.r.plotsurface(mat_file_name)
                
        
        
    def _get_plot_sdev_data(self, dist_sdev_mat, dist_sus_sdev_mat):
        # sdev histogram bin width
        hbin_w = 0.02
        # set the nr of bins, and respective colors
        # gray left for undefined
        nr_bins = 4
        col_lst = ['gray', 'white', 'moccasin', 'lightsalmon', 'red3']

        # get min max value        
        min_val1 = MLab.min(MLab.min(dist_sdev_mat))
        min_val2 = MLab.min(MLab.min(dist_sus_sdev_mat))
        max_val1 = MLab.max(MLab.max(dist_sdev_mat))
        max_val2 = MLab.max(MLab.max(dist_sus_sdev_mat))
        if max_val1 >= max_val2:
            max_val = max_val1
        else:
            max_val = max_val2
        
        # histogram breaks
        hmax_break = math.ceil(max_val)
        hnr_bins = int(round(hmax_break / hbin_w, 0))
        hbreak_lst = []
        for i in range(hnr_bins + 1):
            hbreak_lst.append(hbin_w * i)
            
            
        # sdev matrix plot define breaks_lst
        max_break = math.ceil(max_val)
        bin_w = max_break / nr_bins
        breaks_lst = [-2.0,   -0.5]
        for i in range(1, nr_bins+1):
            breaks_lst.append(bin_w * i)
            
        
        # sdev matrix plot define legend_lst
        legend_lst = ['undef']
        legend_item = "[0.00,%.2f]" % bin_w
        legend_lst.append(legend_item)
        for i in range(1, nr_bins):
            legend_item = "]%.2f,%.2f]" % (bin_w * i, bin_w * (i+1))
            legend_lst.append(legend_item)
        
        return (breaks_lst, col_lst, legend_lst, hbreak_lst)

