#!/usr/bin/python

"""
Maximal clique search
According to:
J Comput Biol, 1996 , 3 , 289-306
algorithm 2, variant BKYS
"""


# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de



import sys
import Numeric


class CliqueS:
    """
    Maximal clique search class
    """
    
    
    def __init__(self):
        self.graph_file = None
        self.graph_array = None   # Numeric binary array, size n_graph x n_graph
        self.n_graph = None
        self.cliques = []    # list of cliques
        
        
    def get_graph(self, graph_file):
        # read graph file
        self.graph_file = graph_file
        f_obj = file(self.graph_file, 'r')
        lines = f_obj.readlines()
        f_obj.close()
        
        # get dimensions
        row_lst = lines[0].split(',')
        n_graph = len(row_lst)
        graph_array = Numeric.zeros([n_graph,n_graph], Numeric.Int)
        # get values
        r_counter = 0
        for line_str in lines:
            line_str = line_str.strip()
            if line_str == '':
                continue
            row_lst = line_str.split(',')    
            if len(row_lst) != n_graph:
                # error in input graph
                return
            c_counter = 0
            for row_elem in row_lst:
                if row_elem == '1':
                    graph_array[r_counter,c_counter] = 1
                c_counter +=1
            if c_counter != n_graph:
                # error in input graph
                return
            r_counter +=1
        if r_counter != n_graph:
            # error in input graph
            return
        self.graph_array = graph_array
        self.n_graph = graph_array.shape[0]
        
        
        
    def set_graph(self, graph_array):
        self.graph_array = graph_array
        self.n_graph = graph_array.shape[0]
        
        
        
    def set_diagonals(self):
        """
        Set graph array diagonals to 0
        """
        for i in range(self.n_graph):
            self.graph_array[i,i] = 0


    def init_search(self):
        c_set = {}
        p_set = {}
        s_set = {}
        if self.n_graph == 1:
            self.cliques = [[0]]
        else:        
            for ui in range(self.n_graph):
                p_set[ui] = ui
            self.c_search(c_set, p_set, s_set)


        
    def get_cliques(self):
        """
        Return the cliques sorted by size (largest first)
        """
        self.cliques.sort(sort_cliques)            
        return self.cliques
        
        
        
    def c_search(self, c_set, p_set, s_set):
        if len(p_set) == 0 and len(s_set) ==0:
            # found maximal clique
            clique_lst = c_set.keys()
            clique_lst.sort()
            self.cliques.append(clique_lst)
            # print "%d %s" % (len(clique_lst),str(clique_lst))
        else:
            if len(s_set) !=0:
                # take one element from s_set
                ut = s_set.keys()[0]
                for ui in p_set.keys():
                    if self.graph_array[ui,ut] == 0:
                        del p_set[ui]
                        pp_set = p_set.copy()
                        ss_set = s_set.copy()
                        n_set = self.get_adjacent(ui)
                        cc_set = c_set.copy()
                        cc_set[ui] = ui
                        ppp_set = self.ints_sets(pp_set, n_set)
                        sss_set = self.ints_sets(ss_set, n_set)
                        self.c_search(cc_set, ppp_set, sss_set)
                        s_set[ui] = ui
            else:            
                for ui in p_set.keys():
                    del p_set[ui]
                    pp_set = p_set.copy()
                    ss_set = s_set.copy()
                    n_set = self.get_adjacent(ui)
                    cc_set = c_set.copy()
                    cc_set[ui] = ui
                    ppp_set = self.ints_sets(pp_set, n_set)
                    sss_set = self.ints_sets(ss_set, n_set)
                    self.c_search(cc_set, ppp_set, sss_set)
                    s_set[ui] = ui
                
                                            
    def get_adjacent(self,ui):
        n_set = {}
        u_row = self.graph_array[ui,]
        for v in range(self.n_graph):
            if u_row[v]:
                n_set[v] = v
        return n_set
    
        
        
    def ints_sets(self,set1,set2):
        ints_set = {}
        for key1 in set1.keys():
            if set2.has_key(key1):
                ints_set[key1] = key1    
        return ints_set
                        

def sort_cliques(clique1, clique2):
    comp_val = cmp(len(clique1), len(clique2))
    return -comp_val
    
    
# if len(sys.argv) != 2:
#    print "Usage: clique_search.py csv_file"
#    sys.exit()
# file_name = sys.argv[1]

    
# cs = CliqueS()
# cs.get_graph(file_name)
# cs.set_diagonals()
# print "Nr nodes: %d" % cs.n_graph
# cs.init_search()
        