"""
Cluster sets in a structural database.

Cluster each different basesets. Basesets are composed of structural entries (sentries)
corresponding to alternative models.
The resulting clusters define new subsets that can be compared
"""


# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import os.path
import copy
import log_util
import scop_sets
import set_alig
import base_set
import Numeric
import Bio.SVDSuperimposer
import rpy
import util_r


class ClusterDb:
    """
    Base class for clustering a database of alternative structural models.
    
    Includes parameters and the definition of the sets
    Loops over all sets, for each set a new cluster object is defined which
    performs clustering and writes output
    """

    def __init__(self, config_dic):
        """
        Initialization
        
        Take parameters from config_dic
        """

        # data files and directories
        self.setsid_file = config_dic['setsid_file']   # definition of sets: file with basesets id
        self.baseset_dir = config_dic['baseset_dir']   # directory with baseset files
        self.sets_file = config_dic['sets_file']       # definition of sets: from SCOP
        self.pdb_dir   = config_dic['pdb_dir']         # pdb directory
        self.use_pdb_subdir = config_dic['use_pdb_subdir']   # True/False, use subdirectory to access PDB
        self.stmap_dir   = config_dic['stmap_dir']     # residue mapping directory
        self.uniprot_dir = config_dic['uniprot_dir']   # directory with uniprot sequences
        self.source_dir = config_dic['source_dir']     # source code directory
        # parameters
        self.min_set_len =  config_dic['min_set_len']  # minimum number of sentries in set
        self.max_set_len = config_dic['max_set_len']  # max number of entries in set, if no max then None
        self.dissim_type = config_dic['dissim_type']   # type of dissimilarity: dmat dali or rmsd
        self.dist_filter = config_dic['dist_filter']   # None or a float
        self.dif_dist_filter = config_dic['dif_dist_filter']  # None or a float
        self.one_gap_value = config_dic['one_gap_value']
        self.two_gap_value = config_dic['two_gap_value']
        self.atom_sel = config_dic['atom_sel']   # atoms to use, CA or GC
        
                
    def cluster_scop_sets(self):
        """
        Cluster SCOP species sets.
        
        Loads a SCOP sets defintion file to define the basesets.
        Then compile the basesets, get alignment of distances
        Cluster
        """
        # source R files
        self.R_source()
        # read sets definition into a list of lines
        read_scop = scop_sets.ReadScopBaseset(self.min_set_len, self.max_set_len, self.atom_sel)
        read_scop.load_file(self.sets_file)

        while(read_scop.line_counter < read_scop.nr_lines):
            last_lnr = read_scop.line_counter
            baseset = None
            
            try:
                # get each set from scop species level
                # define as clusterset
                baseset = read_scop.get_baseset(self.pdb_dir, self.use_pdb_subdir)
                baseset.set_working_to_current_dir()
                clusterset =  ClusterSet(baseset, self.dist_filter, self.dif_dist_filter, self.one_gap_value, self.two_gap_value, False) 
                # get alignment, distances, and alig the distances
                clusterset.alig_distances(self.stmap_dir, self.uniprot_dir)
                if self.dissim_type == 'dmat':                    
                    clusterset.cluster_dmat()
                else:
                    log_util.log_info('Skipping set %s, dissimilarity type not available' % self.baseset.set_id)
                    continue
            except:
                log_util.log_error('Could not cluster input line nr:%d' % read_scop.line_counter)
                if read_scop.line_counter == last_lnr:
                    # if same line on error, then increase line count
                    read_scop.line_counter += 1
                continue



    def cluster_basesets(self):
        """
        Cluster precompiled basesets, defined in setsid file
        Basesets include already aligned distances
        """
        setsid_lst = base_set.get_setsid(self.setsid_file)
        for set_id in setsid_lst:
            try:
                baseset = base_set.load_baseset(set_id,self.baseset_dir)
                clusterset =  ClusterSet(baseset, self.dist_filter, self.dif_dist_filter, self.one_gap_value, self.two_gap_value)
                if self.dissim_type == 'dmat':                    
                    clusterset.cluster_dmat()
                else:
                    log_util.log_info('Skipping set %s, dissimilarity type not available' % self.baseset.set_id)
                    continue
            except:
                log_util.log_error('Could not cluster set id:%s' % set_id)
                if read_scop.line_counter == last_lnr:
                    # if same line on error, then increase line count
                    read_scop.line_counter += 1
                continue


    def R_source(self):
        """
        Source the R files
        """
        r_file = os.path.join(self.source_dir, 'clustering.R')
        rpy.r.source(r_file)





class ClusterSet:
    """
    Base class for clustering a set.
    """

    def __init__(self, baseset, dist_filter, dif_dist_filter, one_gap_value, two_gap_value, store_fdif, cl_atoms = 'CA'):
        # baseset
        self.baseset = baseset  # SentrySet
        # parameters
        self.dist_filter = dist_filter
        self.dif_dist_filter = dif_dist_filter
        self.one_gap_value = one_gap_value
        self.two_gap_value = two_gap_value
        self.store_fdif = store_fdif
        self.cl_atoms = cl_atoms
        # data
        self.dist_array = None        
        self.fdif_dic = {}
        self.dissim_mat = None
        self.max_dissim = None
        self.rmsd_maxd = None
        self.dissim_matrix_file = None
        self.cluster_dend_file = None

        

    def alig_distances(self, stmap_dir, uniprot_dir):
        """
        Get the aligned distances in baseset
        """
        # get alignment
        self.baseset.get_alignment(stmap_dir, uniprot_dir)
        # get aligned distance matrices into alig_dist_array
        self.baseset.get_alig_dist_array()



    def cluster_dmat(self):
        """
        Cluster according to difference of distance matrices.
        """
        
        # get dissimilarity matrix
        if self.cl_atoms == 'CA_GC':
            self.get_dissim_matrix_cagc()
        else:
            self.get_dissim_matrix()
            
        # get the RMSD
        # self.get_rms_max_dissim()
        # write the dissimilarity matrix
        self.write_dissim_matrix()
        # hierarchical clustering
        if self.baseset.set_len > 2:
            self.R_h_cluster()
        else:
            log_util.log_info("No hierarchical clustering dendogram, too few sentries:%d" % self.baseset.set_len)
        # PAM clustering
        # self.R_pam_cluster()
        # write mappings of aligned residues
        # self.write_alig_dist_array()


        
    def cluster_subdmat(self, sentry_id, reskey_lst):
        """
        Cluster according to difference of distance matrices.
        Use a list of residues of a sentry to define a submatrix
        """
        # get dissimilarity matrix
        if self.cl_atoms == 'CA_GC':
            self.get_dissim_matrix_cagc(sentry_id, reskey_lst)
        else:
            self.get_dissim_matrix(sentry_id, reskey_lst)
        # write the dissimilarity matrix
        self.write_dissim_matrix()
        # hierarchical clustering
        if self.baseset.set_len > 2:
            self.R_h_cluster()
        else:
            log_util.log_info("No hierarchical clustering dendogram, too few sentries:%d" % self.baseset.set_len)

                                   
    
    
    def write_alig_dist_array(self):
        """
        Debug, write alig_dist_array.
        """
        for k in range(self.baseset.set_len):
            sentry =  self.baseset.sentryset.sentry_lst[k]
            # print sentry_id
            print "%7s " % sentry.sentry_id
            # column labels
            out_str = "%14s" % ''
            for i in range(self.baseset.alig.length):
                if self.baseset.alig.refpos_dic[i].sentry_dic.has_key(sentry.sentry_id):
                    sentrypos = self.baseset.alig.refpos_dic[i].sentry_dic[sentry.sentry_id]
                    residue_str = "%4d>%5s:%s  " % (i, sentry.res_lst[sentrypos].label, sentry.res_lst[sentrypos].chain_id)
                else:
                    residue_str = "%4d>%5s:%s  " % (i, '_','_')
                out_str = out_str + residue_str
            print out_str
            out_str = ""
            for i in range(self.baseset.alig.length):
                # row label
                if self.baseset.alig.refpos_dic[i].sentry_dic.has_key(sentry.sentry_id):
                    sentrypos = self.baseset.alig.refpos_dic[i].sentry_dic[sentry.sentry_id]
                    residue_str = "%4d>%5s:%s  " % (i, sentry.res_lst[sentrypos].label, sentry.res_lst[sentrypos].chain_id)
                else:
                    residue_str = "%4d>%5s:%s  " % (i, '_','_')
                out_str = residue_str
                for j in range(self.baseset.alig.length):
                    # values
                    coord_dist = self.baseset.alig_dist_array[i,j,k]
                    dist_str = "%8.3f      " % coord_dist
                    out_str = out_str + dist_str
                print out_str 


                      
    def get_dissim_matrix(self, sentry_id = None, reskey_lst = None):
        """
        Get the dissimilarity matrix.

        Also set max dimiliarity value and sentry ids in max_dissim.
        """
        self.dist_array = copy.deepcopy(self.baseset.alig_dist_array)
        dissim_mat = Numeric.zeros([self.baseset.set_len, self.baseset.set_len], Numeric.Float)    
        max_dissim = (0.0, None, None)
        # distance filter
        self.dist_array = self._apply_dist_filter(self.dist_array)
        
        for k_a in range(self.baseset.set_len):
            for k_b in range(k_a):
                # get difference of aligned distance matrices
                dif_dist_mat = self._get_dif_dist_mat(self.dist_array, k_a, k_b)
                # difference of distances filter
                fdif_dist_mat = self._apply_dif_dist_filter(dif_dist_mat)
                # set gap values
                fdif_dist_mat = self._set_gap_values(fdif_dist_mat, k_a, k_b)
                # store the filtered difference matrix
                id_a = self.baseset.sentryset.get_id_from_pos(k_a)
                id_b = self.baseset.sentryset.get_id_from_pos(k_b)
                if self.store_fdif == True:
                    self.fdif_dic[(id_a,id_b)] = fdif_dist_mat
                    self.fdif_dic[(id_b,id_a)] = fdif_dist_mat
                # dissimilarity value
                if sentry_id == None:
                    dissim_value = self._get_dissim_value(fdif_dist_mat)
                else:
                    submat_fdif_dist_mat = self.baseset.get_submat(sentry_id, reskey_lst, fdif_dist_mat)
                    dissim_value = self._get_dissim_value(submat_fdif_dist_mat)
                dissim_mat[k_a, k_b] = dissim_value
                dissim_mat[k_b, k_a] = dissim_value
                if dissim_value >= max_dissim[0]:
                    max_dissim = (dissim_value, self.baseset.sentryset.get_id_from_pos(k_a), self.baseset.sentryset.get_id_from_pos(k_b))                
        self.dissim_mat = dissim_mat
        self.max_dissim = max_dissim
        log_util.log_info("Calculated dissimilarity matrix")


        
    def get_dissim_matrix_cagc(self, sentry_id = None, reskey_lst = None):
        """
        Get the dissimilarity matrix based on CA and side chain geometric center GC

        Also set max dimiliarity value and sentry ids in max_dissim.
        """
        
        self.dist_array = copy.deepcopy(self.baseset.alig_dist_array)
        self.gc_dist_array = copy.deepcopy(self.baseset.alig_gc_dist_array)
        
        dissim_mat = Numeric.zeros([self.baseset.set_len, self.baseset.set_len], Numeric.Float)    
        max_dissim = (0.0, None, None)
        
        # distance filter
        self.dist_array = self._apply_dist_filter(self.dist_array)
        self.gc_dist_array = self._apply_dist_filter(self.gc_dist_array)
        
        for k_a in range(self.baseset.set_len):
            for k_b in range(k_a):
                # get difference of aligned distance matrices
                dif_dist_ca_mat = self._get_dif_dist_mat(self.dist_array ,k_a, k_b)
                dif_dist_gc_mat = self._get_dif_dist_mat(self.gc_dist_array ,k_a, k_b)
                # difference of distances filter
                fdif_dist_ca_mat = self._apply_dif_dist_filter(dif_dist_ca_mat)
                fdif_dist_gc_mat = self._apply_dif_dist_filter(dif_dist_gc_mat)
                fdif_dist_mat = (fdif_dist_ca_mat + fdif_dist_gc_mat) /2.0 
                # set gap values
                fdif_dist_mat = self._set_gap_values(fdif_dist_mat, k_a, k_b)
                # store the filtered difference matrix
                id_a = self.baseset.sentryset.get_id_from_pos(k_a)
                id_b = self.baseset.sentryset.get_id_from_pos(k_b)
                if self.store_fdif == True:
                    self.fdif_dic[(id_a,id_b)] = fdif_dist_mat
                    self.fdif_dic[(id_b,id_a)] = fdif_dist_mat
                # dissimilarity value
                if sentry_id == None:
                    dissim_value = self._get_dissim_value(fdif_dist_mat)                
                else:
                    submat_fdif_dist_mat = self.baseset.get_submat(sentry_id, reskey_lst, fdif_dist_mat)
                    dissim_value = self._get_dissim_value(submat_fdif_dist_mat)                
                dissim_mat[k_a, k_b] = dissim_value
                dissim_mat[k_b, k_a] = dissim_value
                if dissim_value >= max_dissim[0]:
                    max_dissim = (dissim_value, self.baseset.sentryset.get_id_from_pos(k_a), self.baseset.sentryset.get_id_from_pos(k_b))                
        self.dissim_mat = dissim_mat
        self.max_dissim = max_dissim
        log_util.log_info("Calculated dissimilarity matrix")
    

    def get_rms_max_dissim(self):
        """
        Get RMSD for entries with largest dissimilarity as defined in  self.max_dissim.
        """
        # get list of positions of aligned residues
        max_dissim_a = self.max_dissim[1]
        max_dissim_b = self.max_dissim[2]
        respos_lst_a = []
        respos_lst_b = []
        for refpos in range(self.baseset.alig.length):
            if self.baseset.alig.refpos_dic[refpos].sentry_dic.has_key(max_dissim_a):
                if self.baseset.alig.refpos_dic[refpos].sentry_dic.has_key(max_dissim_b):
                    respos_lst_a.append(self.baseset.alig.refpos_dic[refpos].sentry_dic[max_dissim_a])
                    respos_lst_b.append(self.baseset.alig.refpos_dic[refpos].sentry_dic[max_dissim_b])
        coord_mat_a = self._get_coord(max_dissim_a, respos_lst_a)
        coord_mat_b = self._get_coord(max_dissim_b, respos_lst_b)
        # calculate RMSD
        sup = Bio.SVDSuperimposer.SVDSuperimposer()
        sup.set(coord_mat_a, coord_mat_b)
        sup.run()
        self.rmsd_maxd = sup.get_rms()
        # debug
        # print "RMSD: %.3f" % self.rmsd_maxd



    def write_dissim_matrix(self):
        """
        Write the dissimilarity matrix.
        """
        dissim_matrix_filename = self.baseset.set_id + '_' + 'dissim_matrix.txt'
        self.dissim_matrix_file = os.path.join(self.baseset.working_dir, dissim_matrix_filename)
        file_obj = file(self.dissim_matrix_file, 'w')
        out_str = "%10s " % " "
        for i in range(self.baseset.set_len):
            sentry_id = self.baseset.sentryset.get_id_from_pos(i)
            id_str = "%10s " % sentry_id
            out_str = out_str + id_str
        file_obj.write(out_str + "\n")
        for i in range(self.baseset.set_len):
            sentry_id = self.baseset.sentryset.get_id_from_pos(i)
            file_obj.write("%-10s " % sentry_id)
            for j in range(self.baseset.set_len):
                file_obj.write("%10.1f " % self.dissim_mat[i,j])
            file_obj.write("\n")
        file_obj.close()
        log_util.log_info("Wrote %s" % self.dissim_matrix_file)
        

    def _apply_dist_filter(self, dist_array):
        if self.dist_filter != None:
            mask_array = Numeric.greater(dist_array, self.dist_filter)
            Numeric.putmask(dist_array, mask_array, self.dist_filter)
            return dist_array


    def _get_dif_dist_mat(self, dist_array, k_a, k_b):
        """
        Get the difference of distance matrices.

        Use the sentry indeces k_a and k_b to get the
        difference of distance from alig_dist_array.
        """
        # get dif_dist_mat
        dm_a = dist_array[:,:,k_a]
        dm_b = dist_array[:,:,k_b]
        dif_dist_mat = Numeric.absolute(Numeric.subtract(dm_a,dm_b))
        return dif_dist_mat



    def _apply_dif_dist_filter(self, dif_dist_mat):
        """
        Apply the diference of distance filter.

        return the filtered diffrence of distances.
        """
        # set the significant differences (dist difference > dif_dist_filter) to dif_dist_filter
        if  self.dif_dist_filter == None:
            return dif_dist_mat
        filtered_logical = Numeric.greater(dif_dist_mat, self.dif_dist_filter)
        # get a float matrix, sum the logical to a zero float matrix
        d = dif_dist_mat.shape[0]
        fm = Numeric.zeros([d,d], Numeric.Float)
        filtered_dif_dist_mat = fm + filtered_logical
        return filtered_dif_dist_mat



    def _get_dissim_value(self, mat):
        """
        Get the dissimilarity value.
        """
        # get the dissimilarity value, sum of mat
        dissim_value = Numeric.sum(Numeric.sum(mat))
        return dissim_value
                                

    def _get_coord(self, sentry_id, respos_lst):
        """
        Get coordinate matrix for residue psoition list respos_lst of the sentry with id sentry_id
        """
        sentry = self.baseset.sentryset.get_sentry(sentry_id)
        coord_mat = Numeric.zeros([len(respos_lst), 3], Numeric.Float)
        coord_pos = 0
        for respos in respos_lst:
            coord_mat[coord_pos, 0] = sentry.res_lst[respos].atom_dic['CA'].get_coord()[0]
            coord_mat[coord_pos, 1] = sentry.res_lst[respos].atom_dic['CA'].get_coord()[1]
            coord_mat[coord_pos, 2] = sentry.res_lst[respos].atom_dic['CA'].get_coord()[2]
            coord_pos += 1
        return coord_mat


    def _set_gap_values(self, dif_dist_mat, k_a, k_b):
        """
        Set value for gaps in dif_dist_mat

        If gap in only one sentry: one_gap_value
        If gap in both sentryes: two_gap_value
        """
        two_gap_value = 0.0   # do not penalize if region missing in both
        dm_a = self.dist_array[:,:,k_a]
        dm_b = self.dist_array[:,:,k_b]
        # get gap positions
        gap_a = Numeric.less(dm_a, 0.0)
        smalld_a = Numeric.less(dm_a, self.dist_filter)
        gap_b = Numeric.less(dm_b, 0.0)
        smalld_b = Numeric.less(dm_b, self.dist_filter)
        gap_ab = Numeric.logical_and(gap_a, gap_b)
        gap_a_smalld_b = Numeric.logical_and(gap_a, smalld_b)
        gap_b_smalld_a = Numeric.logical_and(gap_b, smalld_a)
        # set gaps in one sentry to one_gap_value for small distances in the other sentry
        Numeric.putmask(dif_dist_mat, gap_a, self.two_gap_value)
        Numeric.putmask(dif_dist_mat, gap_b, self.two_gap_value)
        Numeric.putmask(dif_dist_mat, gap_a_smalld_b, self.one_gap_value)
        Numeric.putmask(dif_dist_mat, gap_b_smalld_a, self.one_gap_value)
        # set gaps in both sentries to two_gap_value
        Numeric.putmask(dif_dist_mat, gap_ab, self.two_gap_value)
        return dif_dist_mat


        



    def R_h_cluster(self):
        """
        Call R function fot hierarchical clustering
        """
        cluster_dend_filename = self.baseset.set_id + '_' + 'cluster_dend.png'
        self.cluster_dend_file = os.path.join(self.baseset.working_dir, cluster_dend_filename)
        file_obj = file(self.cluster_dend_file, 'w')
        title_plot = self.baseset.set_id
        file_obj.close()
        set_len = self.baseset.set_len
        if set_len <= 20:
            imgw = 800
            imgh = 800
            imgpsz = 12
        elif set_len <= 40:
            imgw = 800
            imgh = 800
            imgpsz = 12
        elif set_len <= 60:
            imgw = 800
            imgh = 800
            imgpsz = 12        
        elif set_len <= 80:
            imgw = 1120
            imgh = 800
            imgpsz = 12
        elif set_len <= 100:
            imgw = 1120
            imgh = 800
            imgpsz = 12
        elif set_len <= 150:
            imgw = 1840
            imgh = 800
            imgpsz = 8
        else:
            imgw = 3000
            imgh = 800
            imgpsz = 8
        rpy.r.h_cluster_dend(self.dissim_matrix_file, self.cluster_dend_file, title_plot, imgw, imgh , imgpsz)
        log_util.log_info("Hierarchical clustering dendogram:%s" % self.cluster_dend_file)

        
        
    def R_h_sil(self):
        """
        Get best number of clusters according to silhouette width
        Get silhouette width values for best number of clusters
        """
        asw_filename = self.baseset.set_id + '_' + 'cluster_h_asw.png'
        asw_file = os.path.join(self.baseset.working_dir, asw_filename)
        sil_filename = self.baseset.set_id + '_' + 'cluster_h_sil.png'
        sil_file = os.path.join(self.baseset.working_dir, sil_filename)
        sil_obj = rpy.r.h_cluster(self.dissim_matrix_file, asw_file, sil_file, self.baseset.set_id)
        log_util.log_info("Hierarchical clustering silhouette width values: %s" %self.baseset.set_id)
        
        

    def R_pam_cluster(self):
        asw_filename = self.baseset.set_id + '_cluster_pam_asw.png'
        asw_file = os.path.join(self.baseset.working_dir, asw_filename)
        sil_filename = self.baseset.set_id + '_cluster_pam_sil.png'
        sil_file = os.path.join(self.baseset.working_dir, sil_filename)
        pam_array = rpy.r.pam_cluster(self.dissim_matrix_file, asw_file, sil_file, self.baseset.set_id)
        log_util.log_info("PAM clustering: %s" %self.baseset.set_id)
