"""
Compare two subsets
"""


# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de




import os.path
import math
import types
import rpy
import Numeric
import MLab

import set_alig
import log_util


def generate_compsets(baseset, config_dic):
    """
    Given baseset and parameters in config_dic, return a compsets object
    """ 

    # align
    baseset.get_alignment(config_dic['stmap_dir'],config_dic['uniprot_dir'])
    # get aligned distance matrices, 
    baseset.get_alig_dist_array()
    
    # check method
    if config_dic['method'] == 'sus':
        # get aligned distance sus
        baseset.get_alig_sus(dpi_file = config_dic['dpi_file'])

    compsets = CompSets(baseset, config_dic['source_dir'])  
    return compsets





class CompSets:
    """
    Set Comparison
    """
    
    def __init__(self, baseset, source_dir):

        # general parameters
        self.source_dir = source_dir
        self.baseset = baseset
        
        # parameters specific for one comparison
        self.comp_name    = None
        self.subset1_lst = None
        self.subset2_lst = None
        self.method = None   # sus, dist, dist is default
        self.atom_sel = None
        self.score_type = None # dscore, pval
        self.pval_cutoff = None
        self.dscore_cutoff = None

        # temporary data
        self.comp_mode = None # Set2Set, Entry2Set, Entry2Entry
        self.score_file = None
        
        # results        
        self.score_mat = None   # pvalues, dscores
        self.undef_mat = None   # matrix with undefined entries in score_mat set to 1        
        
        
    def comparison(self, comp_name, subset1_lst, subset2_lst, method, score_type, score_cutoff, atom_sel = 'CA'):
        """
        Call the comparsion method defined in self.method
        """
        self.new_parameters(comp_name, subset1_lst, subset2_lst, method, score_type, score_cutoff, atom_sel)
        
        dist1_len = len(self.subset1_lst)
        dist2_len = len(self.subset2_lst)
        # find mode: Set2Set Entry2Set Entry2Entry
        comp_mode = None
        if dist1_len > 1:
            if dist2_len > 1:
                comp_mode = 'Set2Set'
            else:
                log_util.log_error('Set2Set: Set 2 must have more than 1 entry')
                return
        elif dist1_len ==1:
            if dist2_len > 1:
                comp_mode = 'Entry2Set'
            elif dist2_len == 1:
                comp_mode = 'Entry2Entry'
            else:
                log_util.log_error('Entry2Entry or Entry2Set: Set 2 must have one or more entries')
                return
        else:
            log_util.log_error('Possibilities: Entry2Entry Entry2Set Set2Set')
            return
        if comp_mode == 'Entry2Entry':
            if self.score_type != 'dscore':
                log_util.log_error('Entry2Entry only dscore')
                return
        self.comp_mode = comp_mode              
        # apply method sus or dist
        if self.comp_mode == 'Set2Set':
            log_util.log_info('Set2Set Comparison')
            if self.method == 'sus':
                self.comparison_sus_set2set()
            elif self.method == 'dist':
                self.comparison_dist_set2set()
            else:
                self.comparison_dist_set2set()
        elif self.comp_mode == 'Entry2Set':
            log_util.log_info('Entry2Set Comparison')
            if self.method == 'sus':
                self.comparison_sus_entry2set()
            elif self.method == 'dist':
                self.comparison_dist_entry2set()
            else:
                self.comparison_dist_entry2set()
        elif self.comp_mode == 'Entry2Entry':
            log_util.log_info('Entry2Entry Comparison')
            if self.method == 'sus':
                self.comparison_sus_entry2entry()
            elif self.method == 'dist':
                self.comparison_dist_entry2entry()
            else:
                self.comparison_dist_entry2entry()
        else:
            log_util.log_error('Possibilities: Entry2Entry Entry2Set Set2Set')
            return
        
        #  make plots
        self.plot_score()
                       
        
                            
    def new_parameters(self, comp_name, subset1_lst, subset2_lst, method, score_type, score_cutoff, atom_sel):
        # parameters specific for one comparison
        self.comp_name = comp_name
        self.subset1_lst = subset1_lst
        self.subset2_lst = subset2_lst
        self.method = method
        self.atom_sel = atom_sel
        self.score_type = score_type
        if self.score_type == 'pval':
            self.pval_cutoff = score_cutoff
            self.dscore_cutoff = None
        else:
            self.dscore_cutoff = score_cutoff
            self.pval_cutoff = None
        # erase old temporary data
        self.comp_mode = None # Set2Set, Entry2Set, Entry2Entry
        self.score_file = None
        # erase old results
        self.score_mat = None
        self.undef_mat = None
        
              
    def plot_score(self):
        r_file = os.path.join(self.source_dir, 'make_plots.R')
        rpy.r.source(r_file)
        if self.score_type == 'pval':
            score_cutoff = 1.0 - self.pval_cutoff
        else:
            score_cutoff = self.dscore_cutoff
        plot_title = "%s %s" % (self.comp_name, self.comp_mode)
        plot_file = self.comp_name + '.png'
        rpy.r.plotcompscore(self.score_file, plot_file, plot_title, score_cutoff)

    
        
    def plot_score_submat(self, sentry_id, reskey_lst):
        r_file = os.path.join(self.source_dir, 'make_plots.R')        
        rpy.r.source(r_file)        
        # get ref pos
        refpos_lst = self.baseset.reskey_to_refpos(sentry_id, reskey_lst)        
        # get submat
        submat_len = len(refpos_lst)
        submat = Numeric.zeros((submat_len,submat_len), Numeric.Float)
        submat_undef = Numeric.zeros((submat_len,submat_len), Numeric.Float)
        i = 0
        for i_refpos in refpos_lst:
            j = 0
            for j_refpos in refpos_lst:
                # print "i_refpos %d j_refpos %d, i %d, j %d" % (i_refpos,j_refpos, i, j) # debug
                submat[i,j] = self.score_mat[i_refpos,j_refpos]
                submat_undef[i,j] = self.undef_mat[i_refpos,j_refpos]
                j += 1            
            i += 1
            
        # write to file
        score_file = self.write_score_submat_Rdf(submat,submat_undef)
        
        # cutoff
        if self.score_type == 'pval':
            score_cutoff = 1.0 - self.pval_cutoff
        else:
            score_cutoff = self.dscore_cutoff
        
        # r plot
        plot_title = "%s site %s" % (self.comp_name, self.comp_mode)
        plot_file = self.comp_name + '_site.png'
        rpy.r.plotcompscore(score_file, plot_file, plot_title, score_cutoff)
        
        
        
    def comparison_sus_set2set(self):
        """
        Distance sus comparison, compare 2 sets
        get pvalues and dscores
        if no pvalue can be computed, set undef_mat
        Do not compute pvalue or dscore if:
            1 or less distances in one set
            missing all distance sus in both sets
        missing distances are not considered
        missing distance sus are not considered for calculating the average
        if all distance sus are missing in one set, take average from other set
        """
        log_util.log_info('Distance and SUS Comparison')
        # for each position in the aligned distance matrices
        self.score_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Float)
        self.undef_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Int)
        
        if self.atom_sel == 'GC':
            alig_array = self.baseset.alig_gc_dist_array
            alig_sus_array = self.baseset.alig_gc_sus_array
        else:
            alig_array = self.baseset.alig_dist_array
            alig_sus_array = self.baseset.alig_sus_array
            
        # debug
        # self.write_subsets()
        for i in range(self.baseset.alig.length):
            for j in range(i):
                # values for subset 1                
                dist1_len = len(self.subset1_lst)
                dist1_v = Numeric.zeros(dist1_len, Numeric.Float)
                distsusv1_v = Numeric.zeros(dist1_len, Numeric.Float)                
                c = 0
                for sentry_id in self.subset1_lst:
                    dist1_v[c] = self.baseset._get_dist(sentry_id, i, j, alig_array)
                    distsusv1_v[c] = self.baseset._get_distsusv(sentry_id, i, j, alig_sus_array)
                    c += 1
                # values for subset2
                dist2_len = len(self.subset2_lst)
                dist2_v = Numeric.zeros(dist2_len, Numeric.Float)
                distsusv2_v = Numeric.zeros(dist2_len, Numeric.Float)                
                c = 0
                for sentry_id in self.subset2_lst:
                    dist2_v[c] = self.baseset._get_dist(sentry_id, i, j, alig_array)
                    distsusv2_v[c] = self.baseset._get_distsusv(sentry_id, i, j, alig_sus_array)
                    c += 1
                # extract defined distances, 
                # ignore distances not defined (-1) from dist vectors,
                # and ignore the corresponding (same index) distsusv values
                dist1_v, distsusv1_v = self._get_defined_dist_distsus(dist1_v, distsusv1_v)
                dist2_v, distsusv2_v = self._get_defined_dist_distsus(dist2_v, distsusv2_v)
                # get defined distsusv
                defined_distsusv1_v = self._get_defined_v(distsusv1_v)
                defined_distsusv2_v = self._get_defined_v(distsusv2_v)
                # compute mean distsusv, use only defined values, ingore the missing values (-1)
                mean_distsusv1 = self._mean(defined_distsusv1_v)
                mean_distsusv2 = self._mean(defined_distsusv2_v)
                # check distance values
                if  len(dist1_v) <= 1 or len(dist2_v) <= 1:
                    # one or no distance in one or both sets, no p value, dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                elif len(defined_distsusv1_v) == 0 and len(defined_distsusv2_v) == 0:
                    # no distance sus, no pvalue, no dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                if len(defined_distsusv1_v) == 0:
                    # no distance sus in set 1, set mean to mean of set 2
                    mean_distsusv1 = mean_distsusv2
                if len(defined_distsusv2_v) == 0:
                    # no distance sus in set 2, set mean to mean of set 1
                    mean_distsusv2 = mean_distsusv1
                try:
                    if self.score_type == 'pval':
                        comp_score = self._compute_sus_set2set_pval(dist1_v, dist2_v, mean_distsusv1, mean_distsusv2)
                    else:
                        comp_score = self._compute_sus_set2set_dscore(dist1_v, dist2_v, mean_distsusv1, mean_distsusv2)
                    if comp_score != None:
                        self.score_mat[i,j] = comp_score
                        self.score_mat[j,i] = comp_score
                    else:
                        self.undef_mat[i,j] = 1
                        self.undef_mat[j,i] = 1
                except:
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    log_util.log_error('Failed comparison i:%d, j:%d dist1:%s dist2:%s distsusv1:%s distsusv2:%s' % (i, j, str(dist1), str(dist2), str(distsusv1), str(distsusv2)))
                # debug
                # self.write_datav(i, j, dist1, dist2, distsusv1, distsusv2)
        self.write_score_mat_Rdf()
        log_util.log_info('Computed comparison %s' % self.comp_name)       

                    
        
    def  comparison_dist_set2set(self):       
        """
        Comparsion based on distances, compare 2 sets
        get pvalues and dscores
        if no pvalue can be computed, set undef_mat
        Do not compute pvalue or dscore if:
            1 or less distances in one set
        """
        log_util.log_info('Distance Comparison')
        # for each position in the aligned distance matrices
        self.score_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Float)
        self.undef_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Int)
        
        if self.atom_sel == 'GC':
            alig_array = self.baseset.alig_gc_dist_array
        else:
            alig_array = self.baseset.alig_dist_array
        
        for i in range(self.baseset.alig.length):
            for j in range(i):
                # values for subset 1                
                dist1_len = len(self.subset1_lst)
                dist1_v = Numeric.zeros(dist1_len, Numeric.Float)
                c = 0
                for sentry_id in self.subset1_lst:
                    dist1_v[c] = self.baseset._get_dist(sentry_id, i, j, alig_array)
                    c += 1                                              
                # values for subset2
                dist2_len = len(self.subset2_lst)
                dist2_v = Numeric.zeros(dist2_len, Numeric.Float)
                c = 0
                for sentry_id in self.subset2_lst:
                    dist2_v[c] = self.baseset._get_dist(sentry_id, i, j, alig_array)
                    c += 1
                # extract defined distances, 
                # ignore distances not defined (-1) from dist vectors,
                dist1_v = self._get_defined_v(dist1_v)
                dist2_v = self._get_defined_v(dist2_v)    
                # check distance values
                if  len(dist1_v) <= 1 or len(dist2_v) <= 1:
                    # if less than 2 distances in the sets, no p value, dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                try:
                    # debug
                    # if i == 89 and j == 1:
                    #    log_util.log_info('i:%d j:%d\ndist1:\n %s\n dist2:\n %s' % (i,j,str(dist1_v), str(dist2_v)))
                    if self.score_type == 'pval':
                        comp_score = self._compute_dist_set2set_pval(dist1_v, dist2_v, i, j)
                    else:
                        comp_score = self._compute_dist_set2set_dscore(dist1_v, dist2_v)
                    if comp_score != None:
                        self.score_mat[i,j] = comp_score
                        self.score_mat[j,i] = comp_score
                    else:
                        self.undef_mat[i,j] = 1
                        self.undef_mat[j,i] = 1
                except:
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    log_util.log_error('Failed comparison i:%d, j:%d dist1:%s dist2:%s' % (i, j, str(self.subset1_lst), str(self.subset2_lst)))
        self.write_score_mat_Rdf()
        log_util.log_info('Computed comparison %s' % self.comp_name)       

        
        
    def comparison_sus_entry2set(self):
        """
        Distance sus comparison, compare entry to a set
        subset1_lst contain a single entry
        get pvalues and dscores
        if no pvalue can be computed, set undef_mat
        distance sus from subset1 is added to subset12
        Do not compute pvalue or dscore if:
            no distance in subset1 or 1 or less distances in subset2\
            missing all distance sus in subset2
        missing distances are not considered
        missing distance sus are not considered for calculating the average
        """
        log_util.log_info('Distance and SUS Comparison')
        # for each position in the aligned distance matrices
        self.score_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Float)
        self.undef_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Int)
        
        if self.atom_sel == 'GC':
            alig_array = self.baseset.alig_gc_dist_array
            alig_sus_array = self.baseset.alig_gc_sus_array
        else:
            alig_array = self.baseset.alig_dist_array
            alig_sus_array = self.baseset.alig_sus_array

        
        for i in range(self.baseset.alig.length):
            for j in range(i):
                # value for subset 1, the single entry                
                sentry_id = self.subset1_lst[0]
                dist1 = self.baseset._get_dist(sentry_id, i, j, alig_array)
                distsusv1 = self.baseset._get_distsusv(sentry_id, i, j, alig_sus_array)
                # values for subset2
                dist2_len = len(self.subset2_lst)
                dist2_v = Numeric.zeros(dist2_len, Numeric.Float)
                distsusv2_v = Numeric.zeros(dist2_len, Numeric.Float)
                c = 0
                for sentry_id in self.subset2_lst:
                    dist2_v[c] = self.baseset._get_dist(sentry_id, i, j, alig_array)
                    distsusv2_v[c] = self.baseset._get_distsusv(sentry_id, i, j, alig_sus_array)
                    c += 1
                
                # extract defined distances, 
                # ignore distances not defined (-1) from dist vectors,
                # and ignore the corresponding (same index) distsusv values
                dist2_v, distsusv2_v = self._get_defined_dist_distsus(dist2_v, distsusv2_v)
                # add distsusv1 to distsusv2_v
                if distsusv1 >= 0:
                    distsusv12_v = Numeric.zeros(dist2_len+1, Numeric.Float)
                    distsusv12_v[:len(distsusv2_v)] = distsusv2_v
                    distsusv12_v[len(distsusv2_v)] = distsusv1
                else:
                    distsusv12_v = distsusv2_v
                # get defined distsusv
                defined_distsusv12_v = self._get_defined_v(distsusv12_v)
                # compute mean distsusv, use only defined values, ingore the missing values (-1)
                mean_distsusv12 = self._mean(defined_distsusv12_v)
                # check distance values
                if dist1 < 0  or len(dist2_v) <= 1:
                    # no distance in subset1 or 1 or less distances in subset2
                    # no p value, dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                elif len(defined_distsusv12_v) == 0:
                    # no distance sus, no pvalue, no dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                try:
                    if self.score_type == 'pval':
                        comp_score = self._compute_sus_entry2set_pval(dist1, dist2_v, mean_distsusv12)
                    else:
                        comp_score = self._compute_sus_entry2set_dscore(dist1, dist2_v, mean_distsusv12)
                    if comp_score != None:
                        self.score_mat[i,j] = comp_score
                        self.score_mat[j,i] = comp_score
                    else:
                        self.undef_mat[i,j] = 1
                        self.undef_mat[j,i] = 1
                except:
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    log_util.log_error('Failed comparison i:%d, j:%d dist1:%s dist2:%s' % (i, j, str(subset1_lst), str(subset2_lst)))
        self.write_score_mat_Rdf()
        log_util.log_info('Computed comparison %s' % self.comp_name)       
                
                
                
                                    
    def  comparison_dist_entry2set(self):       
        """
        Comparsion based on distances, compare entry to a set
        subset1_lst contain a single entry
        get pvalues and dscores
        In this case the dscore is actually a zscore
        if no pvalue can be computed, set undef_mat
        Do not compute pvalue or dscore if:
            no distance in subset1 or 1 or less distances in subset2
        """
        log_util.log_info('Distance Comparison')
        # for each position in the aligned distance matrices
        self.score_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Float)
        self.undef_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Int)
        
        if self.atom_sel == 'GC':
            alig_array = self.baseset.alig_gc_dist_array
        else:
            alig_array = self.baseset.alig_dist_array

        for i in range(self.baseset.alig.length):
            for j in range(i):
                # value for subset 1, the single entry                
                sentry_id = self.subset1_lst[0]
                dist1 = self.baseset._get_dist(sentry_id, i, j, alig_array)              
                # values for subset2
                dist2_len = len(self.subset2_lst)
                dist2_v = Numeric.zeros(dist2_len, Numeric.Float)
                c = 0
                for sentry_id in self.subset2_lst:
                    dist2_v[c] = self.baseset._get_dist(sentry_id, i, j, alig_array)
                    c += 1
                # extract defined distances, 
                # ignore distances not defined (-1) from dist vectors,
                dist2_v = self._get_defined_v(dist2_v)    
                # check distance values
                if  dist1 < 0 or len(dist2_v) <=1:
                    # no distance in subset1 or 1 or less distances in subset2
                    # no p value, dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                try:
                    if self.score_type == 'pval':
                        comp_score = self._compute_dist_entry2set_pval(dist1, dist2_v)
                    else:
                        comp_score = self._compute_dist_entry2set_dscore(dist1, dist2_v)    
                    if comp_score != None:
                        self.score_mat[i,j] = comp_score
                        self.score_mat[j,i] = comp_score
                    else:
                        self.undef_mat[i,j] = 1
                        self.undef_mat[j,i] = 1
                except:
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    log_util.log_error('Failed comparison i:%d, j:%d dist1:%s dist2:%s' % (i, j, str(self.subset1_lst), str(self.subset2_lst)))
        self.write_score_mat_Rdf()
        log_util.log_info('Computed comparison %s' % self.comp_name)       
            
        
            
    def comparison_sus_entry2entry(self):
        """
        Distance sus comparison, compare entry to a entry
        get dscores
        if no dscore can be computed, set undef_mat
        Do not compute dscore if:
            no distances
            missing distance sus
        """
        log_util.log_info('Distance and SUS Comparison')
        # for each position in the aligned distance matrices
        self.score_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Float)
        self.undef_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Int)
        
        if self.atom_sel == 'GC':
            alig_array = self.baseset.alig_gc_dist_array
            alig_sus_array = self.baseset.alig_gc_sus_array
        else:
            alig_array = self.baseset.alig_dist_array
            alig_sus_array = self.baseset.alig_sus_array

        for i in range(self.baseset.alig.length):
            for j in range(i):
                # value for subset 1, single entry                
                sentry_id = self.subset1_lst[0]
                dist1 = self.baseset._get_dist(sentry_id, i, j, alig_array)
                distsusv1 = self.baseset._get_distsusv(sentry_id, i, j, alig_sus_array)
                # value for subset 2, single entry                
                sentry_id = self.subset2_lst[0]
                dist2 = self.baseset._get_dist(sentry_id, i, j, alig_array)
                distsusv2 = self.baseset._get_distsusv(sentry_id, i, j, alig_sus_array)
                # check distance values
                if dist1 < 0 or dist2 < 0:
                    # no distances in subset1 or subset2
                    # no dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                elif  distsusv1 < 0 or distsusv2 < 0:
                    # no distances sus in subset1 or subset2
                    # no dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                try:
                    comp_score = self._compute_sus_entry2entry_dscore(dist1, dist2, distsusv1, distsusv2)
                    if comp_score != None:
                        self.score_mat[i,j] = comp_score
                        self.score_mat[j,i] = comp_score
                    else:
                        self.undef_mat[i,j] = 1
                        self.undef_mat[j,i] = 1
                except:
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    log_util.log_error('Failed comparison i:%d, j:%d dist1:%s dist2:%s' % (i, j, str(subset1_lst), str(subset2_lst)))
        self.write_score_mat_Rdf()
        log_util.log_info('Computed comparison %s' % self.comp_name)       

                            
                    
    def comparison_dist_entry2entry(self):
        """
        Distance comparison, compare entry to a entry
        get dscores
        if no dscore can be computed, set undef_mat
        Do not compute dscore if:
            no distances
        """
        log_util.log_info('Distance Comparison')
        self.score_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Float)
        self.undef_mat = Numeric.zeros((self.baseset.alig.length, self.baseset.alig.length), Numeric.Int)
        
        if self.atom_sel == 'GC':
            alig_array = self.baseset.alig_gc_dist_array
        else:
            alig_array = self.baseset.alig_dist_array

        for i in range(self.baseset.alig.length):
            for j in range(i):
                # value for subset 1, the single entry                
                sentry_id = self.subset1_lst[0]
                dist1 = self.baseset._get_dist(sentry_id, i, j, alig_array)                     
                # values for subset2
                sentry_id = self.subset2_lst[0]
                dist2 = self.baseset._get_dist(sentry_id, i, j, alig_array)                           
                if  dist1 < 0 or dist2 < 0:
                    # no distances in subset1 or subset2
                    # no dscore
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    continue
                try:
                    comp_score = self._compute_dist_entry2entry_dscore(dist1, dist2)
                    if comp_score != None:
                        self.score_mat[i,j] = comp_score
                        self.score_mat[j,i] = comp_score
                    else:
                        self.undef_mat[i,j] = 1
                        self.undef_mat[j,i] = 1
                except:
                    self.undef_mat[i,j] = 1
                    self.undef_mat[j,i] = 1
                    log_util.log_error('Failed comparison i:%d, j:%d dist1:%s dist2:%s' % (i, j, str(subset1_lst), str(subset2_lst)))
        self.write_score_mat_Rdf()
        log_util.log_info('Computed comparison %s' % self.comp_name)       

                            
                
    def write_score_mat_Rdf(self):
        """
        Write values to a text file
        To be read by R as data frame
        undefined values declared in undef_dic are set to NA
        """
        if self.score_type == 'pval':
            self.score_file = self.comp_name + '.pval'
        else:
            self.score_file = self.comp_name + '.dsc'
        f_obj = file(self.score_file, 'w')
        for i in range(self.baseset.alig.length):
            for j in range(self.baseset.alig.length):
                if self.undef_mat[i,j] == 1:
                    f_obj.write("NA ")
                else:
                    f_obj.write("%f " % self.score_mat[i,j])
            f_obj.write('\n')     
        f_obj.close()
        log_util.log_info('Wrote %s' % self.score_file)

        
    def write_score_submat_Rdf(self, submat, submat_undef):
        """
        Write values of submatrix to a text file
        To be read by R as data frame
        undefined values declared in submat_undef are set to NA
        """
        if self.score_type == 'pval':
            score_file = self.comp_name + '_site.pval'
        else:
            score_file = self.comp_name + '_site.dsc'
        f_obj = file(score_file, 'w')
        submat_len = submat.shape[0]
        for i in range(submat_len):
            for j in range(submat_len):
                if submat_undef[i,j] == 1:
                    f_obj.write("NA ")
                else:
                    f_obj.write("%f " % submat[i,j])
            f_obj.write('\n')     
        f_obj.close()
        log_util.log_info('Wrote %s' % score_file)
        return score_file
        
        
    def write_subsets(self):
        """
        Debug, list subsets
        """
        c = 0
        str_out = 'subset1: '
        for sentry_id in self.subset1_lst:
            str_out = str_out + "%d,%s " % (c, sentry_id)
            c += 1
        log_util.log_info(str_out)
        c = 0
        str_out = 'subset2: '
        for sentry_id in self.subset2_lst:
            str_out = str_out + "%d,%s " % (c, sentry_id)
            c += 1
        log_util.log_info(str_out)
        

                
    def write_datav(self, i, j, dist1, dist2, distsusv1, distsusv2):
        """
        Debug, write to report log the distance and distance sus vectors
        """        
        str_out = "(%d,%d): dist1:%s\ndistsusv1:%s\ndist2:%s\ndistsusv2:%s" % (i, j, str(dist1), str(distsusv1), str(dist2), str(distsusv2))
        log_util.log_info(str_out)       
        
        
                
    def _compute_sus_set2set_pval(self, dist1_v, dist2_v, mean_distsusv1, mean_distsusv2):
        m = len(dist1_v)
        n = len(dist2_v)
        # do not allow comparison for sets with only one distance value
        # comment, comparison_sus alread did check
        # if m <= 1 or n <= 1:
        #    return None
        mean1 = MLab.mean(dist1_v)
        mean2 = MLab.mean(dist2_v)
        vard1 = self._variance(dist1_v)
        vard2 = self._variance(dist2_v)
        vards1 = vard1 + mean_distsusv1
        vards2 = vard2 + mean_distsusv2
        
        # from DeGroot pg 502,
        # formula 8.6.10
        r1 = vards1/float(m)
        r2 = vards2/float(n)
        test_stat = (mean1 - mean2) / math.sqrt(r1+r2)
        # formula 8.6.13
        s1 = (1.0/ float(m-1)) * ((vards1/float(m)) * (vards1/float(m)))
        s2 = (1.0/ float(n-1)) * ((vards2/float(n)) * (vards2/float(n)))
        v = ((r1 + r2) * (r1 + r2)) / (s1 + s2)
        # if two sided then multiply by 2 
        # pval = 2 * rpy.r.pt(test_stat,v)
        pval = 2 * (1-rpy.r.pt(abs(test_stat),v))
        
        # use 1-pval
        pval = 1.0 - pval
        if type(pval) is types.FloatType:
            if test_stat < 0:
                pval = -pval
            return pval
        else:
            return None

        
        
    def _compute_sus_set2set_dscore(self, dist1_v, dist2_v, mean_distsusv1, mean_distsusv2):
        m = len(dist1_v)
        n = len(dist2_v)
        mean1 = MLab.mean(dist1_v)
        mean2 = MLab.mean(dist2_v)
        vard1 = self._variance(dist1_v)
        vard2 = self._variance(dist2_v)        
        vards1 = vard1 + mean_distsusv1
        vards2 = vard2 + mean_distsusv2
        variance = (vards1/float(m)) + (vards2/float(n))
        sdev = math.sqrt(variance)
        dscore = (mean1 - mean2) / sdev
        return dscore

    
    
    def _compute_dist_set2set_pval(self, dist1_v, dist2_v, i, j):    
        m = len(dist1_v)
        n = len(dist2_v)
        # do not allow comparison for sets with only one distance value
        # checked in comparison_dist
        # if m <= 1 or n <= 1:
        #    return None
        mean1 = MLab.mean(dist1_v)
        mean2 = MLab.mean(dist2_v)
        vard1 = self._variance(dist1_v)
        vard2 = self._variance(dist2_v)
        # from DeGroot pg 502,
        # formula 8.6.10
        r1 = vard1/float(m)
        r2 = vard2/float(n)
        test_stat = (mean1 - mean2) / math.sqrt(r1+r2)
        # formula 8.6.13
        s1 = (1.0/ float(m-1)) * ((vard1/float(m)) * (vard1/float(m)))
        s2 = (1.0/ float(n-1)) * ((vard2/float(n)) * (vard2/float(n)))
        v = ((r1 + r2) * (r1 + r2)) / (s1 + s2)
        # if two sided then multiply by 2 
        # pval = 2 * rpy.r.pt(test_stat,v)
        pval = 2 * (1-rpy.r.pt(abs(test_stat),v))
        
        # use 1-pval
        pval = 1.0 - pval
        if type(pval) is types.FloatType:
            if test_stat < 0:
                pval = -pval
            return pval
        else:
            return None
            
        
        
    def _compute_dist_set2set_pval_ttest(self, dist1_v, dist2_v, i, j):
        # old implementation using ttest    
        m = len(dist1_v)
        n = len(dist2_v)
        # do not allow comparison for sets with only one distance value
        # checked in comparison_dist
        # if m <= 1 or n <= 1:
        #    return None
        stest_dic = rpy.r.t_test(dist1_v,dist2_v,paired= False)
        test_stat = stest_dic['statistic']['t']
        pval = stest_dic['p.value']
        # use 1-pval
        pval = 1.0 - pval
        if type(pval) is types.FloatType:
            if test_stat < 0:
                pval = -pval
                # debug
                # if i == 89 and j == 1:
                #    log_util.log_info('i:%d j:%d pval:%f \ndist1:\n %s\n dist2:\n %s' % (i,j,pval,str(dist1_v), str(dist2_v)))
            return pval
        else:
            return None

        
        
    def _compute_dist_set2set_dscore(self, dist1_v, dist2_v):
        m = len(dist1_v)
        n = len(dist2_v)
        mean1 = MLab.mean(dist1_v)
        mean2 = MLab.mean(dist2_v)
        # check alread in comparison_dist
        # if m <= 1 or n <= 1:
        #    return None
        vard1 = self._variance(dist1_v)
        vard2 = self._variance(dist2_v)
        variance = (vard1/float(m)) + (vard2/float(n))
        sdev = math.sqrt(variance)
        dscore = (mean1 - mean2) / sdev
        return dscore
        

            
    def _compute_sus_entry2set_pval(self, dist1, dist2_v, mean_distsusv12):
        # do not allow comparison for sets with only one distance value
        # cannot be n <= 1, already checked in calling function
        n = len(dist2_v)        
        # if n <= 1:
        #    return None
        mean2 = MLab.mean(dist2_v)
        vard2 = self._variance(dist2_v)
        vards2 = vard2 + mean_distsusv12
        normd = (float(dist1) - mean2)/math.sqrt(vards2)
        pval = rpy.r.pt(-abs(normd), (n-1))*2
        # use 1-pval
        pval = 1.0 - pval
        if type(pval) is types.FloatType:
            if normd < 0:
                pval = -pval
            return pval
        else:
            return None

        
        
    def _compute_sus_entry2set_dscore(self, dist1, dist2_v, mean_distsusv12):
        n = len(dist2_v)
        mean2 = MLab.mean(dist2_v)
        vard2 = self._variance(dist2_v)
        vards2 = vard2 + mean_distsusv12
        sdev = math.sqrt(vards2/float(n))
        dscore = (dist1 - mean2) / sdev
        return dscore
    
    
        
    def _compute_dist_entry2set_pval(self, dist1, dist2_v):
        # do not allow comparison for sets with only one distance value
        # cannot be n <= 1, already checked in calling function
        n = len(dist2_v)
        # if n <= 1:
        #    return None
        mean2 = MLab.mean(dist2_v)
        vard2 = self._variance(dist2_v)
        zscore = (float(dist1) - mean2)/math.sqrt(vard2)
        pval = rpy.r.pt(-abs(zscore), (n-1))*2
        # use 1-pval
        pval = 1.0 - pval        
        if type(pval) is types.FloatType:
            if zscore < 0:
                pval = -pval
            return pval
        else:
            return None
        

            
    def _compute_dist_entry2set_dscore(self, dist1, dist2_v):
        n = len(dist2_v)
        # if n <= 1:
        #    return None        
        mean2 = MLab.mean(dist2_v)
        vard2 = self._variance(dist2_v)
        sdev = math.sqrt(vard2/float(n))
        zscore = (dist1 - mean2) / sdev
        return zscore
    
        

    def _compute_sus_entry2entry_dscore(self, dist1, dist2, distsusv1, distsusv2):
        sdev = math.sqrt(distsusv1 + distsusv2)
        dscore = (dist1 - dist2) / sdev
        return dscore
    
    
    
    def _compute_dist_entry2entry_dscore(self, dist1, dist2):
        dscore = (dist1 - dist2) / (dist1 + dist2)
        return dscore
        
        
    def _variance(self, v):
        """
        Sample variance of v
        """
        mean = MLab.mean(v)
        d = v - mean
        n = v.shape[0]
        var = Numeric.add.reduce(d*d)/(n-1)
        return var
    
                    
                
    def _get_defined_v(self, v):
        return Numeric.compress(Numeric.greater_equal(v,0), v)
        
                   
        
    def _mean(self, v):
        if len(v) == 0:
            return None
        else:
            return MLab.mean(v)
        
        
        
    def _get_defined_dist_distsus(self, dist_v, distsusv_v):
        """
        Given vectors of distances  and distance sus
        return the defined (not -1) distance vector and the correponding sus vector
        """
        defined_dist_v = Numeric.compress(Numeric.greater_equal(dist_v,0), dist_v)
        defined_dist_distsusv_v = Numeric.compress(Numeric.greater_equal(dist_v,0), distsusv_v)
        return (defined_dist_v, defined_dist_distsusv_v)
    
    

