"""
Detect invariant regions, variable segments and hinges
"""


# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de



import os
import cPickle
import types
import copy

import Numeric
import MLab
import rpy


import log_util
import util_r
import clique_search

class HingeInv:
    """
    Detect Hinges Invariant/Variable regions
    """
    
    def __init__(self, detection_label, source_dir, working_dir, save_data):
        # name of detection
        self.detection_label = detection_label
        self.working_dir = working_dir
        self.source_dir = source_dir
        self.save_data = save_data
        # hinges        
        self.hinges_lst = None
        self.hinges_bin_v = None
        self.hinges_resnr_lst = None
        # interhinges
        self.interhinge_lst = None
        self.mean_interhinge_mat = None
        # invariant/variable        
        self.invariant_lst = None
        self.variable_lst = None
        self.inv_mat = None
        self.inv_bin_mat = None
        self.inv_region_lst = []
        self.variable_resnr_lst = None
        self.inv_region_resnr_lst = None
        
        
    
    def get_hinges_invariant(self, score_mat, undef_mat, score_cutoff, filter_cutoff, min_gap, inv_cutoff, mat_type = 'variation', filter_smooth = True):

        # filter score_mat, get bin_mat
        f_mat = self.filter_score_mat(score_mat, undef_mat, score_cutoff, mat_type)
        
        # filter column difference vector, get hinges        
        self.get_hinges(f_mat, filter_cutoff, min_gap, undef_mat, filter_smooth)
        
        # get invariant
        absf_mat = Numeric.absolute(f_mat)
        self.get_invariant(absf_mat, inv_cutoff, undef_mat)

        # write invariant bin matrix as csv
        # csv_file = os.path.join(self.working_dir, self.detection_label + '_inv_bin.csv')
        # self.write_bin_mat(self.inv_bin_mat,csv_file)
        # log_util.log_info('Saved %s' % csv_file)
        
        # get invariant regions
        self.get_invariant_regions()
                
        log_util.log_info('Hinge segments %s\n%s' % (self.detection_label, str(self.hinges_lst)))
        if len(self.variable_lst) > 0:
            log_util.log_info('Variable segments %s\n%s' % (self.detection_label,str(self.variable_lst)))
        log_util.log_info('Invariant segments %s\n%s' % (self.detection_label,str(self.invariant_lst)))
        log_util.log_info('Invariant regions %s\n%s' % (self.detection_label,str(self.inv_region_lst)))

                

    def filter_score_mat(self, score_mat, undef_mat, score_cutoff, mat_type):
        """
        Filtered matrix,
        1 if above score_cutoff
        -1 if below -score_cutoff
        0 otherwise or undefined
        """
        if mat_type == 'comp':
            score_mat2 = Numeric.absolute(score_mat)
            # greater_mat = Numeric.greater_equal(score_mat, score_cutoff)
            # less_mat = Numeric.less_equal(score_mat, -score_cutoff)
            # filtered_mat = copy.copy(greater_mat)
            # Numeric.putmask(filtered_mat,less_mat,-1)        
        else:
            score_mat2 = score_mat
        filtered_mat = Numeric.greater_equal(score_mat2, score_cutoff)
        Numeric.putmask(filtered_mat, undef_mat, 0)      
        return filtered_mat
        
    
            
    def get_hinges(self, f_mat, filter_cutoff, min_gap, undef_mat, filter_smooth =True):
        
        # column difference, get col_diff_v
        col_diff_v = self.get_col_diff(f_mat, undef_mat)
        
        # smooth
        if filter_smooth == True:
            smooth_v = rpy.r.smooth(x=col_diff_v , kind = "3")   # Tukey's smoother
            # spline_smooth_obj = rpy.r.smooth_spline(x=col_diff_v)  # cubic smoothing spline
            # smooth_v = spline_smooth_obj['y']
            # cutoff filter
        else:
            smooth_v = col_diff_v
        bin_v = Numeric.greater_equal(smooth_v, filter_cutoff)
        
        # gaps filter, get hinges_bin_v
        self.get_hinges_v(bin_v, min_gap)
        
        # save plots
        plot_file = os.path.join(self.working_dir, self.detection_label + "_hingeinv.png")
        self.plot_curves(col_diff_v, smooth_v, bin_v, self.hinges_bin_v, plot_file)
        log_util.log_info('Saved %s' % plot_file)
        # plot_file = self.detection_label + "_hingeinv.pdf"
        # self.plot_curves_pdf(col_diff_v, smooth_v, self.hinges_bin_v, plot_file)
        # log_util.log_info('Saved %s' % plot_file)
        # save data
        if self.save_data == True:
            data_lst = [col_diff_v, smooth_v, bin_v, self.hinges_bin_v, "col_diff_v, smooth_v, bin_v, self.hinges_bin_v"]
            data_file = os.path.join(self.working_dir, self.detection_label + "_hingeinv.Rdata")
            util_r.save_obj_r(data_lst, data_file)
        
        
        
    def get_invariant(self, f_mat, inv_cutoff, undef_mat):
        n = len(f_mat)
        
        if len(self.hinges_lst) == 0:
            self.interhinge_lst = [(0,n-1)]
            self.invariant_lst = [(0,n-1)]
            self.variable_lst = []
            self.inv_bin_mat = Numeric.array([1], Numeric.Int)
            return
            
        # get interhinge segments
        interhinge_lst = []
        if self.hinges_lst[0][0] == 0:
            i_st = self.hinges_lst[0][1] + 1
        else:
            i_st = 0
        for h_seg in self.hinges_lst:       
            if h_seg[0] > 0:
                i_end = h_seg[0] - 1
                interhinge_lst.append((i_st, i_end))
                i_st = h_seg[1] +1
        if i_st < (n -1):
            i_end = n-1
            interhinge_lst.append((i_st, i_end))
        self.interhinge_lst = interhinge_lst
        
                    
        # compare interhinges to itself, identify invariant/variant segments
        invariant_lst = []
        variable_lst = []
        seg_counter = 0
        for seg in self.interhinge_lst:
            subst_mat = f_mat[seg[0]:(seg[1]+1),seg[0]:(seg[1]+1)]
            undef_subst = undef_mat[seg[0]:(seg[1]+1),seg[0]:(seg[1]+1)]
            subst_mean = self.get_mean_submat(subst_mat, undef_subst)
            if subst_mean <= inv_cutoff:
                invariant_lst.append(seg)
            else:
                variable_lst.append(seg)
            seg_counter +=1
        self.invariant_lst = invariant_lst
        self.variable_lst = variable_lst
            
        # get matrix with scores between invariant segments
        inv_mat = Numeric.zeros((len(self.invariant_lst), len(self.invariant_lst)), Numeric.Float)
        for i_counter in range(len(self.invariant_lst)):
            for j_counter in range(i_counter+1):
                i_seg = self.invariant_lst[i_counter]
                j_seg = self.invariant_lst[j_counter]
                subst_mat = f_mat[i_seg[0]:(i_seg[1]+1),j_seg[0]:(j_seg[1]+1)]
                undef_subst = undef_mat[i_seg[0]:(i_seg[1]+1),j_seg[0]:(j_seg[1]+1)]
                subst_mean = self.get_mean_submat(subst_mat, undef_subst)                
                inv_mat[i_counter][j_counter] = subst_mean
                inv_mat[j_counter][i_counter] = subst_mean
                j_counter +=1
            i_counter +=1                
        self.inv_mat = inv_mat
        # get binary matrix for inv_mat, apply the inv_cutoff
        self.inv_bin_mat = Numeric.less_equal(self.inv_mat, inv_cutoff)
                

        
    def get_mean_submat(self, submat, undefmat):
        # ignore undef  on computing mean
        # undef values in submat are 0
        nr_undef = Numeric.sum(Numeric.sum(undefmat))           
        nr_val = (Numeric.shape(submat)[0] * Numeric.shape(submat)[1]) - nr_undef
        if nr_val > 0:
            subst_mean = (Numeric.sum(Numeric.sum(submat)))/ float(nr_val)
        else:
            subst_mean = 0.0
        return subst_mean
        
        
            
    def get_col_diff(self, mat, undef_mat):
        n = len(mat)
        # enforce col_diff_v same size as mat, last value is from n-2
        col_diff_v = Numeric.zeros(n,Numeric.Float)
        for i in range(n-1):
            v1 = mat[:,i]
            v2 = mat[:,i+1]
            u1 =  undef_mat[:,i]
            u2 =  undef_mat[:,i+1]
            u = copy.copy(u1)
            Numeric.putmask(u, u2, 1)
            # ignore undefined
            f1 = Numeric.compress(Numeric.equal(u,0), v1)
            f2 = Numeric.compress(Numeric.equal(u,0), v2)
            if len(f1) > 0:
                col_diff_v[i] = MLab.mean(Numeric.absolute(f1-f2))
            else:
                col_diff_v[i] = 0.0
        col_diff_v[n-1] = col_diff_v[i]
        return col_diff_v

        
        
    def get_hinges_v(self, bin_v, min_gap):
        # get list of segments
        n = len(bin_v)
        seg_lst = []
        seg_start = -1
        seg_end = -1
        for i in range(n):
            if bin_v[i]:
                if seg_start >=0:
                    seg_end = i
                else:
                    seg_start = i
                    seg_end = i
            else:
                if seg_start >=0:
                    seg_t = (seg_start,seg_end)
                    seg_lst.append(seg_t)
                    seg_start = -1
                    seg_end = -1
        if seg_start >=0:
            seg_t = (seg_start,seg_end)
            seg_lst.append(seg_t)
          
            
        # merge segments with gap less then min_gap
        n = len(seg_lst)
        segf_lst = []
        if n>0: 
            last_seg = seg_lst[0]            
            for i in range(1,n):
                seg = seg_lst[i]
                if seg[0]-last_seg[1] > min_gap:
                    segf_lst.append(last_seg)
                    last_seg = seg
                else:
                    last_seg = (last_seg[0],seg[1])
            segf_lst.append(last_seg)
        # check for ends, if hinges to close to ends, then extend them to the ends
        n = len(bin_v)
        if len(segf_lst) > 0:
            if (segf_lst[0][0] <= min_gap):
                segf_lst[0] = (0, segf_lst[0][1])
            if (n-1 - segf_lst[-1][1]) <= min_gap:
                segf_lst[-1] = (segf_lst[-1][0], (n-1))
            self.hinges_lst = segf_lst
            # get vector of hinges from list
            hinges_bin_v = Numeric.zeros(len(bin_v), Numeric.Int)
            for seg in segf_lst:
                st_pos = seg[0]
                end_pos = seg[1]
                for i in range(st_pos,end_pos+1):
                    hinges_bin_v[i] = 1
        else:
            self.hinges_lst = []
            hinges_bin_v = Numeric.zeros(len(bin_v), Numeric.Int)
            
        self.hinges_bin_v = hinges_bin_v                
        return hinges_bin_v
        
        
        
    
    
    def plot_curves(self, col_diff_v, smooth_v, bin_v, hinges_bin_v, file_name):
        rpy.r.png(filename= file_name, width = 1024, height = 1024)
        rpy.r.plot(x = col_diff_v, type = 'l', xlab='', ylab='', col='black', main = '', lwd=2)
        rpy.r.lines(x = smooth_v, type = 'l', ylab='', col='orange', lwd=2)
        # rpy.r.lines(bin_v/3.0, type = 'l', ylab='', col='yellow')
        rpy.r.lines(x = hinges_bin_v/3.0, type = 'l', ylab='', col='steelblue3', lwd=2)
        # rpy.r.legend(330, 0.41, legend=['col_diff','smooth','smooth > 0.1', 'min_gap 10'], fill=['black', 'orange', 'yellow', 'red'])
        rpy.r.dev_off()
        
        
        
    def plot_curves_pdf(self, col_diff_v, smooth_v, hinges_bin_v, file_name):
        max_val = MLab.max(smooth_v)
        rpy.r.pdf(file = file_name, paper="special", horizontal = False, width = 10, height = 10)
        rpy.r.par(cex=1.5, mar = [7,7,7,7])
        rpy.r.plot(x = col_diff_v, type = 'l', xlab='', ylab='', col='black', main = '', lwd=1)
        rpy.r.lines(x = smooth_v, type = 'l', xlab = '', ylab='', col='orange', lwd=1, lty=5)
        # rpy.r.lines(bin_v/3.0, type = 'l', ylab='', col='yellow')
        rpy.r.lines(x = hinges_bin_v*float(max_val), type = 'l', ylab='', col='steelblue3', lwd=1, lty=1)
        # rpy.r.legend(x=330, y=0.41, legend=['smooth','hinges'], col=['orange', 'red'], lty=[5,1])
        rpy.r.dev_off()

        
                
    def write_bin_mat(self, mat, out_file):
        """
        Write binary matrix as csv file
        """
        fobj = file(out_file, 'w')
        d1 = Numeric.shape(mat)[0]
        if d1>1 :
            d2 = Numeric.shape(mat)[1]    
            for i in range(d1):
                for j in range(d2-1):
                    fobj.write("%d," % mat[i][j])
                fobj.write("%d\n" % mat[i][d2-1])
        else:
            for i in range(d1):
                fobj.write("%d\n" % mat[i])
        fobj.close()
            
    
    
    def get_invariant_regions(self):
        """
        Get invariant regions by clique search
        """        
        cs_obj = clique_search.CliqueS()
        cs_obj.set_graph(self.inv_bin_mat)
        cs_obj.init_search()
        inv_region_idx_lst = cs_obj.get_cliques()
        self.get_region_from_idx(inv_region_idx_lst)        
        log_util.log_info('Computed Invariant regions')

        
        
    def get_region_from_idx(self, inv_region_idx_lst):
        inv_region_lst = [] 
        for region_idx in inv_region_idx_lst:
            region_lst = []
            for seg_idx in region_idx:
                seg = self.invariant_lst[seg_idx]
                region_lst.append(seg)
            inv_region_lst.append(region_lst)
        self.inv_region_lst = inv_region_lst
                
            

    def get_res_nr(self, alig_obj):
        self.hinges_resnr_lst = self.refpos_lst_to_res_nr(self.hinges_lst, alig_obj)
        self.variable_resnr_lst = self.refpos_lst_to_res_nr(self.variable_lst, alig_obj)
        self.inv_region_resnr_lst = self.refpos_lst_to_res_nr(self.inv_region_lst, alig_obj)
        log_util.log_info('Hinge segments res nr %s\n%s' % (self.detection_label, str(self.hinges_resnr_lst)))
        if len(self.variable_lst) > 0:
            log_util.log_info('Variable segments res nr %s\n%s' % (self.detection_label,str(self.variable_resnr_lst)))
        log_util.log_info('Invariant regions res nr %s\n%s' % (self.detection_label,str(self.inv_region_resnr_lst)))
        
          
    def refpos_lst_to_res_nr(self, lst, alig_obj):
        res_nr_lst = []
        if len(lst) == 0:
            return res_nr_lst
        elif type(lst[0]) is types.TupleType:
            for tpl in lst:
                refpos0 = tpl[0] 
                res_nr0 = alig_obj.refpos_to_uniprot_resnr(refpos0)
                refpos1 = tpl[1] 
                res_nr1 = alig_obj.refpos_to_uniprot_resnr(refpos1)
                res_nr_tpl = (res_nr0, res_nr1)
                res_nr_lst.append(res_nr_tpl)
        else:
            for lst2 in lst:
                res_nr_lst2 = []
                for tpl in lst2:
                    refpos0 = tpl[0] 
                    res_nr0 = alig_obj.refpos_to_uniprot_resnr(refpos0)
                    refpos1 = tpl[1] 
                    res_nr1 = alig_obj.refpos_to_uniprot_resnr(refpos1)
                    res_nr_tpl = (res_nr0, res_nr1)
                    res_nr_lst2.append(res_nr_tpl)
                res_nr_lst.append(res_nr_lst2)
        return res_nr_lst
                                
                        
                         
    def save_hinge_inv(self, datafile):
        hinge_inv_dic = self.get_hinge_inv_dic()
        pickle_protocol = 2
        fobj = file(datafile, 'w')
        cPickle.dump(hinge_inv_dic,fobj,pickle_protocol)
        fobj.close()
        log_util.log_info('Saved %s' % datafile)
        return hinge_inv_dic
       
        
        
    def get_hinge_inv_dic(self):
        hinge_inv_dic = {}
        hinge_inv_dic['hinges_lst'] = self.hinges_lst
        hinge_inv_dic['variable_lst'] = self.variable_lst
        hinge_inv_dic['inv_region_lst'] = self.inv_region_lst
        hinge_inv_dic['hinges_resnr_lst'] = self.hinges_resnr_lst
        hinge_inv_dic['variable_resnr_lst'] = self.variable_resnr_lst
        hinge_inv_dic['inv_region_resnr_lst'] = self.inv_region_resnr_lst
        return hinge_inv_dic
    
        
               
    ###########################################
    # OLD
    def get_smooth_mat(self, mat, smooth_w):
        # old
        mat_dim = len(mat) - smooth_w +1
        smooth_mat = Numeric.zeros( (mat_dim,mat_dim), Numeric.Float)
        for i in range(mat_dim):
            for j in range(mat_dim):
                submat = mat[i:(i+smooth_w), j:(j+smooth_w)]
                smooth_mat[i,j] = MLab.mean(MLab.mean(submat))
        return smooth_mat
    
            
                        
    def filter_mat(self, mat, cutoff):
        # old
        filtered_mat = Numeric.greater_equal(mat, cutoff)
        return filtered_mat

    
    
    def get_invariant_regions_old(self, csv_file, clique_file):
        """
        Old Implementation based on the C clique search program
        """    
        # clique search     
        cliques_exec = os.path.join(self.source_dir, 'C', 'find_clique' )
        cmd = cliques_exec + ' ' + csv_file + ' > ' +  clique_file
        try:                                
            ret_val = os.system(cmd)
        except:
            log_util.log_error('Could not execute %s' % cmd)
        log_util.log_info(cmd)
        # process output
        inv_region_idx_lst = self.read_clique_out(clique_file)
        self.inv_region_lst = self.get_region_from_idx(inv_region_idx_lst)
        
        log_util.log_info('Computed cliques: %s' % clique_file)
        
        
        
    def read_clique_out(self, filename):
        fobj = file(filename, 'r')
        lines_lst = fobj.readlines()
        fobj.close()
        
        region_lst = []
        cl_line = 0
        for line_str in lines_lst:
            if cl_line == 1:
                # get clique line
                line_lst = line_str.strip().split()
                cl_lst = []
                for inx_str in line_lst:
                    cl_lst.append(int(inx_str))
                    cl_lst.sort()
                region_lst.append(cl_lst)
                cl_line = 0            
            elif line_str[:6] == 'Size =':
                # mark nex line as clique line
                cl_line = 1
        return region_lst

                
        

