"""
pdbsws_map.py
Extract residue level mapping data
from text file provided by PDBSws
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import os
import os.path
import cPickle
import log_util
import stmap_resmap


def get_pdb_fasta(pdbswsmap_file, seqdb_lst, fasta_dir):
    """
    Get the FASTA files for the entries in PDB according to pdbsws
    get sequence from the sequence database files in seqdb_lst list
    """
    # extract the uniprot accession numbers in pdbsws
    uniprot_dic = {}
    fobj = file(pdbswsmap_file, 'r')
    for line_str in fobj:
        line_lst = line_str.split()        # split each column
        if len(line_lst) != 8:
            # not according to format, skip
            continue
        uniprot_ac = line_lst[5]
        if uniprot_dic.has_key(uniprot_ac) == 0:
            uniprot_dic[uniprot_ac] = ''
    log_util.log_info("Number of Uniprot sequences: %d" % len(uniprot_dic))
    fobj.close()

    # save sequence files
    for seqdb in seqdb_lst:
        # for each sequence database
        fobj = file(seqdb, 'r')
        pdb_seq_ac = None
        for line_str in fobj:
            line_str = line_str.strip()
            if line_str[0] == '>':
                # new sequence
                if pdb_seq_ac != None:
                    # save old sequence
                    save_fasta(seq_str, ac, fasta_dir)
                    del uniprot_dic[ac]
                seq_str = ''
                ac_name = line_str.split()[0]
                ac = ac_name.split('|')[0]
                ac = ac[1:]
                if uniprot_dic.has_key(ac):
                    # pdb sequence, get it
                    pdb_seq_ac = ac
                else:
                    # not pdb seq, skip
                    pdb_seq_ac = None
            elif line_str[0] != '>' and line_str != '':
                if pdb_seq_ac != None:
                    # sequence string
                    seq_str = seq_str + line_str + '\n'
            if len(uniprot_dic) == 0:
                break
        if pdb_seq_ac != None:
            # save old sequence
            save_fasta(seq_str, ac, fasta_dir)
            del uniprot_dic[ac]
        fobj.close()
        log_util.log_info("Scanned %s" %  seqdb)
    if len(uniprot_dic) == 0:
        log_util.log_info("Got sequences for all PDB entries")
    else:
        log_util.log_info("Missing %d sequences:\n %s" % (len(uniprot_dic), str(uniprot_dic.keys())))



def save_fasta(seq_str, ac, fasta_dir):
    """
    Save sequence string in fasta format to file ac.fasta in fasta_dir
    """

    fasta_filename = ac + '.fasta'
    fasta_file = os.path.join(fasta_dir, fasta_filename)
    fobj_fasta = file(fasta_file, 'w')
    seq_label = '>' + ac + '\n'
    fobj_fasta.write(seq_label)
    fobj_fasta.write(seq_str)
    fobj_fasta.close()    
    log_util.log_info("Saved %s" % fasta_file)



def get_map_file_name(sentry_id):
    """
    From sentry_id get the map_file_name.
    """
    pdb_id = sentry_id[1:5]
    map_file_name = pdb_id + ".data"
    return map_file_name



class PDBSwsMap:
    """
    PDBSws residue mapping between PDB and uniprot
    """

    def __init__(self, pdbswsmap_file, stmap_dir):
        self.pdbswsmap_file =  pdbswsmap_file # text file with residue level mapping from pdbsws
        self.stmap_dir = stmap_dir



    def compile_stmaps(self):
        """
        From the pdbsws mapping file, generate the stmap data files
        """
        st_db = 'PDB'
        st_dbversion = 'NoVersion'
        st_dbcoordsys = 'PDBresnum'
        up_dbversion = 'NoVersion'
        up_dbcoordsys = 'UniProt'
        last_st_id = ''

        # for each line
        log_util.log_info('Start compiling stmaps for %s' % self.pdbswsmap_file)
        for pdbsws_line in file(self.pdbswsmap_file):
            line_lst = pdbsws_line.split()        # split each column
            if len(line_lst) != 8:
                # not according to format, skip
                continue
            st_id = line_lst[0]
            # test if new pdb
            if st_id != last_st_id:
                if len(st_id) != 4:
                    # not a pdb id, skip
                    continue
                # save previous stmap
                if len(last_st_id) == 4:
                    self.save_stmap(stmap)
                # start new stmap
                stmap = stmap_resmap.StMap(st_db , st_dbcoordsys, st_dbversion, st_id)
                last_st_id = st_id
            # add new resmap
            st_chain = line_lst[1]
            if st_chain == '@':
                st_chain = ' '
            st_resname = line_lst[3]
            st_reslabel_raw = line_lst[4]
            up_id = line_lst[5]
            up_resname = line_lst[6]
            up_resnr = int(line_lst[7])
            # change residue label, add space if no insertion code
            st_reslabel = self._get_reslabel(st_reslabel_raw)
            stmap.add_resmap(st_dbcoordsys, st_dbversion, st_chain, st_reslabel, st_resname, up_dbversion, up_dbcoordsys, up_id, up_resnr, up_resname)

        # save the last stmap
        if len(last_st_id) == 4:
            self.save_stmap(stmap)
        log_util.log_info('Finished compiling stmaps for %s' % self.pdbswsmap_file)



    def save_stmap(self, stmap):
        """
        Save the stmap as a byte stream
        For each entry, file name is stmap.st_id.data
        """
        pickle_protocol = 2 # use binary

        data_file = stmap.st_id + '.data'
        path_data_file = os.path.join(self.stmap_dir, data_file)
        outf_obj = file(path_data_file, 'w')
        cPickle.dump(stmap, outf_obj, pickle_protocol)
        outf_obj.close()
        log_util.log_info('Saved %s' % data_file)



    def _get_reslabel(self, st_reslabel_raw):
        """
        Modify original residue label.
        Keep it if there is insertion code
        If no insertion code, then insert space ' '
        """
        
        if st_reslabel_raw[-1].isalpha():
             st_reslabel = st_reslabel_raw
        else:
            st_reslabel = st_reslabel_raw + ' '
        return st_reslabel
        