"""
Read Species sets from SCOP and get a baseset.
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import os

import log_util
import sentry_set
import base_set
import sus_dpi
import pdbsws_map
import stmap_resmap


class ReadScopBaseset:
    """
    Class to process a scop domain definition file.

    Extract the baseset for all sentries at a species level.
    """

    def __init__(self, min_set_len = 2, max_set_len = None, atom_sel='CA'):
        self.sets_lines = None
        self.nr_lines = None
        self.line_counter = 0
        self.min_set_len =   min_set_len                         # minimum number of sentries in set
        self.max_set_len = max_set_len                           # max number of entries in set, if no max then None
        self.atom_sel = atom_sel
        # temporary data for each baseset
        self.baseset_lines_dic = {}  # dict with scop lines for a sentry: dic[sentry_id] scop_line

        
        
    def load_file(self, sets_file):
        """
        Read the file with defintion of sets into self.sets_lines.
        """
        f_obj = file(sets_file, 'r')
        self.sets_lines = f_obj.readlines()
        f_obj.close()
        self.nr_lines = len(self.sets_lines)
        log_util.log_info('Read sets defintion file %s' % sets_file)


        
    def get_baseset(self, pdb_dir, use_pdb_subdir = True):
        """
        Read set defintion from scop.
        
        Return the baseset
        Specific to SCOP database
        If it fails returns None
        """

        # get definiton lines for set
        baseset_lines = self._get_baseset_lines()
        # check there is any sentry in set
        if len(baseset_lines) == 0:
            return None
        # get baseset id, the scop species sunid
        baseset_id = self._get_sp_sunid(baseset_lines[0])
        # check set size
        if self._check_baseset_size(len(baseset_lines)) == 0:
            log_util.log_warning('Skip set %s, size %d' % (baseset_id, len(baseset_lines)))
            return None
        # get list of sentry id
        sentry_id_lst = self._sentry_id_lst(baseset_lines)
        log_util.log_info('Reading new baseset %s' % baseset_id)

        # read into baseset object        
        sentryset = sentry_set.SentrySet(baseset_id, use_pdb_subdir, self.atom_sel)
        sentryset.get_sentries(sentry_id_lst, pdb_dir)
        baseset = base_set.BaseSet(sentryset)

        # check baseset size
        if self._check_baseset_size(baseset.sentryset.set_len) == 0:
            log_util.log_warning('Skip set %s, baseset size %d' % (baseset_id, len(baseset_lines)))
            return None

        log_util.log_info('Read set %s' %  baseset.set_id)
        return baseset

        

    def current_line(self):
        return self.sets_lines[self.line_counter]


        
    def _get_baseset_lines(self):
        """
        Return the scop definition lines for a baseset.
        """
        baseset_lines = []
        self.baseset_lines_dic = {}
        last_sunid = ''

        while(self.line_counter < self.nr_lines):
            scop_line = self.sets_lines[self.line_counter] 
            if len(scop_line.strip()) == 0 or scop_line.strip()[0] == '#':
                # Skip empty lines and comments
                self.line_counter += 1
                continue
            sp_sunid = self._get_sp_sunid(scop_line)
            sid = scop_line.strip().split()[0]
            if sp_sunid == last_sunid:
                # append to set
                baseset_lines.append(scop_line)
                self.baseset_lines_dic[sid] = scop_line
            elif last_sunid == "":
                # first element in new set
                baseset_lines.append(scop_line)
                self.baseset_lines_dic[sid] = scop_line
                last_sunid = sp_sunid
            else:
                # finished set, start of new set
                break
            self.line_counter += 1
        return baseset_lines
        


    def _get_sp_sunid(self, scop_line):
        """
        Given a SCOP domain defintion line, get the species sunid.
        """
        line_lst = scop_line.split()
        sunid_lst = line_lst[5].split(',')
        sp_sunid = sunid_lst[5]
        if sp_sunid[:3] != 'sp=':
            raise RuntimeError, "SCOP line %s" % scop_line
        sp_sunid = sp_sunid[3:]
        return sp_sunid
        

    def _sentry_id_lst(self, baseset_lines):
        """
        Given the scop domain defintion lines for a set extract the scop sids.

        SCOP sids are used as the sentry ids.
        """
        sentry_id_lst = []
        for line in baseset_lines:
            scop_sid = line.split()[0]
            sentry_id_lst.append(scop_sid)
        return sentry_id_lst



    def _check_baseset_size(self, set_len):
        """
        Check nr of entries in set.

        Return 0 if set_len < self.min_set_len
        if self.max_set_len defined
        then return 0 if set_len > self.max_set_len 
        Return 1 otherwise
        """
        if set_len < self.min_set_len:
            return 0
        elif self.max_set_len != None:
            if set_len > self.max_set_len:
                return 0
        return 1

        
        
    
            
class ScopSet:
    """
    SCOP statistics and filters
    """
    
    def __init__(self, scop_file):
        self.scop_file = scop_file
        self.pdb_dir = None
        self.sus_file = None
        self.stmap_dir = None

        
        
    def class_filter(self, filter_file):
           
        no_class_lst = ['57942', '58117', '58231', '58788']
        f_obj = file(self.scop_file, 'r')
        lines = f_obj.readlines()
        f_obj.close()
        log_util.log_info('Filtering SCOP file %s' % self.scop_file)
               
        fout_obj = file(filter_file, 'w')
           
        for scop_line in lines:
            if len(scop_line.strip()) == 0 or scop_line.strip()[0] == '#':
                continue                
            line_lst = scop_line.split()
            sunid_lst = line_lst[5].split(',')
            cl_sunid = sunid_lst[0][3:]
            no_class_match = 0
            for no_class in no_class_lst:
                if cl_sunid == no_class:
                    no_class_match = 1
                    break
            if no_class_match == 0:
                fout_obj.write(scop_line)            
           
        fout_obj.close()
        log_util.log_info('Wrote %s' % filter_file)
        

        
    def sentry_filter(self, filter_file, pdb_dir, sus_file, stmap_dir):
        self.pdb_dir = pdb_dir
        self.sus_file = sus_file
        self.stmap_dir = stmap_dir
        
        min_set_len = 1
        
        log_util.log_info('Filtering SCOP file %s' % self.scop_file)
        
        f_obj = file(filter_file, 'w')
        
        # sus
        suspdb_obj = self._get_sus()

        # read scop
        read_scop = ReadScopBaseset(min_set_len)
        read_scop.load_file(self.scop_file)
        while(read_scop.line_counter < read_scop.nr_lines):
            last_lnr = read_scop.line_counter
            baseset = None

            try:
                use_pdb_subdir = True
                baseset = read_scop.get_baseset(self.pdb_dir, use_pdb_subdir)
                if baseset == None:
                    log_util.log_error('Could not get baseset scop set line nr: %d' % read_scop.line_counter)
                    if read_scop.line_counter == last_lnr:
                        read_scop.line_counter += 1
                    continue
                # for each species set
                (baseset,nr_uniprot) = self._stmap_baseset(baseset)
                nr_sus = self._get_nr_sus(baseset, suspdb_obj)
                if nr_sus >=2 and nr_uniprot == 1:
                    # more than 2 sus, only 1 uniprot
                    for sentry in baseset.sentryset.sentry_lst:
                        sentry_id = sentry.sentry_id
                        if  baseset.sentry_sus_lst.count(sentry_id) > 0:
                            if read_scop.baseset_lines_dic.has_key(sentry_id):
                                f_obj.write(read_scop.baseset_lines_dic[sentry_id])
                                f_obj.flush()
                            else:
                                log_util.error('No sentry in read_scop.baseset_lines_dic %s' % sentry_id)
            except:
                log_util.log_error('Could not read scop set line nr: %d' % read_scop.line_counter)
                if read_scop.line_counter == last_lnr:
                    read_scop.line_counter += 1
                               
        f_obj.close()
        log_util.log_info('Finished Filtering SCOP, wrote %s' % filter_file)
        
        
                    
    def get_stats(self, file_name, pdb_dir, sus_file, stmap_dir):
        self.pdb_dir = pdb_dir
        self.sus_file = sus_file
        self.stmap_dir = stmap_dir
        
        
        f_obj = file(file_name, 'w')
        out_str = "sunid nr_sentries nr_stmap nr_sus ratio_sus nr_uniprot\n"
        f_obj.write(out_str)
        
        min_set_len = 1
        
        # sus
        suspdb_obj = self._get_sus()
        
        # read scop
        read_scop = ReadScopBaseset(min_set_len)
        read_scop.load_file(self.scop_file)
        
        while(read_scop.line_counter < read_scop.nr_lines):
            last_lnr = read_scop.line_counter
            baseset = None
            
            try:
                use_pdb_subdir = True
                baseset = read_scop.get_baseset(self.pdb_dir, use_pdb_subdir)
                if baseset == None:
                    log_util.log_error('Could not get baseset scop set line nr: %d' % read_scop.line_counter)
                    if read_scop.line_counter == last_lnr:
                        read_scop.line_counter += 1
                    continue
                nr_sentries = baseset.set_len
                # for each species set
                (baseset,nr_uniprot) = self._stmap_baseset(baseset)
                nr_stmap = baseset.set_len
                nr_sus = self._get_nr_sus(baseset, suspdb_obj)
                if  nr_stmap > 0:
                    ratio_sus = float(nr_sus) / float(nr_stmap)
                else:
                    ratio_sus = 0.0
                f_obj.write("%-10s %4d %4d %4d %5.2f %4d\n" % (baseset.set_id, nr_sentries, nr_stmap, nr_sus, ratio_sus, nr_uniprot))
                f_obj.flush()
            except:
                log_util.log_error('Could not read scop set line nr: %d' % read_scop.line_counter)
                if read_scop.line_counter == last_lnr:
                    read_scop.line_counter += 1
        
        f_obj.close()
        log_util.log_info('Finished SCOP Stats, wrote %s' % file_name)
            
            
        
    def _get_sus(self):
        suspdb_obj = sus_dpi.SUSpdb()
        suspdb_obj.read_sus_file(self.sus_file)
        return suspdb_obj
        
        
        
    def _get_nr_sus(self, baseset, suspdb_obj):
        baseset._get_dpi_dic(self.sus_file)
        nr_sus = len(baseset.sentry_sus_lst)
        return nr_sus
        
        
        
    def _stmap_baseset(self, baseset):
        map_cov_cutoff = 0.5
        uniprot_dic = {}
        remove_sentry_lst = []
        for sentry in baseset.sentryset.sentry_lst:
            log_util.log_info("Sentry %s length %d" % (sentry.sentry_id, sentry.length))  # debug
            # for each sentry
            uniprot_sentry_dic = {}
            map_file_name = pdbsws_map.get_map_file_name(sentry.sentry_id)
            map_file = os.path.join(self.stmap_dir, map_file_name)
            if os.path.exists(map_file):
                stmap = stmap_resmap.load_map_data(map_file)
            else:
                log_util.log_warning('No map file %s' % map_file)
                remove_sentry_lst.append(sentry.sentry_id)
                continue
            stmap_sentry_cov = 0
            for pos in range(sentry.length):
                # go over each sentry position
                residue = sentry.res_lst[pos]
                resmap = stmap.get_resmap(residue.label, residue.chain_id)
                if resmap != None:                    
                    stmap_sentry_cov +=1
                    uniprot_id = resmap.up_id
                    if uniprot_sentry_dic.has_key(uniprot_id) == False:
                        uniprot_sentry_dic[uniprot_id] = 1
            if sentry.length < 1:
                log_util.log_warning('Sentry with no residues %s %s' % (sentry.sentry_id, str(sentry.length)))
                map_cov = 0.0
            else:            
                map_cov = float(stmap_sentry_cov) / float(sentry.length)
            if  map_cov < map_cov_cutoff:
                log_util.log_warning('Sentry with low stmap coverage: %s %f' % (sentry.sentry_id, map_cov))
                remove_sentry_lst.append(sentry.sentry_id)
                continue
                        
            if len(uniprot_sentry_dic) > 1:
                # more than one uniprot in one sentry
                log_util.log_warning('Sentry with more than one uniprot %s: %s' % (sentry.sentry_id, str(uniprot_sentry_dic.keys())))
            
            for uniprot_id in uniprot_sentry_dic.keys():
                # store sentry results
                if uniprot_dic.has_key(uniprot_id):
                    uniprot_dic[uniprot_id].append(sentry.sentry_id)
                else:
                    uniprot_dic[uniprot_id] = [sentry.sentry_id]                
        
        baseset.delete_sentries(remove_sentry_lst)
                                                  
        nr_uniprot = len(uniprot_dic)
        if nr_uniprot > 1:
            # more than one uniprot in baseset
            log_util.log_warning('Baseset with more than one Uniprot %s' % baseset.set_id)
            log_util.log_info("%s" % str(uniprot_dic))
        return (baseset,nr_uniprot)
            
        
                
        