"""
Structural entries (sentries) set.
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import log_util
import sentryres


class SentrySet:
    """
    Class for storing sets of structural entries (sentries).
    """

    def __init__(self, set_id = "SET", use_pdb_subdir = 'True', atom_sel = 'CA'):
        self.set_id = set_id
        self.pdb_dir = None
        self.sentry_lst = []    # list of sentry objects
        self.sentryid_dic = {}    # sentryid_dic[sentry_id] position in sentry_lst
        self.set_len = None
        self.atom_sel = atom_sel   # atoms selected, 'CA' for carbon alpha, 'ALL' for all atoms, 'GC' for Ca and geometric center of side chain
        self.use_pdb_subdir = use_pdb_subdir  # if subsirectory is needed to access PDB
        
        
    def delete_sentry(self, sentry_id):
        """
        Delete sentry_id from set
        """
        sentry_pos = self.sentryid_dic[sentry_id]
        del self.sentry_lst[sentry_pos]
        self._make_sentryid_dic()
        self.set_len = len(self.sentry_lst)
        
        
        
    def get_sentries(self, input_sentry_id_lst, pdb_dir):
        """
        Get sentries for input_sentry_id_lst.
        """
        self.sentry_id_lst = []
        self.pdb_dir = pdb_dir

        
        for sentry_id in input_sentry_id_lst:
            sentry = sentryres.Sentry(sentry_id, self.pdb_dir, self.atom_sel, self.use_pdb_subdir)            
            try:
                
                r_val = sentry.load_pdb()
                if r_val == 0:
                    log_util.log_error('Could not read entry %s' % sentry_id)
                    continue
                self.sentry_lst.append(sentry)
                self._make_sentryid_dic()
                self.set_len = len(self.sentry_lst)
            except:
                log_util.log_error('Could not read entry %s' % sentry_id)
                if self.sentryid_dic.has_key(sentry_id):
                    self.delete_sentry(sentry_id)
                continue



    def get_sentry(self, sentry_id):
        """
        From sentry_id get sentry object
        """
        s_counter = self.sentryid_dic[sentry_id]
        return self.sentry_lst[s_counter]
        


    def _make_sentryid_dic(self):
        """
        Make sentryid_dic.
        """
        self.sentryid_dic = {}
        s_counter = 0
        for sentry in self.sentry_lst:
            self.sentryid_dic[sentry.sentry_id] = s_counter
            s_counter += 1



    def write(self):
        """
        Debug.
        """
        print "Set: %s" % self.set_id
        for sentry in self.sentry_lst:
            sentry.write()


    def get_id_from_pos(self, spos):
        """
        Get sentry id from position in sentry_lst
        """
        return self.sentry_lst[spos].sentry_id

