"""
Structural entry class and residue class.
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import os.path
import math
import Numeric
import rpy

from Bio.PDB.PDBParser import PDBParser
import Bio.PDB
from Bio.PDB.Atom import Atom

import log_util


# Table of 3 letter codes for post translation modifications
# From ASTRAL SCOP 1.67, raf.html
aa_trans_dic = {
'ALA': 'A', 'VAL': 'V', 'PHE': 'F', 'PRO': 'P', 'MET': 'M', 'ILE': 'I', 'LEU': 'L', 'ASP': 'D', 'GLU': 'E', 'LYS': 'K',
'ARG': 'R', 'SER': 'S', 'THR': 'T', 'TYR': 'Y', 'HIS': 'H', 'CYS': 'C', 'ASN': 'N', 'GLN': 'Q', 'TRP': 'W', 'GLY': 'G',
'2AS': 'D', '3AH': 'H', '5HP': 'E', 'ACL': 'R', 'AIB': 'A', 'ALM': 'A', 'ALO': 'T', 'ALY': 'K', 'ARM': 'R', 'ASA': 'D',
'ASB': 'D', 'ASK': 'D', 'ASL': 'D', 'ASQ': 'D', 'AYA': 'A', 'BCS': 'C', 'BHD': 'D', 'BMT': 'T', 'BNN': 'A', 'BUC': 'C',
'BUG': 'L', 'C5C': 'C', 'C6C': 'C', 'CCS': 'C', 'CEA': 'C', 'CHG': 'A', 'CLE': 'L', 'CME': 'C', 'CSD': 'A', 'CSO': 'C',
'CSP': 'C', 'CSS': 'C', 'CSW': 'C', 'CXM': 'M', 'CY1': 'C', 'CY3': 'C', 'CYG': 'C', 'CYM': 'C', 'CYQ': 'C', 'DAH': 'F',
'DAL': 'A', 'DAR': 'R', 'DAS': 'D', 'DCY': 'C', 'DGL': 'E', 'DGN': 'Q', 'DHA': 'A', 'DHI': 'H', 'DIL': 'I', 'DIV': 'V',
'DLE': 'L', 'DLY': 'K', 'DNP': 'A', 'DPN': 'F', 'DPR': 'P', 'DSN': 'S', 'DSP': 'D', 'DTH': 'T', 'DTR': 'W', 'DTY': 'Y',
'DVA': 'V', 'EFC': 'C', 'FLA': 'A', 'FME': 'M', 'GGL': 'E', 'GLZ': 'G', 'GMA': 'E', 'GSC': 'G', 'HAC': 'A', 'HAR': 'R',
'HIC': 'H', 'HIP': 'H', 'HMR': 'R', 'HPQ': 'F', 'HTR': 'W', 'HYP': 'P', 'IIL': 'I', 'IYR': 'Y', 'KCX': 'K', 'LLP': 'K',
'LLY': 'K', 'LTR': 'W', 'LYM': 'K', 'LYZ': 'K', 'MAA': 'A', 'MEN': 'N', 'MHS': 'H', 'MIS': 'S', 'MLE': 'L', 'MPQ': 'G',
'MSA': 'G', 'MSE': 'M', 'MVA': 'V', 'NEM': 'H', 'NEP': 'H', 'NLE': 'L', 'NLN': 'L', 'NLP': 'L', 'NMC': 'G', 'OAS': 'S',
'OCS': 'C', 'OMT': 'M', 'PAQ': 'Y', 'PCA': 'E', 'PEC': 'C', 'PHI': 'F', 'PHL': 'F', 'PR3': 'C', 'PRR': 'A', 'PTR': 'Y',
'SAC': 'S', 'SAR': 'G', 'SCH': 'C', 'SCS': 'C', 'SCY': 'C', 'SEL': 'S', 'SEP': 'S', 'SET': 'S', 'SHC': 'C', 'SHR': 'K',
'SOC': 'C', 'STY': 'Y', 'SVA': 'S', 'TIH': 'A', 'TPL': 'W', 'TPO': 'T', 'TPQ': 'A', 'TRG': 'K', 'TRO': 'W', 'TYB': 'Y',
'TYQ': 'Y', 'TYS': 'Y', 'TYY': 'Y', 'AGM': 'R', 'GL3': 'G', 'SMC': 'C', 'ASX': 'B', 'CGU': 'E', 'CSX': 'C', 'GLX': 'Z'}


class Sentry:
    """
    Structural entry class (sentry)
    """

    def __init__(self, sentry_id, pdb_dir, atom_sel, use_pdb_subdir):
        self.sentry_id = sentry_id
        self.pdb_dir = pdb_dir
        self.atom_sel = atom_sel  # 'CA' for carbon alpha only, ALL all atoms, GC is for Ca and geometric center of side chains
        self.pdb_file = None
        self.length = None        # number of residues
        self.res_lst = []         # list of residues, in PDB order
        self.res_dic = {}         # res_dic['chain:label'] position in res_lst
        self.use_pdb_subdir = use_pdb_subdir   # True or False, True if astral files under subdirectories
        # need to add chain definitions
        

    def load_pdb(self):
        """
        Load structural object from pdb file.

        PDB file name for scop domain pdbstyle files is generated from sentry id (sid).
        Return 0 error
        return 1 sucess
        """
        self._pdbfile_for_scop()
        if os.path.exists(self.pdb_file) !=True:
            return 0
            
        p_obj = PDBParser(PERMISSIVE=1)
        str_obj = p_obj.get_structure(self.sentry_id, self.pdb_file)

        # use only model 0
        model = str_obj[0]
        for chain_obj in model.get_iterator():
            # for each chain
            chain_id = chain_obj.get_id()
            for res_obj in chain_obj.get_iterator():
                # for each residue
                het_field, res_nr, res_ins = res_obj.get_id()
                residue = Residue(chain_id, res_nr, res_ins, self.atom_sel)
                res_name = res_obj.get_resname()
                if residue.check_aa_type(res_name, het_field) == 0:
                    # skpip, not a residue
                    continue
                if residue.get_atoms(res_obj, self.atom_sel) == 0:
                    # get atom objects
                    continue
                self.res_lst.append(residue)
        # do not sort residues, keep PDB order
        # self.res_lst.sort()
        # make residue indexing
        self._make_res_dic()
        self.length = len(self.res_lst)
        return 1
        
        
    def ca_dist_mat(self):
        """
        Calculate and return the CA distance matrix
        if CA atoms not availale, then set to -1
        """
        dist_mat = Numeric.zeros([self.length,self.length], Numeric.Float) -1
        for i in range(self.length):
            for j in range(i+1):
                if self.res_lst[i].atom_dic.has_key('CA') and self.res_lst[j].atom_dic.has_key('CA'):
                    # calculate distance, the - operator calculates the distance of atom objects
                    # I know, they should have a function for that...
                    dist_mat[i,j] = self.res_lst[i].atom_dic['CA'] -  self.res_lst[j].atom_dic['CA']
                    dist_mat[j,i] = dist_mat[i,j]
                else:
                    dist_mat[j,i] = -1
                    dist_mat[j,i] = -1
        return dist_mat

        
        
    def gc_dist_mat(self):
        """
        Calculate and return the GC distance matrix
        if GC atoms not availale, then set to -1
        """
        dist_mat = Numeric.zeros([self.length,self.length], Numeric.Float) -1
        for i in range(self.length):
            for j in range(i+1):
                if self.res_lst[i].atom_dic.has_key('GC') and self.res_lst[j].atom_dic.has_key('GC'):
                    # calculate distance, the - operator calculates the distance of atom objects
                    # I know, they should have a function for that...
                    dist_mat[i,j] = self.res_lst[i].atom_dic['GC'] -  self.res_lst[j].atom_dic['GC']
                    dist_mat[j,i] = dist_mat[i,j]
                else:
                    dist_mat[j,i] = -1
                    dist_mat[j,i] = -1
        return dist_mat

    
    
    
    def set_sus(self,dpi_obj):
        """
        Set the sus for all atoms in sentry based on dpi object
        If failure return 0
        """
        if dpi_obj.DPIf != None:
            DPI = dpi_obj.DPIf
        elif dpi_obj.DPIa != None:
            DPI = dpi_obj.DPIa
        else:
            DPI = None
            return 0
        Bavg = dpi_obj.Bavg
        k = DPI/Bavg   # assume linear relation
        for res_obj in self.res_lst:
            atomid_lst = res_obj.atom_dic.keys()
            for atom_id in atomid_lst:
                atom_obj = res_obj.atom_dic[atom_id]
                bfactor = atom_obj.get_bfactor()
                sus = k * bfactor   # assume linear relation
                res_obj.atom_sus_dic[atom_id] = sus
        return 1
    
            
    def plot_sus(self, sus_v):
        f_name = self.sentry_id + '_boxp.png'
        rpy.r.png(f_name)
        rpy.r.boxplot(sus_v, names= self.sentry_id, ylim = [0.0,1.5])
        rpy.r.dev_off()

            
            
    def orig_set_sus(self,dpi_obj):
        """
        Set the sus for all atoms in sentry based on dpi object
        If failure return 0
        """
        if dpi_obj.DPIf != None:
            DPI = dpi_obj.DPIf
        elif dpi_obj.DPIa != None:
            DPI = dpi_obj.DPIa
        else:
            DPI = None
            return 0
        Bavg = dpi_obj.Bavg
        k = DPI/Bavg   # assume linear relation
        for res_obj in self.res_lst:
            atomid_lst = res_obj.atom_dic.keys()
            for atom_id in atomid_lst:
                atom_obj = res_obj.atom_dic[atom_id]
                bfactor = atom_obj.get_bfactor()
                sus = k * bfactor   # assume linear relation
                res_obj.atom_sus_dic[atom_id] = sus
        return 1


    def dist_sus_mat(self):
        """
        Calculate and return the CA distance sus distance matrix 
        if no sus then -1
        """
        dist_sus_mat = Numeric.zeros([self.length,self.length], Numeric.Float) -1
        for i in range(self.length):
            for j in range(i+1):
                if self.res_lst[i].atom_sus_dic.has_key('CA') and self.res_lst[j].atom_sus_dic.has_key('CA'):
                    # calculate distance sus for CA, neglect covariance
                    sus_i = self.res_lst[i].atom_sus_dic['CA']
                    sus_j = self.res_lst[j].atom_sus_dic['CA']
                    dist_sus_mat[i,j] = math.sqrt((sus_i * sus_i) + (sus_j * sus_j))
                    dist_sus_mat[j,i] = dist_sus_mat[i,j]
        return dist_sus_mat

        
        
    def dist_gc_sus_mat(self):
        """
        Calculate and return the GC distance sus distance matrix 
        if no sus then -1
        """
        dist_sus_mat = Numeric.zeros([self.length,self.length], Numeric.Float) -1
        for i in range(self.length):
            for j in range(i+1):
                if self.res_lst[i].atom_sus_dic.has_key('GC') and self.res_lst[j].atom_sus_dic.has_key('GC'):
                    # calculate distance sus for GC, neglect covariance
                    sus_i = self.res_lst[i].atom_sus_dic['GC']
                    sus_j = self.res_lst[j].atom_sus_dic['GC']
                    dist_sus_mat[i,j] = math.sqrt((sus_i * sus_i) + (sus_j * sus_j))
                    dist_sus_mat[j,i] = dist_sus_mat[i,j]
        return dist_sus_mat
        
        

    def get_CA_sus_v(self):
        """
        Return a Numeric array with all sus for CA atoms        
        """
        sus_lst = []
        for i in range(self.length):
            if self.res_lst[i].atom_sus_dic.has_key('CA'):
               sus = self.res_lst[i].atom_sus_dic['CA']
               sus_lst.append(sus)
        sus_v = Numeric.array(sus_lst, Numeric.Float)
        return sus_v         
                
            
            
    def get_atom_lst(self, sentrypos_lst):
        """
        Get Ca atom object list from sentrypos list
        """
        atom_lst = []
        for sentrypos in sentrypos_lst:
            if sentrypos == None:
                atom_lst.append(None)
            else:
                residue = self.res_lst[sentrypos]
                atom_lst.append(residue.atom_dic['CA'])
        return atom_lst
        
        
        
    def get_seq_str(self):
        """
        Return the sequence string
        """
        seq_str = ""
        for i in range(self.length):
            seq_str = seq_str + self.res_lst[i].code
        return seq_str
            
        
    
    def _write_sus(self):
        """
        Write the sus for each atom
        debug function
        """
        log_util.log_info('Sentry: %s' % self.sentry_id)
        for res_obj in self.res_lst:
            atomid_lst = res_obj.atom_dic.keys()
            for atom_id in atomid_lst:
                atom_obj = res_obj.atom_dic[atom_id]
                atom_serialnr = atom_obj.get_serial_number()
                atom_bfactor = atom_obj.get_bfactor()
                atom_sus = res_obj.atom_sus_dic[atom_id]
                log_util.log_info('%d %f %f' % (atom_serialnr, atom_bfactor, atom_sus))


                
    def _pdbfile_for_scop(self):
        """
        Get pdb file name from sentry id corressponding to scop sid.

        Return file with full path.
        """
        # suffix
        pdb_filename = self.sentry_id + '.ent'
        if self.use_pdb_subdir == True:
            # pdb subdirectory
            pdb_subdir = self.sentry_id[2:4]
            # full path and file name
            self.pdb_file = os.path.join(self.pdb_dir, pdb_subdir, pdb_filename) 
        else:
            self.pdb_file = os.path.join(self.pdb_dir, pdb_filename) 


    def _make_res_dic(self):
        """
        Make the res_dic.
        """
        res_counter = 0
        for residue in self.res_lst:
            res_key = self._get_res_key(residue)
            self.res_dic[res_key] = res_counter
            res_counter += 1



    def _get_res_key(self, residue):
        """
        Get the residue key for res_dic.
        """
        key = residue.chain_id + ':' + residue.label
        return key



    def write(self):
        """
        Debug
        """
        print "Sentry: %s" % self.sentry_id
        for residue in self.res_lst:
            res_key = self._get_res_key(residue)
            res_pos = self.res_dic[res_key]
            print "%d: >%s<" % (res_pos, res_key),
            residue.write()


        
class Residue:
    """
    Residue class.
    """

    def __init__(self, chain_id, res_nr, res_ins, atom_sel ):
        self.chain_id = chain_id
        self.nr = res_nr
        self.ins = res_ins
        self.label = str(res_nr) + res_ins
        self.name = None           # 3 letter code residue name from pdb
        self.code = None           # amino acid one letter code, from self.name using conversion table in aa_trans_dic
        self.hetero_flag = None    # A for amino acid, H for hetero atom (modified amino acid)
        self.atom_dic = {}         # dictionary of biopython atom objects, key is atom id
        self.atom_sel = atom_sel       # atoms selected, 'CA' for carbon alpha, 'GC' Ca and geometric center of side chains, 'ALL' all atoms are selected
        self.atom_sus_dic = {}      # dictionary of sus values for each atom

    def check_aa_type(self, res_name, het_field):
        """Check type of amino acid.

        If it is heteroatom, check if residue names corresponds to a chemically modified aminoacid,
        and set residue name and corresponding amino acid one letter code according to the
        conversion table aa_trans_dic.                
        Return 1 if amino acid, 0 if it is no amino acid         
        """
        if het_field == ' ':
            if aa_trans_dic.has_key(res_name):
                self.name = res_name
                self.code = aa_trans_dic[self.name]
                self.hetero_flag = 'A'
                return 1
            else:
                # could also be a nucleic acid,
                # need to check for nuclaic acids
                log_util.log_warning('Non standard amino acid code %s, chain %s residue %s' %
                                     (self.name, self.chain_id, self.label))
        elif het_field[:2] == 'H_':
            if aa_trans_dic.has_key(res_name):
                self.name = res_name
                self.code = aa_trans_dic[self.name]
                self.hetero_flag = 'H'
                if self.name != het_field[2:]:
                    log_util.log_warning('Residue name %s does not match het_field %s, chain %s residue %s' %
                                         (self.name, het_field, self.chain_id, self.label))
                return 1        
        return 0


    def get_atoms(self, res_obj, atom_sel):
        """
        Get atoms for residue.

        Inputs: biopython residue object res_obj, and defintion of atom selection (ALL or CA) 
        """
        self.atom_sel = atom_sel
        if self.atom_sel == 'ALL':
            # all atoms
            for atom_obj in res_obj.get_iterator():
                atom_id = atom_obj.get_id()
                self.atom_dic[atom_id] = atom_obj
            return 1        
        elif self.atom_sel == 'GC':
            # geometric center of side chains and C alpha
            if res_obj.has_id('CA')!= 1:
                # skip if no CA
                log_util.log_warning('No CA atom at chain %s residue %s %s' % (self.chain_id, self.label, self.name))
                return 0
            sum_coord = Numeric.zeros(3, Numeric.Float)
            sum_bfactor = 0.00
            c_counter = 0
            for atom_obj in res_obj.get_iterator():
                atom_id = atom_obj.get_id()
                if atom_id == 'CA':
                    # get CA
                    self.atom_dic['CA'] = atom_obj
                elif atom_id == 'N' or atom_id == 'C' or atom_id == 'O':
                    # skip backbone
                    continue
                elif atom_id.find('H') >=0:
                    # ignore H, will also ignore HG, not easy to fix
                    continue
                else:
                    sum_coord = sum_coord + atom_obj.get_coord()
                    sum_bfactor = sum_bfactor + atom_obj.get_bfactor()
                    c_counter +=1
            if c_counter > 0:
                avg_coord = sum_coord / float(c_counter)   
                avg_bfactor =  sum_bfactor / float(c_counter)
                occupancy = 1.0
                altloc = ' '
                fullname = 'GC  '
                atom_obj = Atom('GC', avg_coord, avg_bfactor, occupancy, altloc, fullname, None)
                self.atom_dic['GC'] = atom_obj
            else:
                # CA but no side chain, GC is CA
                self.atom_dic['GC'] = self.atom_dic['CA']
            return 1        
        else:
            # assume CA only
            if res_obj.has_id('CA'):
                self.atom_dic['CA'] = res_obj['CA']
                return 1
            else:
                log_util.log_warning('No CA atom at chain %s residue %s %s' % (self.chain_id, self.label, self.name))
        return 0


            
    def write(self):
        """Debug."""
        print "%s %d >%s< %s %s %s, Coordinates: %f %f %f" % \
              ( self.chain_id, self.nr, self.ins, self.label, self.name, self.hetero_flag,
                self.atom_dic['CA'].get_coord()[0], self.atom_dic['CA'].get_coord()[1], self.atom_dic['CA'].get_coord()[2])

        

    def __cmp__(self, other):
        """
        Compare residues for sorting.

        First compare by chain id, then residue number, then insertion code.
        """
        if self.chain_id < other.chain_id:
            return -1
        elif self.chain_id > other.chain_id:
            return 1
        elif self.nr < other.nr:
            return -1
        elif self.nr > other.nr:
            return 1
        elif self.ins < other.ins:
            return -1
        elif self.ins > other.ins:
            return 1
        return 0
