"""
Alignment data for set of structural entries.

Use StMap object as reference alignment of each structure to Uniprot
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import os

import log_util
import stmap_resmap
import sifts_map
import pdbsws_map
import Numeric


class SetAlig:
    """
    Alignment of a set.
    """

    def __init__(self):
        self.sentryset = None
        self.stmap_dir = None
        self.uniprot_dir = None
        
        # alignment data
        self.refpos_dic = {}
        self.length = None
        
        # accessory data
        self.uniprot_refpos = {}
        self.sentry_refpos = {}
        self.ins_refpos = {}
        
        # data for alignment generation
        self.uniprot_map = {}
        self.ins_map = {}
        # temporary data
        self.sentry = None
        self.sentrypos_counter = None
        self.chain_uniprot = {}
        self.last_uniprotpos = None
        self.uniprot_seq = None

        
        
    def _clear_alig(self):
        self.sentryset = None
        self.stmap_dir = None
        self.uniprot_dir = None
        
        # alignment data
        self.refpos_dic = {}
        self.length = None

        # accessory data       
        self.uniprot_refpos = {}
        self.sentry_refpos = {}
        self.ins_refpos = {}
        
        # data for alignment generation
        self.uniprot_map = {}
        self.ins_map = {}
        # temporary data
        self.sentry = None
        self.sentrypos_counter = None
        self.chain_uniprot = {}
        self.last_uniprotpos = None
        self.uniprot_seq = None
    

        
    def get_alignment(self, sentryset, stmap_dir, uniprot_dir):
        """
        Get alignment from uniport to pdb mapping
        """
        self.sentryset = sentryset
        self.stmap_dir = stmap_dir
        self.uniprot_dir = uniprot_dir
        # get intermediate mappings
        for sentry in sentryset.sentry_lst:
            self._get_int_map(sentry)
        # get the refpos_dic
        self._get_refpos()
        self.length = len(self.refpos_dic)
        log_util.log_info("Aligned set %s" % sentryset.set_id)
    
    
    
    def write_alignment(self, alig_file):
        """
        Debug, write the contents of refpos_dic
        """
        f_obj = file(alig_file, 'w')
        
        refpos_lst = self.refpos_dic.keys()
        refpos_lst.sort()
        sentryset_id = self.sentryset.set_id
        f_obj.write("Alignment for set:%s\n" % sentryset_id)        
        #str_out = "%s %s% %s     %s      |%4s    %4s| %10s %6s     %2s" %\
        f_obj.write("refpos uniprot_id uniprotresnr rescode      ins_nr inspos     sentry_id chain_id sreslabel mutrescode \n")

        for refpos in refpos_lst:
            f_obj.write('\n')
            refmap = self.refpos_dic[refpos]
            uniprot_id = refmap.uniprot_id
            uniprotpos = refmap.uniprotpos
            rescode = refmap.uniprot_rescode
            ins_nr  = refmap.ins_nr
            if ins_nr == None:
                ins_nr = '_'                
            inspos     = refmap.inspos
            if inspos == None:
                inspos = '_'
            sentry_id_lst = refmap.sentry_dic.keys()
            sentry_id_lst.sort()
            for sentry_id in sentry_id_lst:
                sentrypos = refmap.sentry_dic[sentry_id]
                sentry = self.sentryset.get_sentry(sentry_id)
                chain_id = sentry.res_lst[sentrypos].chain_id
                if chain_id == " ":
                    chain_id = "_"
                sreslabel = sentry.res_lst[sentrypos].label
                if refmap.mutsentry_dic.has_key(sentry_id):
                    mutrescode = refmap.mutsentry_dic[sentry_id]
                else:
                    mutrescode   = '_'
                str_out = "%6d %10s %6d %2s   %4s %4s %10s %2s %6s %2s\n" %\
                          (int(refpos), str(uniprot_id), (int(uniprotpos)+1), str(rescode),
                           str(ins_nr), str(inspos), str(sentry_id), str(chain_id),str(sreslabel), str(mutrescode))
                f_obj.write(str_out)
            if sentry_id_lst == []:
                sentry_id = '_'
                chain_id = '_'
                sreslabel = '_'
                mutrescode   = '_'
                str_out = "%6d %10s %6d %2s   %4s %4s %10s %2s %6s %2s" %\
                          (int(refpos), str(uniprot_id), int(uniprotpos+1), str(rescode),
                           str(ins_nr), str(inspos), str(sentry_id), str(chain_id), str(sreslabel), str(mutrescode))
                f_obj.write(str_out)
        f_obj.close()
        
        

    def refpos_to_uniprot_resnr(self, refpos):
        uniprotpos = self.refpos_dic[refpos].uniprotpos
        uniprot_resnr = uniprotpos + 1
        return uniprot_resnr
        
        

    def alig_mat(self, mat_dic):
        """
        Align the matrices, distance or distance sus
    
        Set alig_array. If gap in alig, if missing sentry, value set to -1
        """
        # allocate array, initialize to -1
        alig_array = Numeric.zeros([self.length, self.length, self.sentryset.set_len], Numeric.Float) -1
        for k in range(self.sentryset.set_len):
            sentry =  self.sentryset.sentry_lst[k]
            for i in range(self.length):
                for j in range(i+1):
                    if (self.refpos_dic[i].sentry_dic.has_key(sentry.sentry_id) and
                        self.refpos_dic[j].sentry_dic.has_key(sentry.sentry_id) and 
                        mat_dic.has_key(sentry.sentry_id)):
                            i_sentrypos = self.refpos_dic[i].sentry_dic[sentry.sentry_id]
                            j_sentrypos = self.refpos_dic[j].sentry_dic[sentry.sentry_id]
                            alig_value = mat_dic[sentry.sentry_id][i_sentrypos,j_sentrypos]
                    else:
                        alig_value = -1
                    alig_array[i,j,k] = alig_value
                    alig_array[j,i,k] = alig_value
        return alig_array


    
    def get_aligment_from_msa(self, sentryset, msa_file):
        """
        Get alignment from msa
        """
        self._clear_alig()
        self.sentryset = sentryset
        msa_lst = self._get_msa(msa_file)
        self._get_uniprot_from_msa(msa_lst)
        self._get_sentries_from_msa(msa_lst)
    
            
    
    def write_sequences_fasta(self, sentryset, uniprot_dir, uniprot_id_lst, fasta_file):
        """
        Write all the uniprot and all sentry sequences into fasta file
        """
        line_length = 60
        self.sentryset = sentryset
        self.uniprot_dir = uniprot_dir
        
        # uniprot sequences
        out_str = ""
        for uniprot_id in uniprot_id_lst:
            self._get_uniprot_seq(uniprot_id)
            out_seq_str = self._format_sequence_str(uniprot_id, self.uniprot_seq)
            out_str = out_str + out_seq_str
            
        # sentries
        for sentry in self.sentryset.sentry_lst:
            sentry_id = sentry.sentry_id
            sentry_seq = sentry.get_seq_str()
            out_seq_str = self._format_sequence_str(sentry_id, sentry_seq)
            out_str = out_str + out_seq_str
        
        f_obj = file(fasta_file, 'w')
        f_obj.write(out_str)
        f_obj.close()
        
        
        
    def get_alignment_from_file(self, sentryset, alig_filename):
        """
        Get alignment from file generated with write_alignment
        """
        self._clear_alig()
        self.sentryset = sentryset
        refpos_dic = {}
        f_obj = file(alig_filename, 'r')
        lines_lst = f_obj.readlines()
        f_obj.close()
        
        for line_str in lines_lst:
            line_str = line_str.strip()
            if line_str == "":
                continue
            line_lst = line_str.split()
            if line_lst[0] == 'Alignment' or line_lst[0] == 'refpos':
                continue
            refpos = int(line_lst[0])
            uniprot_id = line_lst[1]
            uniprotpos = int(line_lst[2]) -1
            rescode = line_lst[3]
            if line_lst[4] != "_":
                ins_nr = int(line_lst[4])
            else:
                ins_nr = line_lst[4]
            if line_lst[5] != "_":    
                inspos = int(line_lst[5])
            else:
                inspos = line_lst[5]
            sentry_id = line_lst[6]
            chain_id = line_lst[7]
            sreslabel = line_lst[8]
            mutrescode = line_lst[9]
            
            
            if refpos_dic.has_key(refpos) == False:
                refmap_obj = RefMap(uniprot_id, uniprotpos, rescode, ins_nr, inspos)
                refpos_dic[refpos] = refmap_obj                
                self._add_uniprot_refpos(uniprot_id, uniprotpos, refpos)
            if sentry_id != '_':
                if chain_id == "_":
                    chain_id = " "
                if sreslabel[-1].isalpha() == False:
                    sreslabel = sreslabel + " "
                    res_key = chain_id + ':' + sreslabel                
                sentry = self.sentryset.get_sentry(sentry_id)
                sentrypos = sentry.res_dic[res_key]            
                refpos_dic[refpos].sentry_dic[sentry_id] = sentrypos
                if mutrescode != "_":
                    refpos_dic[refpos].mutsentry_dic[sentry_id] = mutrescode
                self._add_sentry_refpos(sentry_id, sentrypos, refpos)
            if inspos != "_":
                self._add_ins_refpos(uniprot_id, uniprotpos, ins_nr, inspos, refpos)
        self.refpos_dic = refpos_dic
        self.length = len(self.refpos_dic)
    

    def get_sentrypos_from_refpos(self, refpos, sentry_id):
        if self.refpos_dic.has_key(refpos):
            if self.refpos_dic[refpos].sentry_dic.has_key(sentry_id):
                sentrypos = self.refpos_dic[refpos].sentry_dic[sentry_id]
                return sentrypos
        return None
                
        
        
    def _get_msa(self, msa_file):
        f_obj = file(msa_file, 'r')
        lines_lst = f_obj.readlines()
        f_obj.close()
        
        msa_lst = []
        
        seq_label = None
        seq_str = None
        for line_str in lines_lst:
            line_str = line_str.strip()
            if line_str[0] == '>':
                # new sequence
                if seq_label != None and seq_str != None:
                    # store last sequence if exists
                    msa_lst.append([seq_label, seq_str])      
                seq_label = line_str[1:]
                seq_str = ""
            else:
                seq_str = seq_str + line_str
        if seq_label != None and seq_str != None:
            msa_lst.append([seq_label, seq_str])      
        
        return msa_lst            
        
        
            
    def _get_uniprot_from_msa(self, msa_lst):
        uniprot_id = msa_lst[0][0]
        uniprot_seq = msa_lst[0][1]
        self.length = len(uniprot_seq)
        uniprotpos = -1
        inspos = None 
        for refpos in range(self.length):
            # for each position
            res_str = uniprot_seq[refpos]
            if res_str == "-":
                rescode = "_"
                if inspos == None:
                   inspos = 0
                else:
                    inspos +=1
                if uniprotpos == -1:
                    uniprotpos = 0
                ins_nr = 0
                refmap_obj = RefMap(uniprot_id, uniprotpos, rescode, ins_nr, inspos)
                self._add_ins_refpos(uniprot_id, uniprotpos, ins_nr, inspos, refpos)
            else:
                rescode = res_str
                uniprotpos +=1
                refmap_obj = RefMap(uniprot_id, uniprotpos, rescode, None, None)
                inspos == None
            self.refpos_dic[refpos] = refmap_obj
            self._add_uniprot_refpos(uniprot_id, uniprotpos, refpos)
            
            
            
    def _get_sentries_from_msa(self, msa_lst):         
        sentries_msa_lst = msa_lst[1:]        
        for sentry_alig in sentries_msa_lst:
            sentry_id = sentry_alig[0]
            sentry_seq = sentry_alig[1]
            sentrypos = 0
            for refpos in range(self.length):
                res_str = sentry_seq[refpos]
                if res_str != "-":
                    refmap_obj = self.refpos_dic[refpos]
                    refmap_obj.sentry_dic[sentry_id] = sentrypos
                    self._add_sentry_refpos(sentry_id, sentrypos, refpos)
                    sentrypos +=1
                    if res_str != refmap_obj.uniprot_rescode:
                        refmap_obj.mutsentry_dic[sentry_id] = res_str
            sentry = self.sentryset.get_sentry(sentry_id)
            if sentrypos != sentry.length:
                log_util.log_warning("Not all residues from %s in alignment" % sentry_id)
                    
            
                        
            
    def _format_sequence_str(self, seq_label, seq_str):
        """
        Format the sequence string with line return every line_length
        """
        line_length = 60
        
        out_seq_str = ""
        out_seq_str = ">" + seq_label + '\n'
        seq_len = len(seq_str)
        nr_full_lines = int(seq_len/line_length)
        remain_length = seq_len % line_length
        for line_c in range(nr_full_lines):
            start_pos = line_c * line_length
            end_pos = start_pos + line_length
            out_seq_str = out_seq_str + seq_str[start_pos:end_pos] + '\n'
        start_pos = nr_full_lines*line_length
        out_seq_str = out_seq_str + seq_str[start_pos:] + '\n'
        return out_seq_str
                

    def _get_int_map(self, sentry):
        """
        Get intermediate mappings.

        ins_map: insertions
        uniprot_map: uniprotpos to sentrypos
        """

        self.sentry = sentry
        self.sentrypos_counter = 0
        self.chain_uniprot = {}
        self.last_uniprotpos = -1  # -1-> N terminus insertion, starts at 0, is an index starting at 0, to get sequence residue number add +1
        
        # get stmap
        stmap = self._load_stmap()

        # get ins_map and uniprot_map
        # loop over each residue
        while self.sentrypos_counter <  self.sentry.length:
            residue = self.sentry.res_lst[self.sentrypos_counter]
            resmap = stmap.get_resmap(residue.label, residue.chain_id)
            if resmap == None:
                # No resmap for the residue position, insertion in uniprot sequence
                self._get_insertion(stmap)
            else:
                # get uniprot_map
                self._get_uniprot_map(stmap)



    def _load_stmap(self):
        """
        Load stmap object
        Try first stamp for sunid (provided by SIFTS)
        if does not exist, then try the stmap for PDB (provided by PDBSWS)
        If fials returns None
        """

        map_file_name = sifts_map.get_map_file_name(self.sentry.sentry_id)
        map_file = os.path.join(self.stmap_dir, map_file_name)
        if os.path.exists(map_file):
            stmap = stmap_resmap.load_map_data(map_file)
            return stmap
        map_file_name = pdbsws_map.get_map_file_name(self.sentry.sentry_id)
        map_file = os.path.join(self.stmap_dir, map_file_name)
        if os.path.exists(map_file):
            stmap = stmap_resmap.load_map_data(map_file)
            return stmap
        return None



    def _get_insertion(self, stmap):
        # first residue in insertion
        residue = self.sentry.res_lst[self.sentrypos_counter]
        start_sentrypos = self.sentrypos_counter
        ins_length = 1
        ins_aa_seq = residue.code
        ins_chain = residue.chain_id

        # get next residues
        self.sentrypos_counter += 1
        while self.sentrypos_counter <  self.sentry.length:
            residue = self.sentry.res_lst[self.sentrypos_counter]
            resmap = stmap.get_resmap(residue.label, residue.chain_id)
            if residue.chain_id == ins_chain:
                # same chain
                if resmap == None:
                    # continue insertion: same chain and no resmap
                    ins_length += 1
                    ins_aa_seq = ins_aa_seq + residue.code
                    self.sentrypos_counter += 1
                else:
                    # finish insertion: same chain and there is resmap
                    # get uniprot code for this chain
                    uniprot_id = resmap.up_id
                    self._map_chain_uniprot(residue.chain_id, uniprot_id)
                    # if N terminus insertion, check wheter next mapped residue is after uniprotpos 0, if so place insertion in correct place
                    if self.last_uniprotpos == -1:
                        # get position of new mapping
                        new_uniprotpos = resmap.up_resnr -1
                        if new_uniprotpos != 0:
                            # if new mapping not at the N terminus, then set the last insertion before the new postion
                            self.last_uniprotpos = new_uniprotpos -1
                    break
            else:
                # new chain
                break
            
        # get uniprot code for the insertion
        if self.chain_uniprot.has_key(ins_chain):
            uniprot_id = self.chain_uniprot[ins_chain]
        else:
            log_util.log_error('Can not find uniprot_id  for chain: %s in sentry: %s' % (ins_chain, self.sentry.sentry_id))

        # add to ins_map
        ins_interval = (start_sentrypos, start_sentrypos + ins_length -1)
        new_insmap_obj = InsMap(ins_length, ins_aa_seq, self.sentry.sentry_id, ins_interval)
        
        if self.ins_map.has_key(uniprot_id) == False:
            # first insertion for this uniprot sequence
            self.ins_map[uniprot_id] = {}
        if self.ins_map[uniprot_id].has_key(self.last_uniprotpos) == False:
            # new insertion at this point
            self.ins_map[uniprot_id][self.last_uniprotpos] = [new_insmap_obj]
        else:
            # already insertions at this point, compare them
            ins_match = 0
            for insmap_obj in self.ins_map[uniprot_id][self.last_uniprotpos]:
                if insmap_obj.compare(ins_aa_seq):
                    # same insertion
                    insmap_obj.add(self.sentry.sentry_id, ins_interval)
                    ins_match = 1
                    break
            if ins_match == 0:
                # new insertion
                self.ins_map[uniprot_id][self.last_uniprotpos].append(new_insmap_obj)

                    


    def _get_uniprot_map(self, stmap):
        residue = self.sentry.res_lst[self.sentrypos_counter]
        resmap = stmap.get_resmap(residue.label, residue.chain_id)
        uniprot_id = resmap.up_id
        uniprotpos = resmap.up_resnr - 1 # uniprotpos starts at 0, up_resnr starts at 1, subtract 1
        residue_chain = residue.chain_id
        
        # add new uniprot to chain mapping
        self._map_chain_uniprot(residue_chain, uniprot_id)
        if self.uniprot_map.has_key(uniprot_id) == False:
            self.uniprot_map[uniprot_id] = {}
        if self.uniprot_map[uniprot_id].has_key(uniprotpos) == False:
            self.uniprot_map[uniprot_id][uniprotpos] = {}
        if self.uniprot_map[uniprot_id][uniprotpos].has_key(self.sentry.sentry_id) == False:
            self.uniprot_map[uniprot_id][uniprotpos][self.sentry.sentry_id] = self.sentrypos_counter
        else:
            log_util.log_error('Double mapping, uniprot_id:%s uniprotpos:%d sentry_id:%s sentrypos_counter:%d' %
                                       (uniprot_id, uniprotpos, self.sentry.sentry_id, self.sentrypos_counter))
        self.last_uniprotpos = uniprotpos
        self.sentrypos_counter += 1
        # debug
        # print (self.sentry.sentry_id, uniprot_id, uniprotpos +1, resmap.up_resname , residue.label, residue.chain_id, residue.code)



    def _map_chain_uniprot(self, residue_chain, uniprot_id):
        if self.chain_uniprot.has_key(residue_chain):
            # chain already mapped to uniprot, check consistency
            if self.chain_uniprot[residue_chain] != uniprot_id:
                log_util.log_error('Inconsistency in mapping Uniprot:%s  to structure: %s chain: %s' %
                                   (uniprot_id, residue_chain, self.sentry.sentry_id))
        else:
            # map new chain to uniprot
            self.chain_uniprot[residue_chain] = uniprot_id



    def _get_uniprotpos_lst(self, uniprot_id):
        # get list of uniprotpos, from min value of uniprot_map and ins_map, and max of uniprot_map
        uniprotpos_lst = self.uniprot_map[uniprot_id].keys()
        if  self.ins_map.has_key(uniprot_id):
            # to get start of insertion, add min of insertion postions to uniprotpos_lst, then get overall minimum
            inspos_lst = self.ins_map[uniprot_id].keys()
            min_inspos_lst = min(inspos_lst)
            uniprotpos_lst.append(min_inspos_lst)
        uniprotpos_min = min(uniprotpos_lst)
        uniprotpos_max = max(uniprotpos_lst)
        uniprotpos_lst =  range(uniprotpos_min,uniprotpos_max+1)
        return uniprotpos_lst


    def _get_refpos(self):
        refpos = 0
        
        # get  list of uniprot entries
        # use sorted list
        uniprot_lst = self.uniprot_map.keys()
        uniprot_lst.sort()
        for uniprot_id in uniprot_lst:
            # for each uniprot sequence
            uniprotpos_lst = self._get_uniprotpos_lst(uniprot_id)
            # get uniprotpos_lst
            self._get_uniprot_seq(uniprot_id)
            # get uniprot sequence
            for uniprotpos in uniprotpos_lst:                
                # for each uniprotpos
                uniprot_rescode = self.uniprot_seq[uniprotpos]
                # get uniprot residue code
                orig_refpos = refpos         # mark original refpos
                if self.uniprot_map[uniprot_id].has_key(uniprotpos):
                    # get mapping to uniprot
                    refmap_obj = RefMap(uniprot_id, uniprotpos, uniprot_rescode, None, None)
                    sentry_id_lst = self.uniprot_map[uniprot_id][uniprotpos].keys()                    
                    for sentry_id in sentry_id_lst:
                        # loop over all sentries
                        sentrypos = self.uniprot_map[uniprot_id][uniprotpos][sentry_id]
                        sentry_rescode = self.sentryset.get_sentry(sentry_id).res_lst[sentrypos].code
                        refmap_obj.sentry_dic[sentry_id] = sentrypos
                        if sentry_rescode != uniprot_rescode:
                            # check for mutant sentries
                            refmap_obj.mutsentry_dic[sentry_id] = sentry_rescode
                        self._add_sentry_refpos(sentry_id, sentrypos, refpos)
                    self.refpos_dic[refpos] = refmap_obj   # set refpos_obj
                    self._add_uniprot_refpos(uniprot_id, uniprotpos, refpos)
                    refpos += 1           # increment refpos                    
                if self._query_ins_map(uniprot_id, uniprotpos):
                    # if insertion in this uniprot at this position
                    ins_counter = 0
                    for  insmap_obj in self.ins_map[uniprot_id][uniprotpos]:
                        # loop over all insertions
                        for inspos in range(insmap_obj.length):
                            # loop over each insertion position
                            refmap_obj = RefMap(uniprot_id, uniprotpos, '_', ins_counter, inspos)
                            for sentry_id in insmap_obj.sentry_dic.keys():
                                # get each sentry_id
                                refmap_obj.sentry_dic[sentry_id] = insmap_obj.sentry_dic[sentry_id][0]
                                refmap_obj.mutsentry_dic[sentry_id] = insmap_obj.aa_seq[inspos]
                            self.refpos_dic[refpos] = refmap_obj  # set refpos_obj
                            self._add_ins_refpos(uniprot_id, uniprotpos, ins_counter, inspos, refpos)
                            refpos += 1  # increment refpos    
                        ins_counter += 1  # increment insertion counter
                
                if refpos == orig_refpos:
                    # in refpos not incremented, then missing map to uniprotpos, 
                    refmap_obj = RefMap(uniprot_id, uniprotpos, uniprot_rescode, None, None)
                    self.refpos_dic[refpos] = refmap_obj  # set refpos_obj
                    self._add_uniprot_refpos(uniprot_id, uniprotpos, refpos)
                    refpos += 1 # increment refpos
                


    def _add_uniprot_refpos(self, uniprot_id, uniprotpos, refpos):
        """
        Add element to uniprot_refpos
        """
        if self.uniprot_refpos.has_key(uniprot_id) == False:
            self.uniprot_refpos[uniprot_id] = {}
        if self.uniprot_refpos[uniprot_id].has_key(uniprotpos) == False:
            self.uniprot_refpos[uniprot_id][uniprotpos] = refpos
        # else:
        #    log_util.log_error('Double mapping of uniprot_id:%s uniprotpos:%d refpos:%d' % (uniprot_id, uniprotpos, refpos))

            

    def _add_sentry_refpos(self, sentry_id, sentrypos, refpos):
        """
        Add element to sentry_refpos
        """
        if self.sentry_refpos.has_key(sentry_id) == False:
            self.sentry_refpos[sentry_id] = {}
        if self.sentry_refpos[sentry_id].has_key(sentrypos) == False:
            self.sentry_refpos[sentry_id][sentrypos] = refpos
        else:
            log_util.log_error('Double mapping of sentry:%s sentrypos:%d refpos:%d' % (sentry_id, sentrypos, refpos))



    def _add_ins_refpos(self, uniprot_id, uniprotpos, ins_nr, inspos, refpos):
        """
        Add element to ins_refpos
        """
        if self.ins_refpos.has_key(uniprot_id) == False:
            self.ins_refpos[uniprot_id] = {}
        if self.ins_refpos[uniprot_id].has_key(uniprotpos) == False:
            self.ins_refpos[uniprot_id][uniprotpos] = {}
        if self.ins_refpos[uniprot_id][uniprotpos].has_key(ins_nr) == False:
            self.ins_refpos[uniprot_id][uniprotpos][ins_nr] = {}
        if self.ins_refpos[uniprot_id][uniprotpos][ins_nr].has_key(inspos) == False:
            self.ins_refpos[uniprot_id][uniprotpos][ins_nr][inspos] = refpos
        else:
            log_util.log_error('Double mapping of insertion, uniprot_id:%s uniprotpos:%d ins_nr:%d inspos:%d refpos:%d' %
                               (uniprot_id, uniprotpos, ins_nr, inspos,  refpos))

        
    def _get_uniprot_seq(self, uniprot_id):
        # get sequence from sigle file, probably latter I will get it from db file
        # get filename
        uniprot_filename = uniprot_id + '.fasta'
        uniprot_file = os.path.join(self.uniprot_dir, uniprot_filename)
        file_obj = file(uniprot_file, 'r')
        seq_str = ''
        lines = file_obj.readlines()
        for line in lines:
            line = line.strip()
            if line[0] != '>' and line != '':
                seq_str = seq_str + line
        file_obj.close()
        self.uniprot_seq = seq_str


    def _query_ins_map(self, uniprot_id, uniprotpos):
        if self.ins_map.has_key(uniprot_id):
            if self.ins_map[uniprot_id].has_key(uniprotpos):
                return True
        return False
        

    def write_fasta_alig(self, filename):
        """
        Write the alignment as fasta format.

        Under development, need to support multiple uniprots
        """        
        refpos_lst = self.refpos_dic.keys()
        max_refpos = max(refpos_lst)
        sentryset_id = self.sentryset.set_id
        last_uniprot_id = ""
        # uniprot sequence
        out_str = ""
        out_line = ""
        for refpos in range(max_refpos +1):
            refmap = self.refpos_dic[refpos]
            uniprot_id = refmap.uniprot_id
            if last_uniprot_id == "":
                out_str = ">" + uniprot_id + '\n'
                last_uniprot_id = uniprot_id
            else:                
                if uniprot_id != last_uniprot_id:
                    log_util.log_error('More than one uniprot in one sentry set')
                    break
            rescode = refmap.uniprot_rescode
            if rescode == '_':
                rescode = '-'
            out_line = out_line + rescode
            if len(out_line) == 60:
                out_str = out_str + out_line + '\n'
                out_line = ""
        if out_line != '':
            out_str = out_str + out_line + '\n'
            out_line = ""
        # get sentries
        for sentry in self.sentryset.sentry_lst:
            sentry_id = sentry.sentry_id
            out_line = ""
            out_str = out_str + ">" + sentry_id + '\n'
            for refpos in range(max_refpos +1):
                refmap = self.refpos_dic[refpos]
                if refmap.sentry_dic.has_key(sentry_id):
                    if refmap.mutsentry_dic.has_key(sentry_id):
                        rescode = refmap.mutsentry_dic[sentry_id]
                    else:
                        rescode = refmap.uniprot_rescode
                else:
                    rescode = '-'
                out_line = out_line + rescode
                if len(out_line) == 60:
                    out_str = out_str + out_line + '\n'
                    out_line = ""
            if out_line != '':
                out_str = out_str + out_line + '\n'
                out_line = ""
        file_obj = file(filename, 'w')
        file_obj.write(out_str)
        file_obj.close()



class RefMap:
    """
    Residue position in Uniprot sequence
    """

    def __init__(self, uniprot_id, uniprotpos, uniprot_rescode, ins_nr, inspos):
        self.uniprot_id = uniprot_id
        self.uniprotpos = uniprotpos
        self.uniprot_rescode = uniprot_rescode   # residue code of uniprot, if insertion then '_'
        self.ins_nr     = ins_nr
        self.inspos     = inspos
        self.sentry_dic = {}
        self.mutsentry_dic = {}


        
    

class InsMap:
    """
    insertion data
    """

    def __init__(self, length, aa_seq, sentry_id, ins_interval):
        self.length = length
        self.aa_seq = aa_seq
        self.sentry_dic = {}
        self.sentry_dic[sentry_id] = ins_interval



    def add(self, sentry_id, ins_interval):
        self.sentry_dic[sentry_id] = ins_interval



    def compare(self, aa_seq):
        if self.aa_seq == aa_seq:
            return True
        return False


    
            
