"""
sifts_map.py
Extract residue level mapping data
from from XML files as provided by MSD SIFTS
mappings are extracted between uniprot and a given structural database 
Procedure is as follows:

get_maps read the sifts xml residue mapping files,
         extract the mappings for each structural database entry as stmap object
         save the mapping data as a stmap object for ench entry as byte stream

get_scop_maps same as get_maps but optimized for SCOP structural database
              as it extracts the id2 (sid code) as well as id (sunid code)
              needs a scop file in addition


"""


# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


# standard modules
import os
import os.path
import xml.dom.minidom
import cPickle

# struster modules
import log_util




def get_maps(maps_dir, out_dir, st_db):
    """
    Given a directory with xml stdb to uniprot maps,
    save structure to uniprot maps as data files (byte streams)
    in directory out_dir
    """

    # get list of xml map files
    files_lst = os.listdir(maps_dir) 
    maps_file_lst = []
    for file_name in files_lst:
        if len(file_name) >= 5:
            if file_name[-4:] == '.xml':
                path_file_name = os.path.join(maps_dir, file_name)
                maps_file_lst.append(path_file_name)
            
        
    # for each map file
    for map_file in maps_file_lst:
        try:
            # parse xml
            map_obj = UniprotStdbMap()
            map_obj.parse_map(map_file, st_db)
            # save map as data file
            map_obj.save_map_data(out_dir)
        except:
            log_util.log_error('Could not parse map file %s' % map_file)
            continue



def get_scop_maps(maps_dir, out_dir, scop_sunid_sid_file):
    """
    Given a directory with xml SCOP to uniprot maps,
    save structure to uniprot maps as data files (byte streams)
    in directory out_dir
    Based on get_maps fuction, but adds scop sid as st_id2, ans uses it for name of data files
    """

    st_db = 'SCOP'

    # get dictionary with correspondance between scop sunid and sid
    sunid_sid_dic = _get_sunid_sid(scop_sunid_sid_file)

    # get list of xml map files
    files_lst = os.listdir(maps_dir) 
    maps_file_lst = []
    for file_name in files_lst:
        if len(file_name) >= 5:
            if file_name[-4:] == '.xml':
                path_file_name = os.path.join(maps_dir, file_name)
                maps_file_lst.append(path_file_name)
            
        
    # for each map file
    for map_file in maps_file_lst:
        try:
            # parse xml
            map_obj = UniprotStdbMap()
            map_obj.parse_map(map_file, st_db)
            # add st_id2
            map_obj.add_st_id2(sunid_sid_dic)
            # save map as data file
            map_obj.save_map_data(out_dir)
        except:
            log_util.log_error('Could not parse map file %s' % map_file)
            continue






def get_map_file_name(sentry_id):
    """
    From sentry_id get the map_file_name.
    """
    map_file_name = sentry_id + ".data"
    return map_file_name



def _get_sunid_sid(scop_file):
    """
    Get dictionary with correspondance between scop sunid and sid
    """
    sunid_sid_dic = {}
    
    f_obj = file(scop_file, 'r')
    scop_lines = f_obj.readlines()
    for scop_line in scop_lines:
        if scop_line[0] == '#':
            continue
        line_lst = scop_line.split()
        if len(line_lst) == 0:
            continue
        if len(line_lst) != 6:
            log_util.log_warning('SCOP line not standard: %s' % scop_line)
            continue
        sid = line_lst[0]
        sunids_str = line_lst[5]
        sunid_lst = sunids_str.split(',')
        if sunid_lst[6][:3] != 'px=':
            log_util.log_warning('SCOP line not standard: %s' % scop_line)
            continue
        sunid = sunid_lst[6][3:]
        sunid_sid_dic[sunid] = sid

    f_obj.close()
    return sunid_sid_dic




class UniprotStdbMap:
    """Residue mapping between uniprot and structural database"""

    def __init__(self):
        self.map_file = None      # xml residue level mapping file
        self.st_db = None       # name of structural database: CATH, PDB, SCOP, UniProt
        self.stmap_dic = {}      # dictionay of structural entry residue maps StMap


    def parse_map(self, map_file, st_db):
        """
        Input
        map_file: name of xml residue mapping file
        st_db:   name of structure database
        Sets
        self.stmap_lst: list of structural entries with residue mappings to uniprot
        """
        
        self.map_file = map_file
        self.st_db = st_db
        log_util.log_info('Parsing %s' % map_file)
        
        dom_obj = xml.dom.minidom.parse(self.map_file)
        self._process_map(dom_obj)
        # check entries for uniprot consistency
        for stmap_obj in self.stmap_dic.values():
            stmap_obj._check_uniprot_data()
        log_util.log_info('Parsed %s' % self.map_file)



    def save_map_data(self, out_dir):
        """
        Save the mapping as a byte stream
        For each entry, file name is st_id2.data,
        or if  st_id2 not defined st_id.data
        """
        for stmap_obj in self.stmap_dic.values():
            if stmap_obj.st_id2 == None:
                data_file = stmap_obj.st_id + '.data'
            else:
                data_file = stmap_obj.st_id2 + '.data'
            path_data_file = os.path.join(out_dir, data_file)
            outf_obj = file(path_data_file, 'w')
            cPickle.dump(stmap_obj, outf_obj)
            outf_obj.close()
            log_util.log_info('Saved %s' % data_file)



    def add_st_id2(self, id1_id2_dic):
        """
        Add structure secondary id for each entry
        Used to add scop sunid as secondary id
        """

        for stmap_obj in self.stmap_dic.values():
            if id1_id2_dic.has_key(stmap_obj.st_id):
                stmap_obj.st_id2 = id1_id2_dic[stmap_obj.st_id]
            else:
                log_util.log_warning('Could not find id2 for id %s' % stmap_obj.st_id)
                


    def _process_map(self, dom_obj):
        """Process the dom object"""

        # loop over entities
        entity_lst = dom_obj.getElementsByTagName('entity')
        for entity_obj in entity_lst:
            seg_lst = entity_obj.getElementsByTagName('segment')
            # loop over segments
            for seg_obj in seg_lst:
                lres_lst = seg_obj.getElementsByTagName('listResidue')
                if len(lres_lst) != 1:
                    raise RuntimeError, 'More than one listResidue in one segment'
                residue_lst = lres_lst[0].getElementsByTagName('residue')
                #loop over residues
                for residue_obj in residue_lst:                    
                    crossRefDb_lst = residue_obj.getElementsByTagName('crossRefDb')
                    #loop over crossRefs
                    st_crossRefDb_obj = None
                    up_crossRefDb_obj = None
                    for crossRefDb_obj in crossRefDb_lst:
                        if crossRefDb_obj.attributes['dbSource'].value == self.st_db:
                            st_crossRefDb_obj = crossRefDb_obj
                        elif crossRefDb_obj.attributes['dbSource'].value == 'UniProt':
                            up_crossRefDb_obj = crossRefDb_obj
                        if st_crossRefDb_obj != None and up_crossRefDb_obj != None:
                            self._add_res_map(st_crossRefDb_obj, up_crossRefDb_obj)
                            break


    def _add_res_map(self, st_crossRefDb_obj, up_crossRefDb_obj):
        """Add each residue map"""
        st_dbversion = str(st_crossRefDb_obj.attributes['dbVersion'].value)
        st_dbcoordsys = str(st_crossRefDb_obj.attributes['dbCoordSys'].value)
        st_id = str(st_crossRefDb_obj.attributes['dbAccessionId'].value)
        st_reslabel_raw = str(st_crossRefDb_obj.attributes['dbResNum'].value)    #it is actually a label, not a number
        st_resname = str(st_crossRefDb_obj.attributes['dbResName'].value)
        if st_crossRefDb_obj.attributes.has_key('dbChainId'):
            st_chain = str(st_crossRefDb_obj.attributes['dbChainId'].value)
        else:
            # if no chain set to ' '
            st_chain = ' '

        up_dbversion = str(up_crossRefDb_obj.attributes['dbVersion'].value)
        up_dbcoordsys = str(up_crossRefDb_obj.attributes['dbCoordSys'].value)
        up_id = str(up_crossRefDb_obj.attributes['dbAccessionId'].value)
        up_resnr = int(up_crossRefDb_obj.attributes['dbResNum'].value)
        up_resname = str(up_crossRefDb_obj.attributes['dbResName'].value)

        # change residue label, add space if no insertion code
        st_reslabel = self._get_reslabel(st_reslabel_raw)

        # check if st entry already exists
        if self.stmap_dic.has_key(st_id) == 0:
            stmap_obj = StMap(self.st_db, st_dbcoordsys, st_dbversion, st_id)
            self.stmap_dic[st_id] = stmap_obj
        self.stmap_dic[st_id].add_resmap(st_dbcoordsys, st_dbversion, st_chain, st_reslabel, st_resname,
                                            up_dbversion, up_dbcoordsys, up_id , up_resnr, up_resname)



    def _get_reslabel(self, st_reslabel_raw):
        """
        Modify original residue label.
        Keep it if there is insertion code
        If no insertion code, then insert space ' '
        """
        
        if st_reslabel_raw[-1].isalpha():
             st_reslabel = st_reslabel_raw
        else:
            st_reslabel = st_reslabel_raw + ' '
        return st_reslabel
