"""
stmap_resmap.py

load_map_data load the residue mapping data (byte stream) saved by get_maps or get_scop_maps
               return the stmap object with the residue mapping data

StMap       basic class for residue mapping between uniprot and an entry in a structural database
               (PDB, SCOP or CATH)
               see class definition
Resmap      Mapping of a residue between a structure entry and a uniprot sequence
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import log_util
import cPickle



def load_map_data(map_data_file):
    """
    Load the mapa data file (byte stream) map_data_file
    return StMap object
    """
    
    fobj = file(map_data_file, 'r')
    stmap_obj = cPickle.load(fobj)
    fobj.close()
    # log_util.log_info('Loaded map data %s' % map_data_file)
    return stmap_obj



class StMap:
    """
    Structure entry residue mapping to uniprot.
    
    Single residue mapping between a structural database entry (PDB, SCOP, CATH) and uniprot
    """

    def __init__(self, st_db, st_dbcoordsys, st_dbversion, st_id):

        self.st_db = st_db               # structural db 
        self.st_dbversion = st_dbversion # structural db version
        self.st_id = st_id               # structural entry id
        self.st_id2 = None               # secondary structural entry id, this is sid for SCOP
        self.st_dbcoordsys = st_dbcoordsys
        self.resmap_lst = []  # list of residue map objects
        self.resmap_dic = {}  # dictionary of residue map objects, for quick access
                              # see _resmap_get_key for key definition



    def add_resmap(self, st_dbcoordsys, st_dbversion, st_chain, st_reslabel, st_resname, up_dbversion, up_dbcoordsys, up_id , up_resnr, up_resname):
        """Add a resmap to a entry"""
        
        # checking coordsys and db version are the same
        if st_dbcoordsys != self.st_dbcoordsys:
            raise RuntimeError, 'Different dbCoordSys for same structural entry % %s'
        elif st_dbversion != self.st_dbversion:
            raise RuntimeError, 'Different dbVersion for same structural entry'
        # get new resmap object and add it to list and dic
        resmap_obj = Resmap(st_chain, st_reslabel, st_resname, up_dbversion, up_dbcoordsys, up_id , up_resnr, up_resname)
        self.resmap_lst.append(resmap_obj)
        dic_key = self._resmap_get_key(st_reslabel, st_chain)
        self.resmap_dic[dic_key] = resmap_obj



    def get_resmap(self, st_reslabel, st_chain):
        """
        Get residue map object given st_reslabel and st_chain.
        Return None if no residue map for that residue label and chain
        """
        
        dic_key =  self._resmap_get_key(st_reslabel, st_chain)
        if self.resmap_dic.has_key(dic_key):
            resmap_obj =  self.resmap_dic[dic_key]
        else:
            resmap_obj = None
        return resmap_obj



    def write_map(self):
        """
        Write mapping to log
        """

        out_str = "st_db: %s st_dbversion: %s st_id: %s st_id2: %s st_dbcoordsys: %s" % \
                  (self.st_db, self.st_dbversion, self.st_id, self.st_id2, self.st_dbcoordsys)
        log_util.log_info(out_str)
        for resmap_obj in self.resmap_lst:
            out_str = "%s %s %s :: %d %s %s %s %s" % \
                      (resmap_obj.st_reslabel, resmap_obj.st_chain, resmap_obj.st_resname,
                       resmap_obj.up_resnr, resmap_obj.up_resname, resmap_obj.up_id,
                       resmap_obj.up_dbversion, resmap_obj.up_dbcoordsys)
            log_util.log_info(out_str)


            
    def _resmap_get_key(self, st_reslabel, st_chain):
        """generate a resmap key based on residue label and chain"""
        return "%s:%s" % (st_reslabel, st_chain)



    def _check_uniprot_data(self):
        """
        Check if up_id, up_dbversion, up_dbcoordsys are the same
        in all resiudes, if not write warning to log
        """
        up_id = self.resmap_lst[0].up_id
        up_dbversion = self.resmap_lst[0].up_dbversion
        up_dbcoordsys = self.resmap_lst[0].up_dbcoordsys
        
        for resmap_obj in self.resmap_lst:
            if resmap_obj.up_id != up_id:
                # change of sequence, in case of pdb it can happen (complex of two proteins)
                log_util.log_warning('Uniprot ID changed %s to %s' % (up_id, resmap_obj.up_id))
                up_id = resmap_obj.up_id
            if resmap_obj.up_dbversion != up_dbversion:
                log_util.log_warning('Uniprot dbversion changed %s to %s' % (up_dbversion, resmap_obj.up_dbversion))
                up_dbversion = resmap_obj.up_dbversion
            if resmap_obj.up_dbcoordsys != up_dbcoordsys:
                log_util.log_warning('Uniprot up_dbcoordsys changed %s to %s' % (up_dbcoordsys, resmap_obj.up_dbcoordsys))
                up_dbcoordsys = resmap_obj.up_dbcoordsys
                    

class Resmap:
    """
    Mapping of a residue between a structure entry and a uniprot sequence 
    """
     
    def __init__(self, st_chain, st_reslabel, st_resname, up_dbversion, up_dbcoordsys, up_id , up_resnr, up_resname):
        self.st_chain = st_chain
        self.st_reslabel = st_reslabel
        self.st_resname = st_resname

        self.up_dbversion = up_dbversion
        self.up_dbcoordsys = up_dbcoordsys
        self.up_id = up_id
        self.up_resnr = up_resnr
        self.up_resname = up_resname


        
