"""
STRuster Results
Clustering
Distance sdev
hinges
invariant regions
superposition
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import os
import cPickle
import rpy

import log_util
import util_r
import scop_sets
import clustering
import hinge_inv
import superposition

class StrusterRsl:

    def __init__(self, config_dic):
        # data files and directories
        self.config_dic = config_dic
        self.sets_file = config_dic['sets_file']       # definition of sets: from SCOP
        self.pdb_dir   = config_dic['pdb_dir']         # pdb directory
        self.use_pdb_subdir = config_dic['use_pdb_subdir']   # True/False, use subdirectory to access PDB
        self.stmap_dir   = config_dic['stmap_dir']     # residue mapping directory
        self.dpi_file = config_dic['dpi_file']        # SUS file
        self.uniprot_dir = config_dic['uniprot_dir']   # directory with uniprot sequences
        self.source_dir = config_dic['source_dir']     # source code directory
        # baseset parameters
        self.min_set_len = config_dic['min_set_len']  # minimum number of sentries in set
        self.max_set_len = config_dic['max_set_len']  # max number of entries in set, if no max then None
        # clustering parameters
        self.dissim_type = 'dmat'   # type of dissimilarity: dmat dali or rmsd
        self.dist_filter = config_dic['dist_filter']   # None or a float
        self.dif_dist_filter = config_dic['dif_dist_filter']  # None or a float
        self.one_gap_value = config_dic['one_gap_value']
        self.two_gap_value = config_dic['two_gap_value']
        # hinge invariant detection parameters
        self.save_data = config_dic['save_data']
        self.hi_score_cutoff = config_dic['hi_score_cutoff']
        self.hi_rel_score_cutoff = config_dic['hi_rel_score_cutoff']
        self.hi_filter_cutoff = config_dic['hi_filter_cutoff']
        self.hi_min_gap = config_dic['hi_min_gap']
        self.hi_inv_cutoff = config_dic['hi_inv_cutoff']
        self.get_superp = config_dic['get_superp']
        self.hi_filename = config_dic['hi_filename']
        
        
    def struster_scop_sets(self):
        # source R files
        util_r.source_r_clustering(self.source_dir)
        util_r.source_r_plots(self.source_dir)
        
        # read sets definition into a list of lines
        read_scop = scop_sets.ReadScopBaseset(self.min_set_len, self.max_set_len)
        read_scop.load_file(self.sets_file)

        struster_hi_dic = {}
        
        base_dir = os.getcwd()
        while(read_scop.line_counter < read_scop.nr_lines):
            last_lnr = read_scop.line_counter
            baseset = None
            
            try:
                # get each set from scop species level
                baseset = read_scop.get_baseset(self.pdb_dir, self.use_pdb_subdir)
                if baseset == None:
                    log_util.log_error('Could not get struster results input line nr:%d' % read_scop.line_counter)
                    if read_scop.line_counter == last_lnr:
                        # if same line on error, then increase line count
                        read_scop.line_counter += 1
                    continue                                
                baseset.set_working_dir(base_dir)
                baseset.get_alignment(self.stmap_dir, self.uniprot_dir)
                baseset.get_alig_dist_array()
                baseset.get_alig_sus(dpi_file = self.dpi_file)
                # alignment
                alig_file = os.path.join(baseset.working_dir, baseset.set_id + '.alig')
                baseset.alig.write_alignment(alig_file)
                fasta_file = os.path.join(baseset.working_dir, baseset.set_id + '_alig.fasta')
                baseset.alig.write_fasta_alig(fasta_file)
                                                                
                # clustering
                store_fdif = self.get_superp  # store fdif in order to compute superposition
                clusterset =  clustering.ClusterSet(baseset, self.dist_filter, self.dif_dist_filter, self.one_gap_value, self.two_gap_value, store_fdif) 
                clusterset.cluster_dmat()
                                                
                # Dist sdev
                set_lst = baseset.get_sentry_id_lst()
                
                dist_sdev_mat = baseset.get_dist_sdev_mat(set_lst)
                plot_name = baseset.set_id + '_set_dist_sdev'
                baseset.plot_sdev(dist_sdev_mat, plot_name)
                # dist sus
                dist_sus_sdev_mat = baseset.get_dist_sus_sdev_mat(set_lst, dist_sdev_mat)
                plot_name = baseset.set_id + '_set_dist_sus_sdev'
                baseset.plot_sdev(dist_sus_sdev_mat, plot_name)
                # relative dist sdev
                nmat = baseset.get_sqrtmean_dsusv_mat(set_lst)
                rel_dist_sdev_mat = baseset.get_rel_dist_sdev_mat(dist_sdev_mat, nmat)
                plot_name = baseset.set_id + '_set_rel_dist_sdev_mean'
                baseset.plot_relsdev(rel_dist_sdev_mat, plot_name)
                
                # inv/hinges
                # Dist sdev
                set_hi_dic = {}
                rsl_label = baseset.set_id + '_set_dist_sdev'
                und_mat = baseset.get_undef_mat(dist_sdev_mat)
                hi_obj = hinge_inv.HingeInv(rsl_label, self.source_dir, baseset.working_dir, self.save_data)
                hi_obj.get_hinges_invariant(score_mat = dist_sdev_mat, undef_mat= und_mat, mat_type= 'set', score_cutoff = self.hi_score_cutoff, filter_cutoff = self.hi_filter_cutoff, min_gap = self.hi_min_gap, inv_cutoff=self.hi_inv_cutoff)
                hi_obj.get_res_nr(baseset.alig)
                dsdev_dic = hi_obj.get_hinge_inv_dic()
                set_hi_dic['dsdev_dic'] = dsdev_dic
                # relative dist sdev
                rsl_label = baseset.set_id + '_set_rel_dist_sdev_mean'
                und_mat = baseset.get_undef_mat(rel_dist_sdev_mat)                
                hi_obj = hinge_inv.HingeInv(rsl_label, self.source_dir, baseset.working_dir, self.save_data)
                hi_obj.get_hinges_invariant(score_mat = rel_dist_sdev_mat, undef_mat= und_mat, mat_type= 'set', score_cutoff = self.hi_rel_score_cutoff, filter_cutoff = self.hi_filter_cutoff, min_gap = self.hi_min_gap, inv_cutoff=self.hi_inv_cutoff)
                hi_obj.get_res_nr(baseset.alig)
                reldsdev_dic = hi_obj.get_hinge_inv_dic()
                set_hi_dic['reldsdev_dic'] = reldsdev_dic                
                # dist sus, last onse, used for superposition
                rsl_label = baseset.set_id + '_set_dist_sus_sdev'
                und_mat = baseset.get_undef_mat(dist_sus_sdev_mat)                
                hi_obj = hinge_inv.HingeInv(rsl_label, self.source_dir, baseset.working_dir, self.save_data)
                hi_obj.get_hinges_invariant(score_mat = dist_sus_sdev_mat, undef_mat= und_mat, mat_type= 'set', score_cutoff = self.hi_score_cutoff, filter_cutoff = self.hi_filter_cutoff, min_gap = self.hi_min_gap, inv_cutoff=self.hi_inv_cutoff)
                # convert to uniprot sequence residue number
                hi_obj.get_res_nr(baseset.alig)
                # save data
                filename = os.path.join(baseset.working_dir, baseset.set_id + '_hinge_inv.data')
                dsussdev_dic = hi_obj.save_hinge_inv(filename)
                # store data
                set_hi_dic['dsussdev_dic'] = dsussdev_dic
                struster_hi_dic[baseset.set_id] = set_hi_dic
                
                if self.get_superp == True:
                    # superposition
                    max_superp = 3
                    superp_counter = 0
                    for inv_region in hi_obj.inv_region_lst:
                        if superp_counter > max_superp:
                            break
                        superp = superposition.Superp(self.config_dic, baseset, set_lst, clusterset.fdif_dic)
                        superp_pdb_filename = baseset.set_id + '_superp_'+ str(superp_counter+1) + '.pdb'
                        superp_pdb_file = os.path.join(baseset.working_dir, superp_pdb_filename)
                        superp.superimpose_subsets(inv_region, superp_pdb_file)
                        datafilename = baseset.set_id + '_superp_'+ str(superp_counter+1) + '.data'
                        datafile = os.path.join(baseset.working_dir, datafilename)
                        superp.save_superp_data(datafile)
                        superp_rasm_filename = baseset.set_id + '_superp_'+ str(superp_counter+1) + '.rsm'
                        superp_rasm_file = os.path.join(baseset.working_dir, superp_rasm_filename)
                        superp.get_rasmol_script(superp_pdb_file, inv_region, hi_obj.hinges_lst, superp_rasm_file)
                        superp_counter += 1
                
                d_lst = dir()
                if d_lst.count('superp')>0:
                    del superp
                if d_lst.count('hi_obj')>0:
                    del hi_obj
                if d_lst.count('clusterset')>0:
                    del clusterset
                if d_lst.count('baseset')>0:
                    del baseset
            except:
                log_util.log_error('Could not get struster results input line nr:%d' % read_scop.line_counter)
                if read_scop.line_counter == last_lnr:
                    # if same line on error, then increase line count
                    read_scop.line_counter += 1
                d_lst = dir()
                if d_lst.count('superp')>0:
                    del superp
                if d_lst.count('hi_obj')>0:
                    del hi_obj
                if d_lst.count('clusterset')>0:
                    del clusterset
                if d_lst.count('baseset')>0:
                    del baseset                
                continue                
        pickle_protocol = 2
        # datafile = 'struster_hinge_inv.data'
        fobj = file(self.hi_filename, 'w')
        cPickle.dump(struster_hi_dic, fobj, pickle_protocol)
        fobj.close()
        log_util.log_info('Finished STRuster results for %s' % self.sets_file)
                
        