# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de



import copy
import os
import os.path
import cPickle
import Numeric
import Bio.PDB

import log_util
import clustering

class Superp:
    """
    Superposition of two subsets
    """

    def __init__(self, config_dic, baseset, subset_lst, fdif_dic):
        # base data
        self.baseset= baseset
        self.subset_lst = subset_lst
        self.subset_rep = None
        
        # data and parameters
        self.stmap_dir   = config_dic['stmap_dir']
        self.uniprot_dir = config_dic['uniprot_dir']
        self.pdb_superp_dir = config_dic['pdb_superp_dir']
        self.pdb_superp_type = config_dic['pdb_superp_type']   # astral or pdb
        self.tmp_dir = config_dic['tmp_dir']
        self.fdif_dic = fdif_dic
        self.use_pdb_subdir = config_dic['use_pdb_subdir']   # True/False use subdirectory to access pdb
        # computed data
        self.transf_dic = {}
        self.superp_pdb_file = None
        # self.dissim_mat
        
        
    def superimpose_subsets(self, eqv_alig_lst, superp_pdb_file):
        """
        Superimpose two subsets given the list of equivalent residues as positions in alignment
        """
        self.superp_pdb_file = superp_pdb_file
        
        # get representatives
        self.get_representative(eqv_alig_lst)
        
        # get list of refpos
        eqv_refpos_lst = self.get_refpos_lst(eqv_alig_lst)
        
        # get transformations to superimpose subset to representative
        ret_val = self.subset_transf(rep_id = self.subset_rep, subset_lst = self.subset_lst, refpos_lst = eqv_refpos_lst)
        if ret_val == True:                
            # write pdb file
            self.write_superp_pdb()
            log_util.log_info('Wrote transformed PDB: %s' % superp_pdb_file)
        else:
            log_util.log_warning('Could not generate superposition: %s' % self.baseset.set_id)
        
        
            
    def get_representative(self, eqv_alig_lst):
        """
        Compute the representative of the subset
        it corresponds to the entry with lowest sum of difference distances to all other entries over the equivalent residues of subset
        """
                
        rep_id = None
        rep_dissim_sum = None
        for sentry_i in self.subset_lst:
            dissim_sum = 0.0
            for sentry_j in self.subset_lst:
                if sentry_j != sentry_i:
                    dissim_sum += self._get_dissim(sentry_i, sentry_j, eqv_alig_lst)
            if rep_dissim_sum == None:
                rep_dissim_sum = dissim_sum
                rep_id = sentry_i
            elif dissim_sum < rep_dissim_sum:
                rep_dissim_sum = dissim_sum
                rep_id = sentry_i
            log_util.log_info('Sentry %s, dissimilarity sum %f' % (sentry_i,dissim_sum))
        log_util.log_info("Representative %s, dissim sum %f" % (rep_id, rep_dissim_sum))
        self.subset_rep = rep_id                 
                     
    
    
    def get_rasmol_script(self, pdb_file, inv_region_lst, hinges_lst, ras_filename):
        """
        Make a rasmol script coloring the invariant region, hinges, variable
        Adapted to run on jmol
        """

        inv_color = "bluetint"
        hinge_color = "gold"
        bck_width = 80
        pdb_filename = os.path.basename(pdb_file)
        pdb_filename_comp = pdb_filename+ '.gz'
        f_obj = file(ras_filename, 'w')
        rasmol_cmd = "load %s\nframe 0\nselect *\ncpk off\nwireframe off\nbackbone %d\n" % (pdb_filename_comp, bck_width)
        f_obj.write(rasmol_cmd)
        sentry_nr = 1
        for sentry_id in self.subset_lst:
            for inv_seg in inv_region_lst:
                sentry_seg_lst = self._get_sentry_seg(inv_seg, sentry_id)
                if sentry_seg_lst != None:
                    self._get_rasmol_cmd(sentry_nr, sentry_id, sentry_seg_lst, inv_color, f_obj)
            for hinge_seg in hinges_lst:
                sentry_seg_lst = self._get_sentry_seg(hinge_seg, sentry_id)
                if sentry_seg_lst != None:
                    self._get_rasmol_cmd(sentry_nr, sentry_id, sentry_seg_lst, hinge_color, f_obj)
            sentry_nr += 1
        f_obj.close()
        log_util.log_info('Wrote %s' % ras_filename)
                
        
    def _get_rasmol_cmd(self, sentry_nr, sentry_id, sentry_seg_lst, inv_color, f_obj):
        for seg in sentry_seg_lst:
            sentry= self.baseset.sentryset.get_sentry(sentry_id)
            resn_st = sentry.res_lst[seg[0]].nr
            resn_end = sentry.res_lst[seg[1]].nr
            chain = sentry.res_lst[seg[1]].chain_id
            rasmol_cmd = "select */%d and (%d-%d) and *%s \n" % (sentry_nr, resn_st, resn_end, chain)
            f_obj.write(rasmol_cmd)
            rasmol_cmd = "colour %s\n" % inv_color
            f_obj.write(rasmol_cmd)
    
            
            
    def _get_sentry_seg(self, seg_lst, sentry_id):
        """
        Given a segment in ref pos, return the sentry pos
        if there is a change in chain ids, then return different segments, with chain id ends
        """
        ref_st = seg_lst[0]
        ref_end = seg_lst[1]
        sentry_st = None
        # get start and ends of sentry
        for ref_pos in range(ref_st, ref_end+1):
            if self.baseset.alig.refpos_dic[ref_pos].sentry_dic.has_key(sentry_id):
                sentry_st = self.baseset.alig.refpos_dic[ref_pos].sentry_dic[sentry_id] 
                sref_st = ref_pos 
                break
        if sentry_st == None:
            return None
        sentry_end = None
        for ref_pos in range(ref_end, sref_st, -1):
            if self.baseset.alig.refpos_dic[ref_pos].sentry_dic.has_key(sentry_id):
                sentry_end = self.baseset.alig.refpos_dic[ref_pos].sentry_dic[sentry_id] 
                sref_end = ref_pos
                break
        if sentry_end == None:
            return None
            
        sentry= self.baseset.sentryset.get_sentry(sentry_id)
        chain_st = sentry.res_lst[sentry_st].chain_id
        chain_end = sentry.res_lst[sentry_end].chain_id
        # check if start and end in same chain
        if chain_st == chain_end:
            return [[sentry_st, sentry_end]]
        else:
            # not in same chain, find ends of chains
            sentry_chain_end = None
            sentry_chain_st = None
            last_chain = chain_st
            last_sentry_chain_st = sentry_st
            last_sentry_chain_end = sentry_st
            chainseg_lst = []
            for sentry_pos in range(sentry_st, sentry_end+1):
                chain = sentry.res_lst[sentry_pos].chain_id
                if chain != last_chain:
                    if chain != chain_end:
                        last_sentry_chain_end = sentry_pos - 1
                        chainseg_lst.append([last_sentry_chain_st, last_sentry_chain_end])
                        last_sentry_chain_st = sentry_pos
                    else:
                        last_sentry_chain_end = sentry_pos - 1
                        chainseg_lst.append([last_sentry_chain_st, last_sentry_chain_end])
                        chainseg_lst.append([sentry_pos, sentry_end])   
                        return chainseg_lst
            return None
                    
        
    def _get_dissim(self, sentry_i, sentry_j, eqv_alig_lst):
        fdif_dist_mat = self.fdif_dic[(sentry_i,sentry_j)]
        d = fdif_dist_mat.shape[0]
        eqv_fdif_dist_mat = Numeric.zeros([d,d], Numeric.Float)
        for tpl in eqv_alig_lst:
            idx0 = tpl[0]
            idx1 = tpl[1]+1
            eqv_fdif_dist_mat[idx0:idx1,idx0:idx1] = fdif_dist_mat[idx0:idx1,idx0:idx1]
        dissim_value = Numeric.sum(Numeric.sum(eqv_fdif_dist_mat)) 
        return dissim_value
    
    
    
    def get_refpos_lst(self, eqv_alig_lst):
        refpos_lst = []
        for alig_seg in eqv_alig_lst:
            refpos_seg_lst = []
            alig_st = alig_seg[0]
            alig_end = alig_seg[1]
            refpos_seg_lst = range(alig_st, alig_end+1)
            refpos_lst += refpos_seg_lst
        # log_util.log_info("refpos_lst %s" % str(refpos_lst))  # debug
        return refpos_lst
    
        
    
    def subset_transf(self, rep_id, subset_lst, refpos_lst, rep_atom_lst = None):    
        if rep_atom_lst == None:
            rep_atom_lst = self._get_atom_from_alig(rep_id, refpos_lst)
        for sentry_id in subset_lst:
            if sentry_id == rep_id:
                continue
            sentry_atom_lst = self._get_atom_from_alig(sentry_id, refpos_lst)
            (transf_tpl,rmsd) = self._get_transf(rep_atom_lst, sentry_atom_lst, refpos_lst)    
            if transf_tpl == None:
                return False            
            self.transf_dic[sentry_id] = (transf_tpl,rmsd)
        return True
            
        
                
    def save_superp_data(self, datafile):
        pickle_protocol = 2
        fobj = file(datafile, 'w')
        cPickle.dump(self.transf_dic, fobj, pickle_protocol)
        fobj.close()
        log_util.log_info('Saved %s' % datafile)
    
    
    
    def _get_atom_from_alig(self, sentry_id, refpos_lst):
        # get sentrypos
        sentrypos_lst = []        
        for refpos in refpos_lst:
            if self.baseset.alig.refpos_dic[refpos].sentry_dic.has_key(sentry_id):
                sentrypos = self.baseset.alig.refpos_dic[refpos].sentry_dic[sentry_id]  
                sentrypos_lst.append(sentrypos)
            else:
                sentrypos_lst.append(None)
        # print (sentry_id, sentrypos_lst)   # debug
        
        # get list of atom objects
        sentry = self.baseset.sentryset.get_sentry(sentry_id)
        atom_lst = sentry.get_atom_lst(sentrypos_lst)
        return atom_lst
    
    
    
    def _get_transf(self, atom1_lst, atom2_lst, refpos_lst):
        (f_atom1_lst, f_atom2_lst, f_refpos_lst) = self._filter_atom_lists(atom1_lst, atom2_lst, refpos_lst)
        sup_obj = Bio.PDB.Superimposer()
        # log_util.log_info("Length atom lists %d , %d" %  (len(f_atom1_lst),len(f_atom2_lst))) # debug
        if len(f_atom1_lst) > 0 and len(f_atom2_lst) > 0:
            sup_obj.set_atoms(f_atom1_lst, f_atom2_lst)
            transf_tpl =  sup_obj.rotran
            rmsd = sup_obj.rms
        else:
            transf_tpl = None
            rmsd = None
        return (transf_tpl, rmsd)

    
    
    def _filter_atom_lists(self, atom1_lst, atom2_lst, refpos_lst):
        f_atom1_lst = []
        f_atom2_lst = []
        f_refpos_lst = []
        n = len(atom1_lst)
        if len(atom2_lst) != n or len(refpos_lst) != n:
            log_util.log_error('Lists of different size: %d, %d %d' % (n, len(atom2_lst), len(refpos_lst)))
            return (None, None, None) 
        for i in range(n):
            if atom1_lst[i] != None and atom2_lst[i] != None:
                f_atom1_lst.append(atom1_lst[i])
                f_atom2_lst.append(atom2_lst[i])
                f_refpos_lst.append(refpos_lst[i])
        return (f_atom1_lst, f_atom2_lst, f_refpos_lst)
            
            
            
    def write_superp_pdb(self):
        # open output file
        out_fobj = file(self.superp_pdb_file, 'w')
        
        # for each pdb in subset
        model_nr = 1
        for sentry_id in self.subset_lst:
            if sentry_id == self.subset_rep:
                self._append_transf_model(sentry_id, model_nr, None, out_fobj)
            else:
                self._append_transf_model(sentry_id, model_nr, self.transf_dic[sentry_id][0], out_fobj)
            model_nr += 1
        out_str = 'END   \n'
        out_fobj.write(out_str)
        out_fobj.close()
        # compress
        cmd = 'gzip ' +  self.superp_pdb_file
        os.system(cmd)
        log_util.log_info(cmd)
        
        
        
    def _append_transf_model(self, sentry_id, model_nr, transf_tpl, out_fobj):
        # write MODEL to output file
        out_str = "MODEL     %4d   %s\n" % (model_nr, sentry_id)
        out_fobj.write(out_str)
        # get pdb lines
        pdb_lines = self._get_pdb_lines(sentry_id)
        model_flag = 0
        # for each line
        for line_str in pdb_lines:
            if line_str[:6] == 'ATOM  ' or line_str[:6] == 'HETATM':
                out_str = self._transf_coord_line(line_str, transf_tpl)
                out_fobj.write(out_str)
            elif line_str[:6] == 'TER   ' or line_str[:6] == 'CONECT':
                out_fobj.write(line_str)
            elif line_str[:6] == 'MODEL ':
                if model_flag == 0:
                    model_flag = 1
                else:
                    break
            elif line_str[:6] == 'ENDMDL':
                break
        out_str = 'ENDMDL\n'
        out_fobj.write(out_str)
        
        
    def _get_pdb_lines(self, sentry_id):
        if self.pdb_superp_type == 'astral':
            pdb_file_name =  sentry_id + '.ent'
            if self.use_pdb_subdir == True:
                pdb_dir = os.path.join(self.pdb_superp_dir, sentry_id[2:4])
            else:
                pdb_dir = self.pdb_superp_dir
            pdb_file = os.path.join(pdb_dir, pdb_file_name)
            fobj = file(pdb_file, 'r')
            pdb_lines =  fobj.readlines()
            fobj.close()      
        else:
            pdb_file_name = 'pdb' + sentry_id[1:5] + '.ent.Z'
            pdb_file = os.path.join(self.pdb_superp_dir, pdb_file_name)
            exec_cmd = 'cp ' + pdb_file + ' ' + self.tmp_dir
            log_util.log_info(exec_cmd)
            os.system(exec_cmd)
            pdb_file = os.path.join(self.tmp_dir, pdb_file_name)
            exec_cmd = 'uncompress ' + pdb_file
            log_util.log_info(exec_cmd)
            os.system(exec_cmd)
            pdb_file_name = pdb_file_name[:-2]
            pdb_file = os.path.join(self.tmp_dir, pdb_file_name)
            fobj = file(pdb_file, 'r')
            pdb_lines =  fobj.readlines()
            fobj.close()     
            exec_cmd = 'rm -f ' + pdb_file
            log_util.log_info(exec_cmd)
            os.system(exec_cmd)
        return pdb_lines
            
        
        
    def _transf_coord_line(self, line_str, transf_tpl):
        if transf_tpl == None:
            return line_str
        transf_rot = transf_tpl[0]
        transf_trans = transf_tpl[1]
        x = float(line_str[30:38].strip())
        y = float(line_str[38:46].strip())
        z = float(line_str[46:54].strip())
        coord_init = Numeric.array([x,y,z], Numeric.Float)
        coord_t = Numeric.matrixmultiply(coord_init,transf_rot) + transf_trans
        part1 = line_str[0:30]
        part2 = "%8.3f%8.3f%8.3f" % (coord_t[0],coord_t[1],coord_t[2])
        part3 = line_str[54:]
        transf_line = part1+part2+part3
        return transf_line
    