"""
Coordinate Standard Uncertanties (s.u.s)
Extract paramters from refinement
DPI estimated according to Acta Cryst D55 583-601 1999
"""

# Copyright 2007 Max-Planck-Institut Informatik
# 
#     This file is part of STRuster.
# 
#     STRuster is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     STRuster is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with STRuster; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# Contact:
#     Francisco S. Domingues
#     email: doming@mpi-sb.mpg.de


import os.path
import os
import types

import log_util
import mmLib.mmCIF
import mmLib.FileIO
import math
import string



def get_pdblist(pdb_derived_data_file, selected_pdb_file):
    """
    Given the PDB file pdb_derived_data_file
    extract a list of PDB entries with proteins and determined by diffraction.
    """
    fobj_out = file(selected_pdb_file, 'w')
    fobj = file(pdb_derived_data_file, 'r')
    lines = fobj.readlines()
    for line_str in lines:
        line_str = line_str.strip()
        if line_str == '':
            continue
        line_lst = line_str.split()
        # check for diffraction
        if line_lst[2] == 'diffraction':
            # check for protein pdb files
            if string.find(line_lst[1], 'prot') >= 0:
                out_str = "%s\n" % line_lst[0]
                fobj_out.write(out_str)
    fobj.close()
    fobj_out.close()

 

class SUSpdb:
    """
    Get sus values for a set of pdb entries
    """
    
    def __init__(self):
        self.sus_dic = {}
    
        
        
    def extract_sus_from_pdblist_to_file(self, pdblist_file, mmcif_dir, tmp_dir, start_lnr, end_lnr, sus_file):
        """
        Extract sus for a given list of pdb entries, write a sus file
        Restrict from line numbers start_lnr to end_lnr, inclusive
        start_lnr starts at 1 (not 0)
        use string 'end' to mark last line
        """
    
        log_util.log_info('Start extracting SUS on %s, from  %s to %s, output %s' % (pdblist_file, str(start_lnr), str(end_lnr), sus_file))
        
        sus_obj = SUS()
        
        fobj_out = file(sus_file, 'w')
        sus_file_columns = sus_obj.get_column_labels(comment = 'y')
        fobj_out.write(sus_file_columns)

        pdb_lst = self._get_pdbids(pdblist_file, start_lnr, end_lnr)
        for pdb_id in pdb_lst:
            try:
                sus_obj.reset()
                sus_obj.get_refinement_data(pdb_id, mmcif_dir, tmp_dir) 
                sus_obj.calc_dpi()
                out_tpl = sus_obj.get_line_tpl()
                fobj_out.write(str(out_tpl))
                fobj_out.write("\n")
                log_util.log_info('Extracted sus: %s' % pdb_id)
            except:
                log_util.log_error('Could not extract sus for %s' % pdb_id)
                continue
        fobj_out.close()
        log_util.log_info('Finished extracting sus on %s, from  %s to %s, output %s' % (pdblist_file, str(start_lnr), str(end_lnr), sus_file))
        
        
        
    def merge_sus_files(self, sus_files_lst, sus_merged_file):   
        """
        Merge the files in sus_files_lst into sus_merged_file
        by the order in sus_files_lst
        comment lines starting with # are excluded
        Added a column label to start of file defined in sus_file_columns
        """
        fobj_out = file(sus_merged_file, 'w')
        sus_obj = SUS()
        sus_file_columns = sus_obj.get_column_labels(comment = 'y')  
        fobj_out.write(sus_file_columns)
        for sus_file in sus_files_lst:
            fobj = file(sus_file, 'r')
            lines = fobj.readlines()
            fobj.close()
            for line_str in lines:
                line_str = line_str.strip()
                if line_str == '':
                    continue
                if line_str[0] == '#':
                    continue
                fobj_out.write(line_str + '\n')
        fobj_out.close()
        
        
        
    def read_sus_file(self, sus_file):
        """
        Read sus file and set sus_dic
        """
        sus_dic = {}
        fobj = file(sus_file, 'r')
        lines = fobj.readlines()
        for line_str in lines:
            line_str = line_str.strip()
            if line_str[0] == "#":
                continue
            line_tpl = eval(line_str)
            sus_obj = SUS()
            sus_obj.read_line_tpl(line_tpl)
            pdb_id = sus_obj.pdb_id
            sus_dic[pdb_id] = sus_obj
        fobj.close()
        self.sus_dic = sus_dic

        
        
    def write_sus_file(self, sus_file):
        """
        Write a sus file from the sus_dic
        """
        fobj_out = file(sus_file, 'w')
        sus_obj = SUS()
        sus_file_columns = sus_obj.get_column_labels(comment = 'y')
        fobj_out.write(sus_file_columns)
        pdb_id_lst = self.sus_dic.keys()
        pdb_id_lst.sort()
        for pdb_id in pdb_id_lst:
            sus_obj = self.sus_dic[pdb_id]
            line_tpl = sus_obj.get_line_tpl()
            fobj_out.write(str(line_tpl))
            fobj_out.write("\n")
        fobj_out.close()
        log_util.log_info("Wrote sus file %s" % sus_file)
        
    
        
    def write_sus_Rdataframe(self, df_file):
        """
        Export sus from sus_dic to a R data frame
        New df_file is test file
        Entries are sorted alphabetically according to pdbid
        First line is column labels
        """
        fobj_out = file(df_file, 'w')
        sus_obj = SUS()
        sus_file_columns = sus_obj.get_column_labels(comment = 'n')
        fobj_out.write(sus_file_columns)
        sus_dic = self.sus_dic
        pdb_lst = sus_dic.keys()
        pdb_lst.sort()
        for pdb_id in pdb_lst:
            sus_obj = sus_dic[pdb_id]
            out_str = sus_obj.get_line_Rdataframe()
            fobj_out.write(out_str)
        fobj_out.close()

    
        
    def sus_filter_dpi(self):
        """
        Filter sus_dic
        Select the entries that have DPI
        Exclude entries witout DPI
        """
        pdb_id_lst = self.sus_dic.keys()
        pdb_id_lst.sort()
        for pdb_id in pdb_id_lst:
            sus_obj = self.sus_dic[pdb_id]
            if sus_obj.filter_dpi() == 0:
                del self.sus_dic[pdb_id]

                

    def sus_filter_parameters(self):
        """
        Filter out entries on sus_dic
        for different parameter cutofs
        """
        pdb_id_lst = self.sus_dic.keys()
        for pdb_id in pdb_id_lst:
            sus_obj = self.sus_dic[pdb_id]
            if sus_obj.filter_parameters() == 0:
                del self.sus_dic[pdb_id]
            
    
    def sus_filter_parameters2(self):
        """
        Filter out entries on sus_dic
        for different parameter cutofs
        """
        pdb_id_lst = self.sus_dic.keys()
        for pdb_id in pdb_id_lst:
            sus_obj = self.sus_dic[pdb_id]
            if sus_obj.filter_parameters() == 0:
                sus_obj.DPIa = None
                sus_obj.DPIf = None
                sus_obj.DPI = None
    
                
                                            
    def _get_pdbids(self, pdblist_file, start_lnr, end_lnr):
        pdb_lst = []
        if end_lnr < start_lnr:
            return []
        fobj = file(pdblist_file, 'r')
        lines = fobj.readlines()
        fobj.close()
        for line_str in lines:
            pdb_id = line_str.strip()
            pdb_lst.append(pdb_id)
        if  start_lnr < 1:
            start_idx = 1
        else:
            start_idx = start_lnr - 1
        if end_lnr == 'end':
            end_idx = len(pdb_lst)
        elif end_lnr > len(pdb_lst):
            end_idx = len(pdb_lst)
        else:
            end_idx = end_lnr
        sel_pdb_lst = pdb_lst[start_idx:end_idx]
        return sel_pdb_lst

        
class SUS:
    """
    SUS values and refinment values used to compute it
    """

    def __init__(self):
        self.reset()

        
                
    def reset(self):
        self.pdb_id = None       # PDB id
        self.dep_date = None    # PDB deposition
        self.Bavg = None         # average temperature factors
        self.Ni = None          # nr coorindates for fully occupied sites
        self.npar = None        # nr parameters
        self.nobs = None        # nr unique reflections
        self.cpl = None         # data completeness
        self.R = None           # crystallographic R value
        self.Rfree = None       # free R value
        self.dmin = None        # maximum resolution of data
        self.DPIa = None        # DPI (diffraction precision indicator) from R
        self.DPIf = None        # DPI from Rfree
        self.DPI = None         # DPIf, if None then DPIa
        self.luzzati_obs = None # Luzzati sus from R
        self.luzzati_free = None # Luzzati sus from free R
        self.luzzati = None      # luzzati_free, if None then luzzati_obs 
        self.sigma_a_obs = None  # sigma A sus from R
        self.sigma_a_free = None # sigma A sus from Rfree
        self.sigma_a = None      # sigma_a_free, if None then sigma_a_obs
        self.sigatm = None  # availabe experimental sus for coordinates: True/False
        self.anisou = None # available standard anisotropic atomic displacement: True/False


        
    def read_line_tpl(self, line_tpl):
        # get the data from the line tuple from sus text file
        self.pdb_id = line_tpl[0]
        self.dep_date =  line_tpl[1]
        self.Bavg = line_tpl[2]
        self.Ni = line_tpl[3]
        self.npar = line_tpl[4]
        self.nobs = line_tpl[5]
        self.cpl = line_tpl[6]
        self.R = line_tpl[7]
        self.Rfree = line_tpl[8]
        self.dmin = line_tpl[9]
        self.DPIa = line_tpl[10]
        self.DPIf = line_tpl[11]
        self.DPI  = line_tpl[12]
        self.luzzati_obs = line_tpl[13]
        self.luzzati_free = line_tpl[14]
        self.luzzati = line_tpl[15]
        self.sigma_a_obs = line_tpl[16]
        self.sigma_a_free = line_tpl[17]
        self.sigma_a = line_tpl[18]
        self.sigatm = line_tpl[19]
        self.anisou = line_tpl[20]

        
    
    def get_column_labels(self, comment = 'y'):
        """
        define the labels for columns in sus file
        and in R data frame
        """
        column_str = "pdbid date Bavg Ni npar nobs cpl R Rfree dmin DPIa DPIf DPI luzzati_obs luzzati_free luzzati sigma_a_obs sigma_a_free sigma_a sigatm anisou\n"
        if comment == 'y':
            return "# %s" % column_str
        else:
            return column_str
    
                
        
    def get_line_tpl(self):
        line_tpl = (self.pdb_id, self.dep_date,
                    self.Bavg, self.Ni, self.npar, self.nobs,
                    self.cpl, self.R, self.Rfree, self.dmin,
                    self.DPIa, self.DPIf, self.DPI,
                    self.luzzati_obs, self.luzzati_free, self.luzzati,
                    self.sigma_a_obs, self.sigma_a_free, self.sigma_a,
                    self.sigatm, self.anisou)
        return line_tpl

    
    
    def get_line_Rdataframe(self):
        out_str = "%s %s %s %s %s %s\
        %s %s %s %s\
        %s %s %s\
        %s %s %s\
        %s %s %s\
        %s %s\n" % \
        (self.pdb_id, self.dep_date, str(self.Bavg), str(self.Ni), str(self.npar), str(self.nobs),
        str(self.cpl), str(self.R), str(self.Rfree), str(self.dmin),
        str(self.DPIa), str(self.DPIf), str(self.DPI),
        str(self.luzzati_obs), str(self.luzzati_free), str(self.luzzati), 
        str(self.sigma_a_obs), str(self.sigma_a_free), str(self.sigma_a),
        str(self.sigatm), str(self.anisou))
        return out_str
        
    
    def get_refinement_data(self, pdb_id, mmcif_dir, tmp_dir):
        """
        Given a pdb id, load mmcif file and extract refinment data
        for calculation of DPI
        """
        # get full path name of mmcif file 
        mmcifZ_filename = pdb_id + '.cif.Z'
        mmcifZ_file = os.path.join(mmcif_dir, mmcifZ_filename)
        # copy
        exec_cmd = 'cp ' + mmcifZ_file + ' ' + tmp_dir
        os.system(exec_cmd)
        # uncompress
        mmcifZ_file = os.path.join(tmp_dir, mmcifZ_filename)
        exec_cmd = 'uncompress ' + mmcifZ_file
        os.system(exec_cmd)
        # get  mmcif data
        mmcif_filename = mmcifZ_filename[:-2]
        mmcif_file = os.path.join(tmp_dir, mmcif_filename)
        self.pdb_id = pdb_id
        self._get_mmcif_data(mmcif_file)
        
        #remove mmcif
        exec_cmd = 'rm -f ' + mmcif_file
        os.system(exec_cmd)



    def calc_dpi(self):
        """
        Calculate DPIa, DPIf
        """
        # check basic data is set
        if self.Ni != None and self.nobs != None and self.cpl != None and self.dmin != None:
            # DPIa
            if self.npar != None and self.R != None:
                p = float(self.nobs) - float(self.npar)
                if p > 0:
                    self.DPIa = math.sqrt(float(self.Ni)/p) * float(self.cpl) ** (-1/3.0) * self.R * self.dmin
            # DPIf
            if self.Rfree != None:
                self.DPIf = math.sqrt(float(self.Ni)/float(self.nobs)) * float(self.cpl) ** (-1/3.0) * self.Rfree * self.dmin

        self.DPI = self.get_DPI()


    
    def get_DPI(self):
        if self.DPIf != None:
            return self.DPIf
        if self.DPIa != None:
            return self.DPIa
        return None
    
    
    
    def get_luzzati(self):
        if self.luzzati_free != None:
            return self.luzzati_free
        if self.luzzati_obs != None:
            return self.luzzati_obs
        return None
    
    
    
    def get_sigma_a(self):
        if self.sigma_a_free != None:
            return self.sigma_a_free
        if self.sigma_a_obs != None:
            return self.sigma_a_obs
        return None
        

    
    def filter_dpi(self):
        if self.DPI == None:
            return 0
        return 1
    
    
    
    def filter_parameters(self):    
        if self.nobs < 600:
            return 0
        elif (float(self.npar) / float(self.nobs)) > 2.0:
            return 0
        elif self.cpl > 1.0 or self.cpl < 0.2:
            return 0
        elif self.Ni < 2:
            return 0
        elif self.dmin < 0.2:
            return 0        
        # DPI
        if self.DPIf != None:
            if self.DPIf < 0.001:
                self.DPIf = None
        if  self.DPIa != None:
            if self.DPIa < 0.001:
                self.DPIa = None
            elif (float(self.npar) / float(self.nobs)) > 0.6:
                self.DPIa = None
        if self.Rfree != None:
            if self.Rfree < 0.05:
                self.Rfree = None
                self.DPIf = None
        if self.R != None:
            if self.R < 0.05:
                self.R = None
                self.DPIa = None                    
        self.DPI = self.get_DPI()
        return 1
            
            
    
        
    def _get_mmcif_data(self, mmcif_file):
        # get the refinement data
        mmcif_obj = mmLib.mmCIF.mmCIFFile()
        mmcif_obj.load_file(fil = mmcif_file)
        mmcif_data = mmcif_obj[0]
        
        # extract the different types of data
        # check if None, or if correct type, if not correct type, maybe None, ? or ., then set to None
        self.dep_date = mmcif_data.get_tag('_database_PDB_rev.date_original')

        nobs = mmcif_data.get_tag('_refine.ls_number_reflns_obs')
        if nobs != None and nobs.isdigit() == True:
            self.nobs = int(nobs)
        else:
            nobs = mmcif_data.get_tag('_refine.ls_number_reflns_all')
            if nobs != None and nobs.isdigit() == True:
                self.nobs = int(nobs)
            else:
                self.nobs = None        

        cpl = mmcif_data.get_tag('_refine.ls_percent_reflns_obs')
        if cpl != None and self._isfloat(cpl) == True:
            self.cpl = float(cpl) / 100.0   # convert from percentage
        else:
            cpl = mmcif_data.get_tag('_reflns.percent_possible_obs')
            if cpl != None and self._isfloat(cpl) == True:
                self.cpl = float(cpl) / 100.0   # convert from percentage
            else:
                self.cpl = None


        R = mmcif_data.get_tag('_refine.ls_R_factor_R_work')
        if R != None and self._isfloat(R) == True:
            self.R = float(R)
        else:
            R = mmcif_data.get_tag('_refine.ls_R_factor_obs')
            if R != None and self._isfloat(R) == True:
                self.R = float(R)
            else:
                self.R = None

        Rfree = mmcif_data.get_tag('_refine.ls_R_factor_R_free')
        if Rfree != None and self._isfloat(Rfree) == True:
            self.Rfree = float(Rfree)
        else:
            self.Rfree = None

        dmin = mmcif_data.get_tag('_refine.ls_d_res_high')
        if dmin != None and self._isfloat(dmin) == True:
            self.dmin = float(dmin)
        else:
            dmin = mmcif_data.get_tag('_reflns.d_resolution_high')
            if dmin != None and self._isfloat(dmin) == True:
                self.dmin = float(dmin)
            else:
                self.dmin = None
        
        luzzati_obs = mmcif_data.get_tag('_refine_analyze.Luzzati_coordinate_error_obs')
        if luzzati_obs != None and self._isfloat(luzzati_obs) == True:
            self.luzzati_obs = float(luzzati_obs)
        else:
            self.luzzati_obs = None
        
        luzzati_free = mmcif_data.get_tag('_refine_analyze.Luzzati_coordinate_error_free')
        if luzzati_free != None and self._isfloat(luzzati_free) == True:
            self.luzzati_free = float(luzzati_free)
        else:
            self.luzzati_free = None
        
        self.luzzati = self.get_luzzati()
            
        sigma_a_obs = mmcif_data.get_tag('_refine_analyze.Luzzati_sigma_a_obs')
        if sigma_a_obs != None and self._isfloat(sigma_a_obs) == True:
            self.sigma_a_obs = float(sigma_a_obs)
        else:
            self.sigma_a_obs = None
            
        sigma_a_free = mmcif_data.get_tag('_refine_analyze.Luzzati_sigma_a_free')
        if sigma_a_free != None and self._isfloat(sigma_a_free) == True:
            self.sigma_a_free = float(sigma_a_free)
        else:
            self.sigma_a_free = None
        
        self.sigma_a = self.get_sigma_a()
            
        self.sigatm = False
        self.anisou = False
        if mmcif_data.has_key('atom_site'):
            atom_site = mmcif_data.get_table('atom_site')
            atom_site.autoset_columns()
            if atom_site.has_column('Cartn_x_esd'):
                esd_dic = atom_site.row_index_dict('Cartn_x_esd')
                if len(esd_dic.keys()) > 2:
                    # exclude ? or .
                    self.sigatm = True
            if atom_site.has_column('aniso_B[1][1]'):
                aniso_dic = atom_site.row_index_dict('aniso_B[1][1]')
                if len(aniso_dic) > 2:
                    self.anisou = True
            if self.anisou == False and atom_site.has_column('aniso_U[1][1]'):
                aniso_dic = atom_site.row_index_dict('aniso_U[1][1]')
                if len(aniso_dic) > 2:
                    self.anisou = True                
        if self.anisou == False and mmcif_data.has_key('atom_site_anisotrop'):
            self.anisou = True

                
        self._get_structure_data(mmcif_file)



    def _get_structure_data(self, mmcif_file):
        """
        get the data from the structure
        """
        str_obj = mmLib.FileIO.LoadStructure(fil = mmcif_file)
        # average B factor
        Ni = 0
        Bsum = 0.0
        for atom in str_obj.iter_all_atoms():       
            Ni += 1
            Bsum += atom.temp_factor
        if Ni > 0:
            self.Bavg = float(Bsum)/float(Ni)
        else:
            self.Bavg = None
        # Nr of coordinates
        self.Ni = Ni
        # Nr of paramters 4*Ni -> three coordinates + B value
        self.npar = self.Ni * 4
        
        # debug
        # Self Test Nr of coordinates
        TestNi = str_obj.count_all_atoms()
        if Ni != TestNi:
            log_util.log_error("Ni does not match, Ni: %d, TestNi: %d" % (Ni, TestNi))



    def _isfloat(self, query_str):
        """
        Test if variable is a float, or can be converted to float
        There nust be better solutions than this ...
        """
        try:
            float(query_str)
            return True
        except:
            return False
